#ifndef BBA_PAR_H
#define BBA_PAR_H

#include <set>
#include <optional>
#include <functional>
#include "bp_par.h"
#include "bba.h"
#include "shared.h"
#include "execution.h"
#include "output.h"
#include "dfa_seq.h"

namespace bba::par {
	struct block_initialize_task_t {
		size_t inputIdx;
		size_t blockIdx;
	};

	struct incomplete_range_t {
		struct iterator {
		private:
			size_t _value;
			std::set<size_t>::iterator _exclusionsIter;

		public:
			size_t& operator*() {
				return _value;
			}

			const size_t& operator*() const {
				return _value;
			}

			bool operator==(const iterator& other) const {
				return this->_value == other._value;
			}

			bool operator!=(const iterator& other) const {
				return this->_value != other._value;
			}

			friend struct incomplete_range_t;
		};

		size_t size;
		// TODO: This set is only used in a way that requires capacity for only AFFINITY - 1 elements. Should we optimize this?
		std::set<size_t> exclusions;

		void _findUnexcluded(iterator& iter) const {
			while (iter._exclusionsIter != this->exclusions.end() && *iter._exclusionsIter == iter._value) {
				++iter._exclusionsIter;
				++iter._value;
			}
		}

		iterator begin() const {
			iterator result;
			result._value = 0;
			result._exclusionsIter = this->exclusions.begin();
			this->_findUnexcluded(result);
			return result;
		}

		iterator end() const {
			iterator result;
			result._value = this->size;
			result._exclusionsIter = this->exclusions.end();
			return result;
		}

		void next(iterator& iter) const {
			++iter._value;
			this->_findUnexcluded(iter);
		}
	};

	struct probe_join_task_t {
		block_initialize_task_t block;
		incomplete_range_t probeRange;
	};

	template<typename elem_t, int DIMS, typename key_t, typename data_t>
	struct ranked_block_transform_join_task_t {
		struct input_t {
			size_t blockSize;
			size_t blockCount;
			record_t<elem_t, DIMS, key_t, data_t>* firstSource;
			record_t<elem_t, DIMS, key_t, data_t>* lastSource;
		};

		std::array<input_t, 2> inputs;
		// Each input has a set of blocks which have been allocated, but some allocated blocks may not be initialized.
		std::array<incomplete_range_t, 2> inputStates;

		// Given two inputs, selects the input to read that gives the best upper score bound above the threshold
		// Returns no value if both inputs are exhausted OR if the threshold cannot be satisfied
		template<typename score_func_t, typename score_cmp_t>
		std::optional<size_t> selectBestReadInput(score_func_t scoreFunc, score_cmp_t scoreCmp, key_t scoreThreshold) const {
			key_t bestScoreUB = scoreThreshold;
			std::optional<size_t> result{};
			for (int readInputIdx = 0; readInputIdx < 2; ++readInputIdx) {
				// is there another block of this input that we can read, or is it exhausted?
				auto& readInput = this->inputs[readInputIdx];
				auto& readInputState = this->inputStates[readInputIdx];
				if (readInputState.size < readInput.blockCount) {
					// there is another block, see what the upper bound will be if we read it
					int probeInputIdx = 1 - readInputIdx;
					auto& probeInput = this->inputs[probeInputIdx];

					// The top record of the block that we could read
					const record_t<elem_t, DIMS, key_t, data_t>* readUpper = readInput.firstSource + (readInputState.size * readInput.blockSize);
					// The top record of the input that we could probe
					const record_t<elem_t, DIMS, key_t, data_t>* probeUpper = probeInput.firstSource;

					const record_t<elem_t, DIMS, key_t, data_t>* uppers[2];
					uppers[readInputIdx] = readUpper;
					uppers[probeInputIdx] = probeUpper;

					key_t candidateScoreUB = scoreFunc(uppers[0]->key, uppers[1]->key);
					if (scoreCmp(candidateScoreUB, bestScoreUB)) {
						result = std::optional<size_t>(readInputIdx);
						bestScoreUB = candidateScoreUB;
					}
				}
			}

			return result;
		}

		// Allocates a block of the best input to initialize and returns a task to initialize it
		// Returns no task if both inputs are exhausted OR if the threshold cannot be satisfied
		template<typename score_func_t, typename score_cmp_t>
		std::optional<block_initialize_task_t> nextUninitializedBlock(score_func_t scoreFunc, score_cmp_t scoreCmp, key_t scoreThreshold) {
			std::optional<size_t> readInputIdx = this->selectBestReadInput(scoreFunc, scoreCmp, scoreThreshold);
			if (readInputIdx.has_value()) {
				size_t probeInputIdx = 1 - *readInputIdx;
				auto& readInput = this->inputs[*readInputIdx];
				auto& readInputState = this->inputStates[*readInputIdx];
				size_t blockIdx = readInputState.size;

				block_initialize_task_t subtask;
				subtask.inputIdx = *readInputIdx;
				subtask.blockIdx = blockIdx;

				readInputState.exclusions.insert(blockIdx);
				++readInputState.size;

				return std::optional<block_initialize_task_t>{subtask};
			}

			return {};
		}

		// When a block initialize task is complete, it must be submitted to the main task so that it can be joined
		// Returns a task to join the submitted block with the blocks in the other input
		// The returned task contains a snapshot of the range of initialized blocks from the other input
		probe_join_task_t submitInitializedBlock(block_initialize_task_t task) {
			probe_join_task_t result;
			result.block = task;
			auto& readInputState = this->inputStates[task.inputIdx];
			readInputState.exclusions.erase(task.blockIdx);
			size_t probeInputIdx = 1 - task.inputIdx;
			result.probeRange = this->inputStates[probeInputIdx];
			return result;
		}
	};

	template<typename elem_t, int DIMS, typename key_t, typename data_t, typename key_agg_t, int MAX_HEIGHT, typename spatial_pred_t, typename score_func_t, typename score_cmp_t>
	void probeJoinBlocks(
		const std::array<bba_internal_input_t<elem_t, DIMS, key_t, data_t, key_agg_t, MAX_HEIGHT>, 2>& inputs,
		probe_join_task_t task,
		spatial_pred_t spatialPred,
		score_func_t scoreFunc,
		const output::accumulator_config_t<score_cmp_t>& accConfig,
		output::accumulator_data_t<key_t, std::array<data_t, 2>>& accData
	) {
		size_t readInputIdx = task.block.inputIdx;
		size_t probeInputIdx = 1 - readInputIdx;
		const bba_internal_input_t<elem_t, DIMS, key_t, data_t, key_agg_t, MAX_HEIGHT>& readInput = inputs[readInputIdx];
		const bba_internal_input_t<elem_t, DIMS, key_t, data_t, key_agg_t, MAX_HEIGHT>& probeInput = inputs[probeInputIdx];

        std::array<rtree_layout_t<MAX_HEIGHT>, 2> layouts{
            inputs[0].layout,
            inputs[1].layout
        };
        std::array<rtree_t<elem_t, DIMS, key_t, data_t>, 2> rtrees;

		size_t readBlockIdx = task.block.blockIdx;
		// We will run sequential DFA internally to join each pair of blocks
		// dfaInputs[readInputIdx].rtree is constant, it is the rtree of the block we read
		rtrees[readInputIdx] = inputs[readInputIdx].rtrees[readBlockIdx];
		const record_t<elem_t, DIMS, key_t, data_t>* readUpper = readInput.records.first + (readBlockIdx * readInput.blockSize);
		for (incomplete_range_t::iterator i = task.probeRange.begin(); i != task.probeRange.end(); task.probeRange.next(i)) {
			size_t probeBlockIdx = *i;
			const record_t<elem_t, DIMS, key_t, data_t>* probeUpper = probeInput.records.first + (probeBlockIdx * probeInput.blockSize);

			const record_t<elem_t, DIMS, key_t, data_t>* uppers[2];
			uppers[task.block.inputIdx] = readUpper;
			uppers[probeInputIdx] = probeUpper;

			key_t scoreUB = scoreFunc(uppers[0]->key, uppers[1]->key);
			// scoreUB < accData.lowerBound
			if (!accConfig.scoreCmp(scoreUB, accData.lowerBound)) {
				// we are probing in decreasing order of score
				// we cannot satisfy the score threshold by joining with this block
				// it follows that we cannot satisfy the score threshold with any further blocks as well
				break;
			}

			// set the rtree of the probed block as the other input for DFA
			rtrees[probeInputIdx] = inputs[probeInputIdx].rtrees[probeBlockIdx];

			// run DFA to populate accData
			dfa::seq::runDfa(rtrees, layouts, spatialPred, scoreFunc, accConfig, accData);
		}
	}

	template<typename elem_t, int DIMS, typename key_t, typename data_t, typename key_agg_t, int MAX_HEIGHT, int AFFINITY, typename task_func_t, typename spatial_pred_t, typename score_func_t, typename score_cmp_t>
	void runBba(
		execution::fork_join_pool_t<AFFINITY, task_func_t>& forkJoinPool,
		const std::array<bba::bba_internal_input_t<elem_t, DIMS, key_t, data_t, key_agg_t, MAX_HEIGHT>, 2>& inputs,
        spatial_pred_t spatialPred,
		score_func_t scoreFunc,
		const output::accumulator_config_t<score_cmp_t>& accConfig,
		output::accumulator_data_t<key_t, std::array<data_t, 2>>& accData
	) {
		ranked_block_transform_join_task_t<elem_t, DIMS, key_t, data_t> task;
		for (int inputIdx = 0; inputIdx < 2; ++inputIdx) {
			auto& taskInput = task.inputs[inputIdx];
			const auto& input = inputs[inputIdx];

			taskInput.blockSize = input.blockSize;
			taskInput.blockCount = input.numBlocks;
			taskInput.firstSource = input.records.first;
			taskInput.lastSource = input.records.last;

			incomplete_range_t& taskInputState = task.inputStates[inputIdx];
			taskInputState.size = 0;
		}

		using data_pair_t = std::array<data_t, 2>;

		std::mutex taskMutex;
		std::mutex outputMutex;

		forkJoinPool.run([&](int tid) {
			// std::chrono::high_resolution_clock::duration totalMutexDuration { 0 };

			score_data_pair_t<key_t, data_pair_t>* localHeap = new score_data_pair_t<key_t, data_pair_t>[accConfig.capacity];

			std::unique_lock<std::mutex> outputLock(outputMutex);
			// Invariant: outputLock must be acquired on entry and exit
			while (true) {
				// auto taskMutexRequested = std::chrono::high_resolution_clock::now();
				std::unique_lock<std::mutex> taskLock(taskMutex);
				// totalMutexDuration += std::chrono::high_resolution_clock::now() - taskMutexRequested;
				// try to allocate a block of one of the inputs to initialize then join
				std::optional<block_initialize_task_t> initializeBlockTask = task.nextUninitializedBlock(scoreFunc, accConfig.scoreCmp, accData.lowerBound);
				taskLock.unlock();
				if (initializeBlockTask.has_value()) {
					// we managed to allocate a task to initialize a block that we can eventually join
					outputLock.unlock();

					const bba::bba_internal_input_t<elem_t, DIMS, key_t, data_t, key_agg_t, MAX_HEIGHT>& input = inputs[initializeBlockTask->inputIdx];
					record_t<elem_t, DIMS, key_t, data_t>* source = input.records.first + input.blockSize * initializeBlockTask->blockIdx;
					input.rtrees[initializeBlockTask->blockIdx] = str::seq::runStr(input.layout, input.keyAgg, source);

					taskLock.lock();
					probe_join_task_t probeJoinTask = task.submitInitializedBlock(*initializeBlockTask);
					taskLock.unlock();

					// we now have a task to join the block we initialized by probing the other input
					// we'll use a new local accumulator initialized with the lower bound of the shared accumulator
					outputLock.lock();
					output::accumulator_data_t localAccData(localHeap, 0, accData.lowerBound);
					outputLock.unlock();

					// now perform the joins to fill the local accumulator
					probeJoinBlocks(inputs, probeJoinTask, spatialPred, scoreFunc, accConfig, localAccData);

					// then insert the local accumulator data into the shared data
					outputLock.lock();
					accData.insertAll(localAccData, accConfig);
				}
				else {
					// we couldn't find any block to read that would produce any results above the threshold
					// this means that the blocks that could produce results have all been allocated
					// the only thing we can do is try to steal work from the other threads or just terminate here
					// TODO: Implement work stealing
					break;
				}
			}

			outputLock.unlock();

			delete[] localHeap;
		});

		for (size_t inputIdx = 0; inputIdx < 2; ++inputIdx) {
			size_t lastBlockIdx = task.inputStates[inputIdx].size;
			for (size_t blockIdx = 0; blockIdx < lastBlockIdx; ++blockIdx) {
				delete[] inputs[inputIdx].rtrees[blockIdx].records;
			}
		}
	}

    template<typename iter_t, typename cmp_t>
    void selectFirst(iter_t first, iter_t last, cmp_t cmp) {
        auto minElem = first;
        for (iter_t i = first + 1; i < last; ++i) {
            if (cmp(*minElem, *i)) {
                minElem = i;
            }
        }
        if (minElem != first) {
            std::swap(*first, *minElem);
        }
    }

    template<typename elem_t, int DIMS, typename key_t, typename data_t, typename key_cmp_t, typename key_agg_t, int MAX_HEIGHT, int AFFINITY, typename task_func_t>
    bba::bba_internal_input_t<elem_t, DIMS, key_t, data_t, key_agg_t, MAX_HEIGHT> prepare(
		execution::fork_join_pool_t<AFFINITY, task_func_t>& forkJoinPool,
        const bba::bba_input_t<elem_t, DIMS, key_t, data_t, key_cmp_t, key_agg_t>& input
    ) {
        rtree_layout_t<MAX_HEIGHT> layout = createRTreeLayout<MAX_HEIGHT>(input.blockSize, input.fanout, input.paddedFanout);

		size_t inputSize = input.records.last - input.records.first;
		assert(!(inputSize % input.blockSize) /* input size must be divisible by block size */);
		size_t numBlocks = ceilDiv((long)inputSize, (long)input.blockSize);
        rtree_t<elem_t, DIMS, key_t, data_t>* rtrees = new rtree_t<elem_t, DIMS, key_t, data_t>[numBlocks];

        range_t<record_t<elem_t, DIMS, key_t, data_t>*> records;
		records.first = new record_t<elem_t, DIMS, key_t, data_t>[inputSize];
		records.last = records.first + inputSize;
		// make a sorted copy by score so that we read blocks in descending rank order
		// std::copy(input.records.first, input.records.last, records.first);
		forkJoinPool.run([&](int tid) {
			size_t numItemsPerThread = ceilDiv(inputSize, AFFINITY);

			record_t<elem_t, DIMS, key_t, data_t>* firstSrc = input.records.first + numItemsPerThread * tid;
			record_t<elem_t, DIMS, key_t, data_t>* lastSrc = std::min(firstSrc + numItemsPerThread, input.records.last);
			record_t<elem_t, DIMS, key_t, data_t>* firstDst = records.first + numItemsPerThread * tid;
			std::copy(firstSrc, lastSrc, firstDst);
		});
		const key_cmp_t& keyCmp = input.keyCmp;
        auto cmp = [&keyCmp](const record_t<float, DIMS, int, int>& lhs, const record_t<float, DIMS, int, int>& rhs) {
            return keyCmp(lhs.key, rhs.key);
        };
		bp::par::segmentedBlockPartition(forkJoinPool, records.first, records.last, records.last - records.first, input.blockSize, cmp);
		
		// We need to set the first element of each block as the min element
		forkJoinPool.run([&](int tid) {
			size_t numBlocksPerThread = ceilDiv(numBlocks, AFFINITY);

			record_t<elem_t, DIMS, key_t, data_t>* firstBlock = records.first + numBlocksPerThread * input.blockSize * tid;
			record_t<elem_t, DIMS, key_t, data_t>* lastBlock = std::min(firstBlock + input.blockSize * numBlocksPerThread, records.last);
			for (record_t<elem_t, DIMS, key_t, data_t>* i = firstBlock; i < lastBlock; i += input.blockSize) {
				selectFirst(i, i + input.blockSize, cmp);
			}
		});

        return bba::bba_internal_input_t<elem_t, DIMS, key_t, data_t, key_agg_t, MAX_HEIGHT>(records, input.blockSize, numBlocks, input.keyAgg, rtrees, layout);
    }

    template<typename elem_t, int DIMS, typename key_t, typename data_t, typename key_cmp_t, typename key_agg_t, int MAX_HEIGHT, int AFFINITY, typename task_func_t>
    std::array<bba::bba_internal_input_t<elem_t, DIMS, key_t, data_t, key_agg_t, MAX_HEIGHT>, 2> prepare(
		execution::fork_join_pool_t<AFFINITY, task_func_t>& forkJoinPool,
		const std::array<bba::bba_input_t<elem_t, DIMS, key_t, data_t, key_cmp_t, key_agg_t>, 2>& inputs
	) {
        return {
            prepare<elem_t, DIMS, key_t, data_t, key_cmp_t, key_agg_t, MAX_HEIGHT>(forkJoinPool, inputs[0]),
            prepare<elem_t, DIMS, key_t, data_t, key_cmp_t, key_agg_t, MAX_HEIGHT>(forkJoinPool, inputs[1])
        };
    }
}

#endif //BBA_PAR_H
