#ifndef BBA_SEQ_H
#define BBA_SEQ_H

#include <array>
#include "bp_seq.h"
#include "dfa_seq.h"
#include "bba.h"

namespace bba::seq {
    template<typename elem_t, int DIMS, typename key_t, typename data_t>
    struct bba_input_reader_t {
        const record_t<elem_t, DIMS, key_t, data_t>* source;
        const record_t<elem_t, DIMS, key_t, data_t>* nextSource;
        const record_t<elem_t, DIMS, key_t, data_t>* lastSource;
        rtree_t<elem_t, DIMS, key_t, data_t>* rtrees;
        rtree_t<elem_t, DIMS, key_t, data_t>* nextRtree;
        rtree_t<elem_t, DIMS, key_t, data_t>* lastRtree;

        template<typename key_agg_t, int MAX_HEIGHT>
        void initialize(const bba::bba_internal_input_t<elem_t, DIMS, key_t, data_t, key_agg_t, MAX_HEIGHT>& input);
        void free();
        const key_t& getProbeScoreUB() const;
        const key_t& getReadScoreUB() const;
        bool canRead() const;
        template<typename key_agg_t, int MAX_HEIGHT>
        rtree_t<elem_t, DIMS, key_t, data_t> readNextBlock(const bba::bba_internal_input_t<elem_t, DIMS, key_t, data_t, key_agg_t, MAX_HEIGHT>& input);
    };

    template<typename elem_t, int DIMS, typename key_t, typename data_t>
    template<typename key_agg_t, int MAX_HEIGHT>
    void bba_input_reader_t<elem_t, DIMS, key_t, data_t>::initialize(const bba::bba_internal_input_t<elem_t, DIMS, key_t, data_t, key_agg_t, MAX_HEIGHT>& input) {
        this->source = input.records.first;
        this->nextSource = input.records.first;
        this->lastSource = input.records.last;

        this->rtrees = input.rtrees;
        this->nextRtree = input.rtrees;
        this->lastRtree = input.rtrees + input.numBlocks;
    }

    template<typename elem_t, int DIMS, typename key_t, typename data_t>
    void bba_input_reader_t<elem_t, DIMS, key_t, data_t>::free() {
        for (auto i = this->rtrees; i < this->nextRtree; ++i) {
            delete[] i->records;
        }
    }

    template<typename elem_t, int DIMS, typename key_t, typename data_t>
    const key_t& bba_input_reader_t<elem_t, DIMS, key_t, data_t>::getProbeScoreUB() const {
        return this->source->key;
    }

    template<typename elem_t, int DIMS, typename key_t, typename data_t>
    const key_t& bba_input_reader_t<elem_t, DIMS, key_t, data_t>::getReadScoreUB() const {
        return this->nextSource->key;
    }

    template<typename elem_t, int DIMS, typename key_t, typename data_t>
    bool bba_input_reader_t<elem_t, DIMS, key_t, data_t>::canRead() const {
        return this->nextRtree < this->lastRtree;
    }

    template<typename elem_t, int DIMS, typename key_t, typename data_t>
    template<typename key_agg_t, int MAX_HEIGHT>
    rtree_t<elem_t, DIMS, key_t, data_t> bba_input_reader_t<elem_t, DIMS, key_t, data_t>::readNextBlock(const bba::bba_internal_input_t<elem_t, DIMS, key_t, data_t, key_agg_t, MAX_HEIGHT>& input) {
        rtree_t<elem_t, DIMS, key_t, data_t> rtree = str::seq::runStr(input.layout, input.keyAgg, this->nextSource);
        *this->nextRtree = rtree;

        this->nextSource += input.blockSize;
        ++this->nextRtree;

        return rtree;
    }

    template<typename elem_t, int DIMS, typename key_t, typename data_t, typename score_func_t, typename score_cmp_t>
    std::optional<size_t> selectInput(const std::array<bba_input_reader_t<elem_t, DIMS, key_t, data_t>, 2>& readers, score_func_t scoreFunc, score_cmp_t scoreCmp, key_t scoreThreshold) {
        // find which input to read and which input to probe
        // the goal is maximize the upper bound for the combined score of any produced results
        std::optional<size_t> result{};
        key_t bestScoreUB = scoreThreshold;
        for (size_t i = 0; i < 2; ++i) {
            if (readers[i].canRead()) {
                // if we read a block from inputs[i] and probe it against read blocks of inputs[1 - i], this is the upper bound for the combined score any produced results
                key_t scoreUB = scoreFunc(readers[i].getReadScoreUB(), readers[1ll - i].getProbeScoreUB());
                if (scoreCmp(scoreUB, bestScoreUB)) {
                    result = std::optional<size_t>(i);
                    bestScoreUB = scoreUB;
                }
            }
        }
        return result;
    }

    template<typename elem_t, int DIMS, typename key_t, typename data_t, typename key_agg_t, typename spatial_pred_t, typename score_func_t, int MAX_HEIGHT, typename score_cmp_t>
    void runBba(
        const std::array<bba::bba_internal_input_t<elem_t, DIMS, key_t, data_t, key_agg_t, MAX_HEIGHT>, 2>& inputs,
        spatial_pred_t spatialPred,
        score_func_t scoreFunc,
		const output::accumulator_config_t<score_cmp_t>& accConfig,
		output::accumulator_data_t<key_t, std::array<data_t, 2>>& accData
    ) {
        std::array<bba_input_reader_t<elem_t, DIMS, key_t, data_t>, 2> readers;
        for (int i = 0; i < 2; ++i) {
            readers[i].initialize(inputs[i]);
        }

        std::array<rtree_layout_t<MAX_HEIGHT>, 2> layouts{
            inputs[0].layout,
            inputs[1].layout
        };
        std::array<rtree_t<elem_t, DIMS, key_t, data_t>, 2> rtrees;

        while (true) {
            std::optional<size_t> readInput = selectInput(readers, scoreFunc, accConfig.scoreCmp, accData.lowerBound);
            if (readInput.has_value()) {
				size_t probeInput = 1 - *readInput;
				rtrees[*readInput] = readers[*readInput].readNextBlock(inputs[*readInput]);

				for (auto i = readers[probeInput].rtrees; i < readers[probeInput].nextRtree; ++i) {
					rtrees[probeInput] = *i;
					dfa::seq::runDfa(rtrees, layouts, spatialPred, scoreFunc, accConfig, accData);
				}
            }
            else {
                break;
            }
        }

        for (int i = 0; i < 2; ++i) {
            readers[i].free();
        }
    }

    template<typename iter_t, typename cmp_t>
    void selectFirst(iter_t first, iter_t last, cmp_t cmp) {
        auto minElem = first;
        for (iter_t i = first + 1; i < last; ++i) {
            if (cmp(*i, *minElem)) {
                minElem = i;
            }
        }
        if (minElem != first) {
            std::swap(*first, *minElem);
        }
    }

    template<typename elem_t, int DIMS, typename key_t, typename data_t, typename key_cmp_t, typename key_agg_t, int MAX_HEIGHT>
    bba::bba_internal_input_t<elem_t, DIMS, key_t, data_t, key_agg_t, MAX_HEIGHT> prepare(
        const bba::bba_input_t<elem_t, DIMS, key_t, data_t, key_cmp_t, key_agg_t>& input
    ) {
        rtree_layout_t<MAX_HEIGHT> layout = createRTreeLayout<MAX_HEIGHT>(input.blockSize, input.fanout, input.paddedFanout);

		size_t inputSize = input.records.last - input.records.first;
		assert(!(inputSize % input.blockSize) /* input size must be divisible by block size */);
		size_t numBlocks = ceilDiv((long)inputSize, (long)input.blockSize);
        rtree_t<elem_t, DIMS, key_t, data_t>* rtrees = new rtree_t<elem_t, DIMS, key_t, data_t>[numBlocks];

        range_t<record_t<elem_t, DIMS, key_t, data_t>*> records;
		records.first = new record_t<elem_t, DIMS, key_t, data_t>[inputSize];
		records.last = records.first + inputSize;
		// make a sorted copy by score so that we read blocks in descending rank order
		std::copy(input.records.first, input.records.last, records.first);
		const key_cmp_t& keyCmp = input.keyCmp;
        auto cmp = [&keyCmp](const record_t<float, DIMS, int, int>& lhs, const record_t<float, DIMS, int, int>& rhs) {
            return keyCmp(lhs.key, rhs.key);
        };
        bp::seq::blockPartition(records.first, records.last, input.blockSize, cmp);

        // We need to set the first element of each block as the min element
        for (record_t<elem_t, DIMS, key_t, data_t>* i = records.first; i < records.last; i += input.blockSize) {
            selectFirst(i, i + input.blockSize, cmp);
        }

        return bba::bba_internal_input_t<elem_t, DIMS, key_t, data_t, key_agg_t, MAX_HEIGHT>(records, input.blockSize, numBlocks, input.keyAgg, rtrees, layout);
    }

    template<typename elem_t, int DIMS, typename key_t, typename data_t, typename key_cmp_t, typename key_agg_t, int MAX_HEIGHT>
    std::array<bba::bba_internal_input_t<elem_t, DIMS, key_t, data_t, key_agg_t, MAX_HEIGHT>, 2> prepare(
        const std::array<bba::bba_input_t<elem_t, DIMS, key_t, data_t, key_cmp_t, key_agg_t>, 2>& inputs
    ) {
        return {
            prepare<elem_t, DIMS, key_t, data_t, key_cmp_t, key_agg_t, MAX_HEIGHT>(inputs[0]),
            prepare<elem_t, DIMS, key_t, data_t, key_cmp_t, key_agg_t, MAX_HEIGHT>(inputs[1])
        };
    }
}

#endif // BBA_SEQ_H
