#ifndef BP_H
#define BP_H

#include <utility>

namespace bp {
	template<typename iter_t, typename cmp_t>
	struct segmented_block_partition_task_t {
		iter_t first;
		iter_t last;
		size_t segmentSize;
		size_t blockSize;
		size_t segmentCount;
		cmp_t cmp;

		segmented_block_partition_task_t(iter_t first, iter_t last, size_t segmentSize, size_t blockSize, size_t segmentCount, cmp_t cmp)
			: first(first)
			, last(last)
			, segmentSize(segmentSize)
			, blockSize(blockSize)
			, segmentCount(segmentCount)
			, cmp(cmp) {}
	};

	template<typename iter_t>
	struct block_partition_work_item_t {
		iter_t begin;
		size_t size;

		block_partition_work_item_t() = default;
		block_partition_work_item_t(iter_t begin, size_t size)
			: begin(begin)
			, size(size) {}
	};

	template<typename iter_t, typename cmp_t>
	void median3(iter_t first, iter_t mid, iter_t last, cmp_t cmp) {
		if (cmp(*mid, *first)) {
			std::swap(*mid, *first);
		}

		if (cmp(*last, *mid)) {
			std::swap(*last, *mid);

			if (cmp(*mid, *first)) {
				std::swap(*mid, *first);
			}
		}
	}

	template<typename iter_t, typename cmp_t>
	void guessMedian(iter_t first, iter_t mid, iter_t last, cmp_t cmp) {
		const size_t count = last - first;
		if (count >= 40) {
			// for 40 or more items, guess median as median of medians
			const size_t step = (count + 1) >> 3; // +1 can't overflow because range was made inclusive in caller
			const size_t twoStep = step << 1; // note: intentionally discards low-order bit
			median3(first, first + step, first + twoStep, cmp);
			median3(mid - step, mid, mid + step, cmp);
			median3(last - twoStep, last - step, last, cmp);
			median3(first + step, mid, last - step, cmp);
		}
		else {
			median3(first, mid, last, cmp);
		}
	}

	template<typename iter_t, typename cmp_t>
	std::pair<iter_t, iter_t> partition(iter_t first, iter_t mid, iter_t last, cmp_t cmp) {
		iter_t pFirst = mid;
		iter_t pLast = pFirst + 1;

		while (first < pFirst && !cmp(*(pFirst - 1), *pFirst) && !cmp(*pFirst, *(pFirst - 1))) {
			--pFirst;
		}

		while (pLast < last && !cmp(*pLast, *pFirst) && !cmp(*pFirst, *pLast)) {
			++pLast;
		}

		iter_t gFirst = pLast;
		iter_t gLast = pFirst;

		for (;;) { // partition
			for (; gFirst < last; ++gFirst) {
				if (cmp(*pFirst, *gFirst)) {
				}
				else if (cmp(*gFirst, *pFirst)) {
					break;
				}
				else if (pLast != gFirst) {
					std::swap(*pLast, *gFirst);
					++pLast;
				}
				else {
					++pLast;
				}
			}

			for (; first < gLast; --gLast) {
				if (cmp(*(gLast - 1), *pFirst)) {
				}
				else if (cmp(*pFirst, *(gLast - 1))) {
					break;
				}
				else if (--pFirst != gLast - 1) {
					std::swap(*pFirst, *(gLast - 1));
				}
			}

			if (gLast == first && gFirst == last) {
				return std::pair<iter_t, iter_t>(pFirst, pLast);
			}

			if (gLast == first) { // no room at bottom, rotate pivot upward
				if (pLast != gFirst) {
					std::swap(*pFirst, *pLast);
				}

				++pLast;
				std::swap(*pFirst, *gFirst);
				++pFirst;
				++gFirst;
			}
			else if (gFirst == last) { // no room at top, rotate pivot downward
				if (--gLast != --pFirst) {
					std::swap(*gLast, *pFirst);
				}

				std::swap(*pFirst, *(--pLast));
			}
			else {
				std::swap(*gFirst, *(--gLast));
				++gFirst;
			}
		}
	}

	template<typename iter_t>
	bool crossesBlockBoundary(iter_t basis, size_t blockSize, iter_t first, iter_t last) {
		size_t firstBlock = (first - basis) / blockSize;
		size_t lastBlock = (last - basis) / blockSize;
		return firstBlock != lastBlock;
	}

	const int MIN_SORT_ITEMS = 32;

	template<typename iter_t, typename cmp_t>
	int processAndDivideWorkItem(block_partition_work_item_t<iter_t> workItem, block_partition_work_item_t<iter_t> nextWorkItems[2], iter_t basis, size_t blockSize, cmp_t cmp) {
		if (workItem.size >= MIN_SORT_ITEMS) {
			// partition
			size_t halfSize = workItem.size >> 1;
			iter_t begin = workItem.begin;
			iter_t mid = workItem.begin + halfSize;
			iter_t end = workItem.begin + workItem.size;

			// guess a median, places the median at *mid
			guessMedian(begin, mid, end - 1, cmp);

			std::pair<iter_t, iter_t> trueMid = partition(begin, mid, end, cmp);

			int i = 0;
			if (crossesBlockBoundary(basis, blockSize, begin, trueMid.first - 1)) {
				nextWorkItems[i].begin = begin;
				nextWorkItems[i].size = trueMid.first - begin;
				++i;
			}
			if (crossesBlockBoundary(basis, blockSize, trueMid.second, end - 1)) {
				nextWorkItems[i].begin = trueMid.second;
				nextWorkItems[i].size = end - trueMid.second;
				++i;
			}

			return i;
		}
		else {
			std::sort(workItem.begin, workItem.begin + workItem.size, cmp);
			return 0;
		}
	}
}

#endif //BP_H
