#ifndef BP_PAR_H
#define BP_PAR_H

#include <optional>
#include "execution.h"
#include "bp.h"

namespace bp::par {
	template<typename work_scheduler_t, typename iter_t, typename cmp_t>
	void processWorkItem(
		work_scheduler_t& workScheduler,
		const bp::segmented_block_partition_task_t<iter_t, cmp_t>& task,
		std::atomic<size_t>& remainingWork,
		bp::block_partition_work_item_t<iter_t> initialWorkItem
	) {
		bp::block_partition_work_item_t<iter_t> workItem = initialWorkItem;
		bp::block_partition_work_item_t<iter_t> nextWorkItems[2];
		while (true) {
			int numNextSubtasks = bp::processAndDivideWorkItem(workItem, nextWorkItems, task.first, task.blockSize, task.cmp);
			if (numNextSubtasks > 0) {
				workItem = nextWorkItems[0];
			}
			else {
				break;
			}
			for (int i = 1; i < numNextSubtasks; ++i) {
				++remainingWork;
				workScheduler.submit(nextWorkItems[i]);
			}
		}
		--remainingWork;
	}

	template<typename work_scheduler_t, typename iter_t, typename cmp_t>
	void exhaustWorkItem(
		work_scheduler_t& workScheduler,
		const bp::segmented_block_partition_task_t<iter_t, cmp_t>& task,
		std::atomic<size_t>& remainingWork,
		bp::block_partition_work_item_t<iter_t> initialWorkItem
	) {
		std::optional<bp::block_partition_work_item_t<iter_t>> workItem(initialWorkItem);
		while (workItem.has_value()) {
			processWorkItem(workScheduler, task, remainingWork, *workItem);
			workItem = workScheduler.take();
		}
	}

	template<typename iter_t, typename cmp_t>
	std::optional<block_partition_work_item_t<iter_t>> produceWorkItem(std::atomic<size_t>& nextSegmentIdx, const bp::segmented_block_partition_task_t<iter_t, cmp_t>& task) {
		std::optional<size_t> segmentIdx = nextSegmentIdx.fetch_add(1);
		if (segmentIdx < task.segmentCount) {
			iter_t first = task.first + *segmentIdx * task.segmentSize;
			iter_t last = std::min(first + task.segmentSize, task.last);
			bp::block_partition_work_item_t<iter_t> workItem(first, last - first);
			return std::optional<bp::block_partition_work_item_t<iter_t>>(workItem);
		}
		return {};
	}

	template<typename iter_t, typename cmp_t>
	void segmentedBlockPartitionWorker(
		execution::work_stealing_membership_t<bp::block_partition_work_item_t<iter_t>>& workScheduler,
		const bp::segmented_block_partition_task_t<iter_t, cmp_t>& task,
		std::atomic<size_t>& nextSegmentIdx,
		std::atomic<size_t>& remainingWork
	) {
		// process as many segments as possible until there are none left
		while (true) {
			std::optional<block_partition_work_item_t<iter_t>> workItem = produceWorkItem(nextSegmentIdx, task);
			if (workItem.has_value()) {
				exhaustWorkItem(workScheduler, task, remainingWork, *workItem);
			}
			else {
				break;
			}
		}

		// spin and try to steal subtasks until there is no more remaining work
		while (remainingWork.load() > 0) {
			std::optional<bp::block_partition_work_item_t<iter_t>> workItem = workScheduler.steal();
			if (workItem.has_value()) {
				exhaustWorkItem(workScheduler, task, remainingWork, *workItem);
			}
		}
	}

	template<typename iter_t, typename cmp_t>
	struct segmented_block_partition_input_t {
		iter_t first;
		iter_t last;
		size_t segmentSize;
		size_t blockSize;
		cmp_t cmp;

		segmented_block_partition_input_t(iter_t first, iter_t last, size_t segmentSize, size_t blockSize, cmp_t cmp)
			: first(first)
			, last(last)
			, segmentSize(segmentSize)
			, blockSize(blockSize)
			, cmp(cmp) {}
	};

	template<typename iter_t, typename cmp_t>
	struct segmented_block_partition_execution_t {
		execution::work_stealing_team<bp::block_partition_work_item_t<iter_t>> team;
		std::atomic<size_t> nextSegmentIdx;
		segmented_block_partition_task_t<iter_t, cmp_t> task;

		// remainingWork roughly counts the number of work items that are currently accounted for
		// whenever remainingWork > 0, there may be work items that can be stolen (but not always)
		// when remainingWork reaches zero, there are no more work items, and the task is complete
		// we initialize it with the number of initial work items, which is equal to the total amount of segments
		std::atomic<size_t> remainingWork;

		segmented_block_partition_execution_t(int affinity, size_t segmentCount, const segmented_block_partition_input_t<iter_t, cmp_t>& input)
			: team(affinity)
			, nextSegmentIdx(0)
			, task(input.first, input.last, input.segmentSize, input.blockSize, segmentCount, input.cmp)
			, remainingWork(segmentCount) {
		}

		static segmented_block_partition_execution_t<iter_t, cmp_t> create(int affinity, const segmented_block_partition_input_t<iter_t, cmp_t>& input) {
			assert(input.segmentSize % input.blockSize == 0 /* segmentSize must be a multiple of blockSize */);
			size_t segmentCount = ceilDiv(input.last - input.first, input.segmentSize);
			return segmented_block_partition_execution_t<iter_t, cmp_t>(affinity, segmentCount, input);
		}
	};

	// Performs a segmented block partitioning operation
	// The input is divided into contiguous segments where a block partitioning operation is performed on each segment
	// A block partitioning operation under a comparator is a partial sort that divides the input into a series of blocks
	// All entries in a block in the resulting series of blocks are smaller than all entries in all following blocks
	template<int AFFINITY, typename task_func_t, typename iter_t, typename cmp_t>
	void segmentedBlockPartition(
		execution::fork_join_pool_t<AFFINITY, task_func_t>& forkJoinPool,
		iter_t first,
		iter_t last,
		size_t segmentSize,
		size_t blockSize,
		cmp_t cmp
	) {
		segmented_block_partition_input_t<iter_t, cmp_t> input(first, last, segmentSize, blockSize, cmp);
		segmented_block_partition_execution_t<iter_t, cmp_t> execution = segmented_block_partition_execution_t<iter_t, cmp_t>::create(AFFINITY, input);

		task_func_t taskFunc([&](int tid) {
			execution::work_stealing_membership_t<bp::block_partition_work_item_t<iter_t>> membership(tid, execution.team);
			segmentedBlockPartitionWorker(membership, execution.task, execution.nextSegmentIdx, execution.remainingWork);
		});

		forkJoinPool.run(taskFunc);
	}

	// Performs a segmented block partitioning operation
	// The input is divided into contiguous segments where a block partitioning operation is performed on each segment
	// A block partitioning operation under a comparator is a partial sort that divides the input into a series of blocks
	// All entries in a block in the resulting series of blocks are smaller than all entries in all following blocks
	template<int AFFINITY, typename task_func_t, typename iter_t, typename cmp_t>
	void multiSegmentedBlockPartition(
		execution::fork_join_pool_t<AFFINITY, task_func_t>& forkJoinPool,
		const std::array<segmented_block_partition_input_t<iter_t, cmp_t>, 2>& inputs
	) {
		const int THREADS_PER_INPUT = ceilDiv(AFFINITY, 2);

		std::array<segmented_block_partition_execution_t<iter_t, cmp_t>, 2> executions {
			segmented_block_partition_execution_t<iter_t, cmp_t>::create(THREADS_PER_INPUT, inputs[0]),
			segmented_block_partition_execution_t<iter_t, cmp_t>::create(THREADS_PER_INPUT, inputs[1])
		};

		task_func_t taskFunc([&](int tid) {
			for (int i = tid; i < 2 * THREADS_PER_INPUT; i += AFFINITY) {
				std::div_t div = std::div(i, THREADS_PER_INPUT);
				int inputIdx = div.quot;
				int innerTid = div.rem;

				segmented_block_partition_execution_t<iter_t, cmp_t>& execution = executions[inputIdx];
				execution::work_stealing_membership_t<bp::block_partition_work_item_t<iter_t>> workScheduler(innerTid, execution.team);
				segmentedBlockPartitionWorker(workScheduler, execution.task, execution.nextSegmentIdx, execution.remainingWork);
			}
		});

		forkJoinPool.run(taskFunc);
	}
}

#endif //BP_PAR_H
