#ifndef DFA_SEQ_H
#define DFA_SEQ_H

#include <array>
#include <queue>
#include <algorithm>
#include <functional>
#include "str_seq.h"
#include "str_par.h"
#include "shared.h"
#include "output.h"

namespace dfa::seq {
    template<typename score_t>
    struct dfa_queue_entry_t {
        score_t score;
        std::array<size_t, 2> levels;
        std::array<size_t, 2> items;
    };

    template<typename elem_t, int DIMS, typename key_t, typename data_t, int MAX_HEIGHT>
    struct dfa_internal_input_t {
        rtree_t<elem_t, DIMS, key_t, data_t> rtree;
        rtree_layout_t<MAX_HEIGHT> layout;

        dfa_internal_input_t(rtree_t<elem_t, DIMS, key_t, data_t> rtree, rtree_layout_t<MAX_HEIGHT> layout)
            : rtree(rtree)
            , layout(layout) {}
    };

    template<typename elem_t, int DIMS, typename key_t, typename data_t, typename spatial_pred_t, typename score_func_t, typename score_cmp_t, typename queue_cmp_t>
    void dfaJoinPair(
        const std::array<rtree_t<elem_t, DIMS, key_t, data_t>, 2>& inputs,
        std::array<size_t, 2> levels,
        std::array<size_t, 2> items,
        spatial_pred_t spatialPred,
        score_func_t scoreFunc,
        std::priority_queue<dfa_queue_entry_t<key_t>, std::vector<dfa_queue_entry_t<key_t>>, queue_cmp_t>& queue,
        const output::accumulator_config_t<score_cmp_t>& accConfig,
        output::accumulator_data_t<key_t, std::array<data_t, 2>>& accData
    ) {
        std::array<const record_t<elem_t, DIMS, key_t, data_t>*, 2> records = {
            &inputs[0].records[items[0]],
            &inputs[1].records[items[1]]
        };
        if (spatialPred(records[0]->mbb, records[1]->mbb)) {
            key_t score = scoreFunc(records[0]->key, records[1]->key);
            if (accConfig.scoreCmp(score, accData.lowerBound)) {
                if (levels[0] == 0 && levels[1] == 0) {
                    // leaf entries, insert in output
                    score_data_pair_t<key_t, std::array<data_t, 2>> outputValue;
                    outputValue.score = score;
                    outputValue.data[0] = records[0]->data;
                    outputValue.data[1] = records[1]->data;
                    accData.insert(outputValue, accConfig);
                }
                else {
                    // inner entries, insert a child in the queue for expansion later
                    dfa_queue_entry_t<key_t> queueEntry;
                    queueEntry.score = score;
                    queueEntry.levels = levels;
                    queueEntry.items = items;
                    queue.push(queueEntry);
                }
            }
        }
    }

    template<typename elem_t, int DIMS, typename key_t, typename data_t, typename spatial_pred_t, typename score_func_t, typename score_cmp_t, typename queue_cmp_t>
    void dfaJoinRanges(
        const std::array<rtree_t<elem_t, DIMS, key_t, data_t>, 2>& inputs,
        std::array<size_t, 2> levels,
        std::array<size_t, 2> firstItems,
        std::array<size_t, 2> lastItems,
        spatial_pred_t spatialPred,
        score_func_t scoreFunc,
        std::priority_queue<dfa_queue_entry_t<key_t>, std::vector<dfa_queue_entry_t<key_t>>, queue_cmp_t>& queue,
        const output::accumulator_config_t<score_cmp_t>& accConfig,
        output::accumulator_data_t<key_t, std::array<data_t, 2>>& accData
    ) {
        std::array<size_t, 2> items;
        for (items[0] = firstItems[0]; items[0] < lastItems[0]; ++items[0]) {
            for (items[1] = firstItems[1]; items[1] < lastItems[1]; ++items[1]) {
                dfaJoinPair(inputs, levels, items, spatialPred, scoreFunc, queue, accConfig, accData);
            }
        }
    }

    template<typename elem_t, int DIMS, typename key_t, typename data_t, typename spatial_pred_t, typename score_func_t, typename score_cmp_t, typename queue_cmp_t, int MAX_HEIGHT>
    void dfaExpand(
        const std::array<rtree_t<elem_t, DIMS, key_t, data_t>, 2>& inputs,
        const std::array<rtree_layout_t<MAX_HEIGHT>, 2>& layouts,
        const std::array<size_t, 2>& levels,
        const std::array<size_t, 2>& items,
        spatial_pred_t spatialPred,
        score_func_t scoreFunc,
        std::priority_queue<dfa_queue_entry_t<key_t>, std::vector<dfa_queue_entry_t<key_t>>, queue_cmp_t>& queue,
        const output::accumulator_config_t<score_cmp_t>& accConfig,
        output::accumulator_data_t<key_t, std::array<data_t, 2>>& accData
    ) {
        std::array<size_t, 2> childLevels;
        std::array<size_t, 2> firstChildItems;
        std::array<size_t, 2> lastChildItems;
        for (int i = 0; i < 2; ++i) {
            size_t level = levels[i];
            if (level > 0) {
                --level;
                const record_t<elem_t, DIMS, key_t, data_t>& record = inputs[i].records[items[i]];
                segment_t segment = layouts[i].segments[level];
                size_t firstItem = segment.offset + (size_t)record.childBlock * (size_t)layouts[i].fanout;
                size_t lastItem = firstItem + layouts[i].fanout;
                firstChildItems[i] = firstItem;
                lastChildItems[i] = std::min(lastItem, segment.offset + segment.size);
            }
            else {
                firstChildItems[i] = items[i];
                lastChildItems[i] = items[i] + 1;
            }
            childLevels[i] = level;
        }

        dfaJoinRanges(inputs, childLevels, firstChildItems, lastChildItems, spatialPred, scoreFunc, queue, accConfig, accData);
    }

    template<typename elem_t, int DIMS, typename key_t, typename data_t, typename spatial_pred_t, typename score_func_t, typename score_cmp_t, int MAX_HEIGHT>
    void runDfa(
        const std::array<rtree_t<elem_t, DIMS, key_t, data_t>, 2>& inputs,
        const std::array<rtree_layout_t<MAX_HEIGHT>, 2>& layouts,
        spatial_pred_t spatialPred,
        score_func_t scoreFunc,
        const output::accumulator_config_t<score_cmp_t>& accConfig,
        output::accumulator_data_t<key_t, std::array<data_t, 2>>& accData
    ) {
        // If the roots don't intersect or can't satisfy the lower bound of the score, we can terminate immediately
        if (!spatialPred(inputs[0].root.mbb, inputs[1].root.mbb)) {
            key_t score = scoreFunc(inputs[0].root.key, inputs[1].root.key);
            if (accConfig.scoreCmp(score, accData.lowerBound)) {
                return;
            }
        }

        // TODO: Can we increase efficiency with a special case for inputs[0].layout->height == inputs[1].layout->height?

        const auto scoreCmp = accConfig.scoreCmp;
        const auto queueCmp = [scoreCmp](const dfa_queue_entry_t<key_t>& lhs, const dfa_queue_entry_t<key_t>& rhs) {
            // the queue uses the inverse order, it will always find the greatest element using the less function, we want the opposite
            return scoreCmp(rhs.score, lhs.score);
        };
        using queue_cmp_t = decltype(queueCmp);
        std::priority_queue<dfa_queue_entry_t<key_t>, std::vector<dfa_queue_entry_t<key_t>>, queue_cmp_t> queue(queueCmp);

        {
            // Visit root pairs first
            std::array<size_t, 2> levels;
            std::array<size_t, 2> firstItems;
            std::array<size_t, 2> lastItems;
            for (int i = 0; i < 2; ++i) {
                size_t level = layouts[i].height - 1;
                levels[i] = level;
                segment_t segment = layouts[i].segments[level];
                firstItems[i] = segment.offset;
                lastItems[i] = segment.offset + segment.size;
            }
            dfaJoinRanges(inputs, levels, firstItems, lastItems, spatialPred, scoreFunc, queue, accConfig, accData);
        }

        while (!queue.empty()) {
            dfa_queue_entry_t<key_t> entry = queue.top();
            if (!accConfig.scoreCmp(entry.score, accData.lowerBound)) {
                // the top entry has a score lower than the lower bound
                // all other entries in the priority queue have a lower score than the top entry, and must therefore also have scores lower than the lower bound
                // any undiscovered children must also have scores lower than the lower bound
                // we can therefore terminate here
                return;
            }
            queue.pop();

            dfaExpand(inputs, layouts, entry.levels, entry.items, spatialPred, scoreFunc, queue, accConfig, accData);
        }
    }
}

#endif // DFA_SEQ_H
