#ifndef EXECUTION_H
#define EXECUTION_H

#include <mutex>
#include <condition_variable>
#include <bitset>
#include <thread>
#include <array>
#include <atomic>
#include <memory>
#include <optional>

namespace execution {
	enum class task_status_t {
		WAIT,
		RUN_TASK,
		TERMINATE
	};

	struct flex_barrier_t {
		std::mutex mut;
		std::condition_variable cv;
		int numThreads;
		int count;

		flex_barrier_t(int numThreads)
			: numThreads(numThreads)
			, count(0) {}

		template<typename completion_func_t>
		void arrive(completion_func_t completionFunc) {
		}
	};

	template<int AFFINITY, typename task_func_t>
	struct fork_join_pool_t {
		std::array<std::thread, AFFINITY - 1> threads;
		std::mutex mut;
		char rawTask[sizeof(task_func_t)];
		task_status_t status;

		std::condition_variable statusCv;
		std::condition_variable completeCv;
		std::bitset<AFFINITY> completedMask;

		void _completeBarrier(int tid, std::unique_lock<std::mutex>& lock) {
			assert(!this->completedMask.test(tid));
			this->completedMask.set(tid);
			assert(this->status == task_status_t::RUN_TASK);
			if (this->completedMask.all()) {
				this->status = task_status_t::WAIT;
				this->completeCv.notify_all();
			}
			else {
				this->completeCv.wait(lock);
			}
		}

		fork_join_pool_t()
			: status(task_status_t::WAIT) {
			completedMask.set();

			for (int tid = 0; tid < AFFINITY - 1; ++tid) {
				this->threads[tid] = std::thread([this, tid]() {
					std::unique_lock lock(this->mut);
					while (true) {
						if (this->status == task_status_t::RUN_TASK) {
							lock.unlock();
							task_func_t* task = reinterpret_cast<task_func_t*>(this->rawTask);
							(*task)(tid);
							lock.lock();
							this->_completeBarrier(tid, lock);
						}
						else if (this->status == task_status_t::WAIT) {
							this->statusCv.wait(lock);
						}
						else if (this->status == task_status_t::TERMINATE) {
							break;
						}
					}
					});
			}
		}

		~fork_join_pool_t() {
			{
				std::unique_lock lock(this->mut);
				this->status = task_status_t::TERMINATE;
			}
			this->statusCv.notify_all();
			for (int tid = 0; tid < AFFINITY - 1; ++tid) {
				this->threads[tid].join();
			}
		}

		void run(task_func_t task) {
			std::unique_lock lock(this->mut);
			this->status = task_status_t::RUN_TASK;
			assert(this->completedMask.all() /* can only run task is no task is currently active */);
			this->completedMask.reset();
			task_func_t* theTask = new (this->rawTask) task_func_t(task);
			lock.unlock();
			this->statusCv.notify_all();

			(*theTask)(AFFINITY - 1);

			lock.lock();
			this->_completeBarrier(AFFINITY - 1, lock);
			theTask->~task_func_t();
			lock.unlock();
		}
	};

	template<typename item_t>
	struct circular_buffer_t {
		size_t logSize;
		item_t* items;

		explicit circular_buffer_t(size_t logSize)
			: logSize(logSize)
			, items(new item_t[static_cast<size_t>(1) << logSize]) {}

		circular_buffer_t(const circular_buffer_t& other)
			: logSize(other.logSize)
			, items(new size_t[other.logSize]) {
			std::copy_n(other.items, 1 << logSize, this->items);
		}
		circular_buffer_t(circular_buffer_t&& other) noexcept
			: logSize(other.logSize)
			, items(other.items) {
			other.items = nullptr;
		}

		~circular_buffer_t() {
			delete[] this->items;
		}

		size_t capacity() const {
			return static_cast<size_t>(1) << this->logSize;
		}

		item_t& operator[](size_t idx) {
			size_t mask = (static_cast<size_t>(1) << this->logSize) - 1;
			return this->items[idx & mask];
		}

		circular_buffer_t grow(size_t bottom, size_t top) {
			circular_buffer_t other(logSize + 1);
			for (size_t i = top; i < bottom; ++i) {
				other[i] = this->operator[](i);
			}
			return other;
		}
	};

	template<typename work_item_t>
	struct alignas(64) work_queue_t {
		// presents the bottom index of the circular buffer, which must always be >= top
		// the owning thread inserts work items at the bottom and takes work items from the bottom
		std::atomic<size_t> bottom;
		// presents the top index of the circular buffer, which must always be <= bottom
		// stealing threads steal work items from the top
		std::atomic<size_t> top;
		// used to guard the buffer reference. the buffer may be swapped when grown
		std::mutex mut;
		// reference to the current buffer, is swapped when grown
		std::shared_ptr<circular_buffer_t<work_item_t>> buffer;

		work_queue_t()
			: bottom(0)
			, top(0)
			, buffer(std::make_shared<circular_buffer_t<work_item_t>>(6)) {}

		void submit(work_item_t workItem) {
			size_t localBottom = this->bottom.load();
			size_t localTop = this->top.load();
			size_t size = localBottom - localTop;

			if (size >= this->buffer->capacity()) {
				// we must grow the buffer and swap it
				auto newBuffer = std::make_shared<circular_buffer_t<work_item_t>>(this->buffer->grow(localBottom, localTop));
				{
					std::lock_guard<std::mutex> lock(this->mut);
					this->buffer = newBuffer;
				}
			}

			(*this->buffer)[localBottom] = workItem;
			this->bottom.store(localBottom + 1);
		}

		std::optional<work_item_t> take() {
			size_t localBottom = this->bottom.load();
			// localBottom == 0 implies that there is nothing in the queue to take. it would overflow below
			if (localBottom == 0) {
				return {};
			}

			--localBottom;
			this->bottom.store(localBottom);
			size_t localTop = this->top.load();
			if (localBottom < localTop) {
				// all items were stolen since we moved the bottom. there is nothing to take.
				this->bottom.store(localTop);
				return {};
			}

			std::optional<work_item_t> result((*this->buffer)[localBottom]);

			// are we competing with a thread that could be stealing the entry?
			if (localBottom > localTop) {
				// there is a stealable entry before the one we just read, so we are not competing with another thread
				return result;
			}

			// there was a single item in the queue as we started taking, so we might be competing with a stealing thread
			// we therefore try to steal the only item
			size_t desiredTop = localTop + 1;
			if (this->top.compare_exchange_strong(localTop, desiredTop)) {
				// we stole the only item successfully
				this->bottom.store(desiredTop);
				return result;
			}
			else {
				// another thread stole the item before us
				this->bottom.store(localTop);
				return {};
			}
		}

		std::optional<work_item_t> steal() {
			size_t localTop = this->top.load();
			size_t desiredTop;
			while (true) {
				if (localTop >= this->bottom.load()) {
					return {};
				}

				std::optional<work_item_t> result;
				{
					// the producer thread may currently be growing the buffer
					// grab a reference to a valid buffer
					// we use a block here to keep the reference to the buffer as short-lived as possible
					std::shared_ptr<circular_buffer_t<work_item_t>> buffer;
					{
						std::lock_guard<std::mutex> lock(this->mut);
						buffer = this->buffer;
					}
					result = std::optional<work_item_t>((*buffer)[localTop]);
				}

				desiredTop = localTop + 1;
				// check if there was no data race
				if (this->top.compare_exchange_strong(localTop, desiredTop)) {
					return result;
				}
				// there was a data race, we'll keep trying
			}
		}
	};

	template<typename work_item_t>
	struct work_stealing_team {
		int affinity;
		std::unique_ptr<work_queue_t<work_item_t>[]> workQueues;

		explicit work_stealing_team(int affinity)
			: affinity(affinity)
			, workQueues(new work_queue_t<work_item_t>[affinity]) {}
	};

	template<typename work_item_t>
	struct work_stealing_membership_t {
		int tid;
		work_stealing_team<work_item_t>& team;

		work_stealing_membership_t(int tid, work_stealing_team<work_item_t>& team)
			: tid(tid)
			, team(team) {}

		std::optional<work_item_t> take() const {
			return this->team.workQueues[this->tid].take();
		}

		void submit(work_item_t workItem) const {
			this->team.workQueues[this->tid].submit(workItem);
		}

		std::optional<work_item_t> steal() const {
			for (int i = tid;;) {
				if (i == 0) {
					i = team.affinity;
				}
				--i;

				if (i == tid) {
					return {};
				}

				std::optional<work_item_t> workItem = this->team.workQueues[i].steal();
				if (workItem.has_value()) {
					return workItem;
				}
			}
		}
	};
}

#endif //EXECUTION_H
