#include <stdio.h>
#include <cstdlib>
#include <cassert>
#include <random>
#include <chrono>
#include <queue>
#include <array>
#include <string>
#include "shared.h"
#include "output.h"
#include "dfa_seq.h"
#include "bba_seq.h"
#include "bba_par.h"
#include "bnb_seq.h"
#include "validation.h"


template<int DIMS, typename key_distribution_t, typename record_iter_t>
void generateRandomLeafRecords(record_iter_t first, record_iter_t last, std::mt19937& random, std::uniform_real_distribution<float> elemDistribution, key_distribution_t keyDistribution) {
    int data = 0;
    for (record_iter_t i = first; i < last; ++i, ++data) {
        for (int dim = 0; dim < DIMS; ++dim) {
            float coord = elemDistribution(random);
            i->mbb.min[dim] = coord;
            i->mbb.max[dim] = coord;
        }
        i->key = keyDistribution(random);
        i->data = data;
    }
}

template<typename elem_t, int DIMS, typename key_t, typename data_t, typename key_cmp_t, typename key_agg_t>
struct ranked_join_input_t {
    range_t<record_t<elem_t, DIMS, key_t, data_t>*> records;
    key_cmp_t keyCmp;
    key_agg_t keyAgg;

    ranked_join_input_t(range_t<record_t<elem_t, DIMS, key_t, data_t>*> records, key_cmp_t keyCmp, key_agg_t keyAgg)
        : records(records)
        , keyCmp(keyCmp)
        , keyAgg(keyAgg) {}
};

template<typename elem_t, int DIMS, typename key_t, typename data_t, typename key_cmp_t, typename key_agg_t, typename score_func_t, typename score_cmp_t, typename spatial_pred_t>
struct top_k_spatial_join_t {
    std::array<ranked_join_input_t<elem_t, DIMS, key_t, data_t, key_cmp_t, key_agg_t>, 2> inputs;
    key_t minScore;
    score_func_t scoreFunc;
    spatial_pred_t spatialPred;
    output::accumulator_config_t<score_cmp_t> accConfig;

    top_k_spatial_join_t(
        std::array<ranked_join_input_t<elem_t, DIMS, key_t, data_t, key_cmp_t, key_agg_t>, 2> inputs,
        key_t minScore,
        score_func_t scoreFunc,
        spatial_pred_t spatialPred,
        output::accumulator_config_t<score_cmp_t> accConfig
    )
        : inputs(inputs)
        , minScore(minScore)
        , scoreFunc(scoreFunc)
        , spatialPred(spatialPred)
        , accConfig(accConfig) {}
};


struct bnb_experiment_t {
	template<typename elem_t, int DIMS, typename key_t, typename data_t, typename key_cmp_t, typename key_agg_t, typename score_func_t, typename score_cmp_t, typename spatial_pred_t>
	void operator()(const top_k_spatial_join_t<elem_t, DIMS, key_t, data_t, key_cmp_t, key_agg_t, score_func_t, score_cmp_t, spatial_pred_t>& join, output::accumulator_data_t<key_t, std::array<data_t, 2>>& accData) const {
		std::array<bnb::seq::bnb_input_t<elem_t, DIMS, key_t, data_t, key_cmp_t>, 2> bnbInputs {
			bnb::seq::bnb_input_t<elem_t, DIMS, key_t, data_t, key_cmp_t>(join.inputs[0].records, join.inputs[0].keyCmp),
			bnb::seq::bnb_input_t<elem_t, DIMS, key_t, data_t, key_cmp_t>(join.inputs[1].records, join.inputs[1].keyCmp)
		};
		bnb::seq::runBnb(bnbInputs, join.spatialPred, join.scoreFunc, join.accConfig, accData);
	}
};

template<typename elem_t, int DIMS, typename key_t, typename data_t, typename key_agg_t, int MAX_HEIGHT>
struct str_input_t {
    rtree_layout_t<MAX_HEIGHT> layout;
    key_agg_t keyAgg;
    record_t<elem_t, DIMS, key_t, data_t>* records;

    str_input_t(rtree_layout_t<MAX_HEIGHT> layout, key_agg_t keyAgg, record_t<elem_t, DIMS, key_t, data_t>* records)
        : layout(layout)
        , keyAgg(keyAgg)
        , records(records) {}
};

struct str_seq_t {
	template<typename elem_t, int DIMS, typename key_t, typename data_t, typename key_agg_t, int MAX_HEIGHT>
    std::array<rtree_t<elem_t, DIMS, key_t, data_t>, 2> operator()(const std::array<str_input_t<elem_t, DIMS, key_t, data_t, key_agg_t, MAX_HEIGHT>, 2>& inputs) const {
        return {
            str::seq::runStr(inputs[0].layout, inputs[0].keyAgg, inputs[0].records),
            str::seq::runStr(inputs[1].layout, inputs[1].keyAgg, inputs[1].records)
        };
    }
};

template<int AFFINITY, typename task_func_t>
struct str_par_t {
    execution::fork_join_pool_t<AFFINITY, task_func_t>& forkJoinPool;

    str_par_t(execution::fork_join_pool_t<AFFINITY, task_func_t>& forkJoinPool)
        : forkJoinPool(forkJoinPool) {}

	template<typename elem_t, int DIMS, typename key_t, typename data_t, typename key_agg_t, int MAX_HEIGHT>
    std::array<rtree_t<elem_t, DIMS, key_t, data_t>, 2> operator()(const std::array<str_input_t<elem_t, DIMS, key_t, data_t, key_agg_t, MAX_HEIGHT>, 2>& inputs) const {
        return {
            str::par::runStr(this->forkJoinPool, inputs[0].layout, inputs[0].keyAgg, inputs[0].records),
            str::par::runStr(this->forkJoinPool, inputs[1].layout, inputs[1].keyAgg, inputs[1].records)
        };
    }
};

struct dfa_seq_t {
    template<typename elem_t, int DIMS, typename key_t, typename data_t, typename spatial_pred_t, typename score_func_t, typename score_cmp_t, int MAX_HEIGHT>
    void operator()(
        const std::array<rtree_t<elem_t, DIMS, key_t, data_t>, 2>& inputs,
        const std::array<rtree_layout_t<MAX_HEIGHT>, 2>& layouts,
        spatial_pred_t spatialPred,
        score_func_t scoreFunc,
        const output::accumulator_config_t<score_cmp_t>& accConfig,
        output::accumulator_data_t<key_t, std::array<data_t, 2>>& accData
    ) const {
        dfa::seq::runDfa(inputs, layouts, spatialPred, scoreFunc, accConfig, accData);
    }
};

template<int MAX_HEIGHT, typename bulk_load_func_t, typename rtree_join_func_t>
struct rtree_join_experiment_t {
    int fanout;
    int paddedFanout;
    bulk_load_func_t bulkLoadFunc;
    rtree_join_func_t rtreeJoinFunc;

    rtree_join_experiment_t(int fanout, int paddedFanout, bulk_load_func_t bulkLoadFunc, rtree_join_func_t rtreeJoinFunc)
        : fanout(fanout)
        , paddedFanout(paddedFanout)
        , bulkLoadFunc(bulkLoadFunc)
        , rtreeJoinFunc(rtreeJoinFunc) {}

	template<typename elem_t, int DIMS, typename key_t, typename data_t, typename key_cmp_t, typename key_agg_t, typename score_func_t, typename score_cmp_t, typename spatial_pred_t>
	void operator()(const top_k_spatial_join_t<elem_t, DIMS, key_t, data_t, key_cmp_t, key_agg_t, score_func_t, score_cmp_t, spatial_pred_t>& join, output::accumulator_data_t<key_t, std::array<data_t, 2>>& accData) const {
		std::array<rtree_layout_t<MAX_HEIGHT>, 2> layouts {
			createRTreeLayout<MAX_HEIGHT>(join.inputs[0].records.last - join.inputs[0].records.first, this->fanout, this->paddedFanout),
			createRTreeLayout<MAX_HEIGHT>(join.inputs[1].records.last - join.inputs[1].records.first, this->fanout, this->paddedFanout)
		};
        std::array<str_input_t<elem_t, DIMS, key_t, data_t, key_agg_t, MAX_HEIGHT>, 2> strInputs{
            str_input_t<elem_t, DIMS, key_t, data_t, key_agg_t, MAX_HEIGHT>(layouts[0], join.inputs[0].keyAgg, join.inputs[0].records.first),
            str_input_t<elem_t, DIMS, key_t, data_t, key_agg_t, MAX_HEIGHT>(layouts[1], join.inputs[1].keyAgg, join.inputs[1].records.first)
        };
        std::array<rtree_t<elem_t, DIMS, key_t, data_t>, 2> rtrees = this->bulkLoadFunc(strInputs);
        this->rtreeJoinFunc(rtrees, layouts, join.spatialPred, join.scoreFunc, join.accConfig, accData);

		for (int i = 0; i < 2; ++i) {
			delete[] rtrees[i].records;
		}
	}
};

struct bba_prepare_seq_t {
    template<int MAX_HEIGHT, typename elem_t, int DIMS, typename key_t, typename data_t, typename key_cmp_t, typename key_agg_t>
    std::array<bba::bba_internal_input_t<elem_t, DIMS, key_t, data_t, key_agg_t, MAX_HEIGHT>, 2> operator()(const std::array<bba::bba_input_t<elem_t, DIMS, key_t, data_t, key_cmp_t, key_agg_t>, 2>& inputs, std::integral_constant<int, MAX_HEIGHT>) const {
        return bba::seq::prepare<elem_t, DIMS, key_t, data_t, key_cmp_t, key_agg_t, MAX_HEIGHT>(inputs);
    }
};

template<int AFFINITY, typename task_func_t>
struct bba_prepare_par_t {
    execution::fork_join_pool_t<AFFINITY, task_func_t>& forkJoinPool;

    bba_prepare_par_t(execution::fork_join_pool_t<AFFINITY, task_func_t>& forkJoinPool)
        : forkJoinPool(forkJoinPool) {}

    template<int MAX_HEIGHT, typename elem_t, int DIMS, typename key_t, typename data_t, typename key_cmp_t, typename key_agg_t>
    std::array<bba::bba_internal_input_t<elem_t, DIMS, key_t, data_t, key_agg_t, MAX_HEIGHT>, 2> operator()(const std::array<bba::bba_input_t<elem_t, DIMS, key_t, data_t, key_cmp_t, key_agg_t>, 2>& inputs, std::integral_constant<int, MAX_HEIGHT>) const {
        return bba::par::prepare<elem_t, DIMS, key_t, data_t, key_cmp_t, key_agg_t, MAX_HEIGHT>(forkJoinPool, inputs);
    }
};

struct bba_join_seq_t {
    template<typename elem_t, int DIMS, typename key_t, typename data_t, typename key_agg_t, typename spatial_pred_t, typename score_func_t, int MAX_HEIGHT, typename score_cmp_t>
    void operator()(
        const std::array<bba::bba_internal_input_t<elem_t, DIMS, key_t, data_t, key_agg_t, MAX_HEIGHT>, 2>& inputs,
        spatial_pred_t spatialPred,
        score_func_t scoreFunc,
        const output::accumulator_config_t<score_cmp_t>& accConfig,
        output::accumulator_data_t<key_t, std::array<data_t, 2>>& accData
	) const {
        bba::seq::runBba(inputs, spatialPred, scoreFunc, accConfig, accData);
    }
};

template<int AFFINITY, typename task_func_t>
struct bba_join_par_t {
    execution::fork_join_pool_t<AFFINITY, task_func_t>& forkJoinPool;

    bba_join_par_t(execution::fork_join_pool_t<AFFINITY, task_func_t>& forkJoinPool)
        : forkJoinPool(forkJoinPool) {}

    template<typename elem_t, int DIMS, typename key_t, typename data_t, typename key_agg_t, typename spatial_pred_t, typename score_func_t, int MAX_HEIGHT, typename score_cmp_t>
    void operator()(
        const std::array<bba::bba_internal_input_t<elem_t, DIMS, key_t, data_t, key_agg_t, MAX_HEIGHT>, 2>& inputs,
        spatial_pred_t spatialPred,
        score_func_t scoreFunc,
        const output::accumulator_config_t<score_cmp_t>& accConfig,
        output::accumulator_data_t<key_t, std::array<data_t, 2>>& accData
	) const {
        bba::par::runBba(forkJoinPool, inputs, spatialPred, scoreFunc, accConfig, accData);
    }
};

template<int MAX_HEIGHT, typename input_factory_t, typename join_func_t>
struct bba_experiment_t {
    size_t blockSize;
    int fanout;
    int paddedFanout;
    input_factory_t inputFactory;
    join_func_t joinFunc;

    bba_experiment_t(size_t blockSize, int fanout, int paddedFanout, input_factory_t inputFactory, join_func_t joinFunc)
        : blockSize(blockSize)
        , fanout(fanout)
        , paddedFanout(paddedFanout)
        , inputFactory(inputFactory)
        , joinFunc(joinFunc) {}

	template<typename elem_t, int DIMS, typename key_t, typename data_t, typename key_cmp_t, typename key_agg_t, typename score_func_t, typename score_cmp_t, typename spatial_pred_t>
    void operator()(const top_k_spatial_join_t<elem_t, DIMS, key_t, data_t, key_cmp_t, key_agg_t, score_func_t, score_cmp_t, spatial_pred_t>& join, output::accumulator_data_t<key_t, std::array<data_t, 2>>& accData) const {
		std::array<bba::bba_input_t<elem_t, DIMS, key_t, data_t, key_cmp_t, key_agg_t>, 2> bbaInputs {
			bba::bba_input_t<elem_t, DIMS, key_t, data_t, key_cmp_t, key_agg_t>(join.inputs[0].records, blockSize, fanout, paddedFanout, join.inputs[0].keyCmp, join.inputs[0].keyAgg),
			bba::bba_input_t<elem_t, DIMS, key_t, data_t, key_cmp_t, key_agg_t>(join.inputs[1].records, blockSize, fanout, paddedFanout, join.inputs[1].keyCmp, join.inputs[1].keyAgg)
		};
		std::array<bba::bba_internal_input_t<elem_t, DIMS, key_t, data_t, key_agg_t, MAX_HEIGHT>, 2> bbaInternalInputs = this->inputFactory(bbaInputs, std::integral_constant<int, MAX_HEIGHT>());
		this->joinFunc(bbaInternalInputs, join.spatialPred, join.scoreFunc, join.accConfig, accData);
    }
};

struct experiment_result_t {
    // std::vector<score_data_pair_t<key_t, std::array<data_t, 2>>> output;
    std::chrono::high_resolution_clock::duration avgDuration;
    std::string name;
};

const size_t biggerThanCacheSize = 10 * 1024 * 1024;
long* cacheClearMemory = new long[biggerThanCacheSize];

void flushCache() {
    for (int i = 0; i < biggerThanCacheSize; ++i) {
        cacheClearMemory[i] = rand();
    }
}

template<typename elem_t, int DIMS, typename key_t, typename data_t, typename key_cmp_t, typename key_agg_t, typename score_func_t, typename score_cmp_t, typename spatial_pred_t, typename task_func_t>
experiment_result_t runExperiment(const top_k_spatial_join_t<elem_t, DIMS, key_t, data_t, key_cmp_t, key_agg_t, score_func_t, score_cmp_t, spatial_pred_t>& join, task_func_t taskFunc, std::string name) {
    const int ITERATIONS = 5;

    std::chrono::high_resolution_clock::duration totalDuration { 0 };
    for (int i = 0; i < ITERATIONS; ++i) {
		flushCache();

		auto begin = std::chrono::high_resolution_clock::now();

		score_data_pair_t<key_t, std::array<data_t, 2>>* resultHeap = new score_data_pair_t<key_t, std::array<data_t, 2>>[join.accConfig.capacity];
		output::accumulator_data_t<int, std::array<data_t, 2>> accData(resultHeap, 0, join.minScore);

		taskFunc(join, accData);

		std::vector<score_data_pair_t<key_t, std::array<data_t, 2>>> output(accData.size);
		accData.read(output.begin(), join.accConfig);
		delete[] resultHeap;

		auto end = std::chrono::high_resolution_clock::now();
		totalDuration += end - begin;
    }

    return experiment_result_t { totalDuration / ITERATIONS, name };
}

template<typename in_t, std::size_t N, typename op_t, std::size_t... is>
std::array<typename std::result_of<op_t(const in_t&)>::type, N> mapArray(const std::array<in_t, N>& in, op_t op, std::index_sequence<is...>) {
    return { { op(in[is])... } };
}

template<typename in_t, std::size_t N, typename op_t>
std::array<typename std::result_of<op_t(const in_t&)>::type, N> mapArray(const std::array<in_t, N>& in, op_t op) {
    return mapArray<in_t, N, op_t>(in, op, std::make_index_sequence<N>());
}

template<typename score_t, typename data_t, typename score_cmp_t>
void assertCorrectness(const std::vector<score_data_pair_t<score_t, data_t>>& expected, const std::vector<score_data_pair_t<score_t, data_t>>& actual, score_cmp_t scoreCmp) {
    assert(expected.size() == actual.size());
    assertTruncatedWeakOrderedArrayEquality<score_t, data_t>(expected.cbegin(), actual.cbegin(), expected.size(), scoreCmp);
}

template<typename score_t, typename data_t, typename score_cmp_t>
bool testCorrectness(const std::vector<score_data_pair_t<score_t, data_t>>& expected, const std::vector<score_data_pair_t<score_t, data_t>>& actual, score_cmp_t scoreCmp) {
    if (expected.size() != actual.size()) return false;
    return testTruncatedWeakOrderedArrayEquality<score_t, data_t>(expected.cbegin(), actual.cbegin(), expected.size(), scoreCmp);
}

template<typename score_t, typename data_t>
void printOutput(score_data_pair_t<score_t, std::array<data_t, 2>>* output, size_t outputSize) {
	for (int i = 0; i < outputSize; ++i) {
		printf("#%02d s: %07d, l: %07d, r: %07d\n", i + 1, output[i].score, output[i].data[0], output[i].data[1]);
	}
}

int main() {
    // If l > r, l has a better score than r
    const auto keyCmp = std::greater<int>();
    const auto keyAgg = [&keyCmp](const int& lhs, const int& rhs) {
        return keyCmp(lhs, rhs) ? lhs : rhs;
    };
    const auto scoreFunc = std::plus<int>();
    const auto scoreCmp = std::greater<int>();
    const auto spatialPred = euclidean_distance_test_t<float>(1.0f);

    const int DIMS = 2;
    using key_cmp_t = decltype(keyCmp);
    using key_agg_t = decltype(keyAgg);
    using score_func_t = decltype(scoreFunc);
    using score_cmp_t = decltype(scoreCmp);
    using spatial_pred_t = decltype(spatialPred);
    using elem_t = float;
    using key_t = int;
    using data_t = int;
    using task_func_t = std::function<void(int)>;

	// const int INPUT_SIZES[] = { 1024 * 1024, 32 * 1024, 1024 };
    const int INPUT_SIZES[] = { 1024 * 1024 };
	// const int OUTPUT_SIZES[] = { 10000, 1000, 100 };
	const int OUTPUT_SIZES[] = { 10000 };
	// const int FANOUTS[] = { 2, 3, 4, 5, 6, 7, 8, 10, 12, 14, 16, 20, 24, 28, 32 };
	const int FANOUTS[] = { 4 };
	// const int BLOCK_SIZES[] = { 512, 1024, 2048, 4096, 8192, 16384, 32768, 65536 };
	const int BLOCK_SIZES[] = { 65536 };

	const int AFFINITY = 4; // how many threads in total will be used for parallel implementations
	execution::fork_join_pool_t<AFFINITY, task_func_t> forkJoinPool;

	const float ELEM_DISTRIBUTION_SIZE = 1000.0f;
	const int KEY_DISTRIBUTION_SIZE = 100;
	std::mt19937 random(0);

	std::unique_ptr<record_t<elem_t, DIMS, key_t, data_t>[]> inputArrays[2];
	for (int i = 0; i < 2; ++i) {
		inputArrays[i] = std::make_unique<record_t<elem_t, DIMS, key_t, data_t>[]>(INPUT_SIZES[0]);
		std::uniform_real_distribution<elem_t> elemDistribution(0, ELEM_DISTRIBUTION_SIZE);
		std::uniform_int_distribution<key_t> keyDistribution(0, KEY_DISTRIBUTION_SIZE);
		generateRandomLeafRecords<DIMS>(inputArrays[i].get(), inputArrays[i].get() + INPUT_SIZES[0], random, elemDistribution, keyDistribution);
	}

	for (int inputSizeIdxL = 0; inputSizeIdxL < sizeof(INPUT_SIZES) / sizeof(int); ++inputSizeIdxL) {
		const int inputSizeL = INPUT_SIZES[inputSizeIdxL];
		range_t<record_t<elem_t, DIMS, key_t, data_t>*> inputRangeL = { inputArrays[0].get(), inputArrays[0].get() + inputSizeL };
        ranked_join_input_t<elem_t, DIMS, key_t, data_t, key_cmp_t, key_agg_t> inputL(inputRangeL, keyCmp, keyAgg);
		for (int inputSizeIdxR = inputSizeIdxL; inputSizeIdxR < sizeof(INPUT_SIZES) / sizeof(int); ++inputSizeIdxR) {
			const int inputSizeR = INPUT_SIZES[inputSizeIdxR];
			range_t<record_t<elem_t, DIMS, key_t, data_t>*> inputRangeR = { inputArrays[1].get(), inputArrays[1].get() + inputSizeR };
			ranked_join_input_t<elem_t, DIMS, key_t, data_t, key_cmp_t, key_agg_t> inputR(inputRangeR, keyCmp, keyAgg);
			for (int outputSize : OUTPUT_SIZES) {
				printf("Input %d x %d -> Output %d:\n", inputSizeL, inputSizeR, outputSize);

				const output::accumulator_config_t<score_cmp_t> accConfig(outputSize, scoreCmp);

				std::vector<experiment_result_t> results;

				top_k_spatial_join_t<elem_t, DIMS, key_t, data_t, key_cmp_t, key_agg_t, score_func_t, score_cmp_t, spatial_pred_t> joinParameters(
					{ inputL, inputR },
					std::numeric_limits<int>::min(),
					scoreFunc,
					spatialPred,
					accConfig
				);

				for (int fanout : FANOUTS) {
					// results.push_back(runExperiment(
					//     joinParameters, rtree_join_experiment_t<10, str_seq_t, dfa_seq_t>(fanout, fanout, str_seq_t(), dfa_seq_t()),
					//     "dfa seq fanout=" + std::to_string(fanout)
					// ));
					// results.push_back(runExperiment(
					//     joinParameters, rtree_join_experiment_t<10, str_par_t<AFFINITY, task_func_t>, dfa_seq_t>(fanout, fanout, str_par_t(forkJoinPool), dfa_seq_t()),
					//     "dfa par fanout=" + std::to_string(fanout)
					// ));

					for (int blockSize : BLOCK_SIZES) {
						/* if (blockSize <= inputSizeR) {
							results.push_back(runExperiment(
								joinParameters, bba_experiment_t<10, bba_prepare_seq_t, bba_join_seq_t>(blockSize, fanout, fanout, bba_prepare_seq_t(), bba_join_seq_t()),
								"bba seq fanout=" + std::to_string(fanout) + " blockSize=" + std::to_string(blockSize)
							));
							results.push_back(runExperiment(
								joinParameters, bba_experiment_t<10, bba_prepare_par_t<AFFINITY, task_func_t>, bba_join_par_t<AFFINITY, task_func_t>>(blockSize, fanout, fanout, bba_prepare_par_t(forkJoinPool), bba_join_par_t(forkJoinPool)),
								"bba par fanout=" + std::to_string(fanout) + " blockSize=" + std::to_string(blockSize)
							));
						} */

#define experiment(AFFINITY) { execution::fork_join_pool_t<AFFINITY, task_func_t> forkJoinPool; results.push_back(runExperiment(joinParameters, bba_experiment_t<10, bba_prepare_par_t<AFFINITY, task_func_t>, bba_join_par_t<AFFINITY, task_func_t>>(blockSize, fanout, fanout, bba_prepare_par_t(forkJoinPool), bba_join_par_t(forkJoinPool)), "bba par fanout=" + std::to_string(fanout) + " blockSize=" + std::to_string(blockSize) + " affinity=" + std::to_string(AFFINITY))); }
                        experiment(1);
                        experiment(2);
                        experiment(3);
                        experiment(4);
                        experiment(5);
                        experiment(6);
                        experiment(7);
                        experiment(8);
                        experiment(9);
                        experiment(10);
                        experiment(11);
                        experiment(12);
                        experiment(13);
                        experiment(14);
                        experiment(15);
                        experiment(16);
					}
				}

				/* std::sort(results.begin(), results.end(), [](const experiment_result_t& lhs, const experiment_result_t& rhs) {
					return lhs.avgDuration < rhs.avgDuration;
				}); */

				for (experiment_result_t& result : results) {
					printf("%lld ns: %s\n", std::chrono::duration_cast<std::chrono::nanoseconds>(result.avgDuration).count(), result.name.c_str());
				}
				printf("\n");
			}
		}
	}

/*#ifndef NDEBUG
    assertCorrectness(bnbSeqResult.output, dfaSeqResult.output, scoreCmp);
    assertCorrectness(bnbSeqResult.output, dfaParResult.output, scoreCmp);
    assertCorrectness(bnbSeqResult.output, bbaSeqResult.output, scoreCmp);
    assertCorrectness(bnbSeqResult.output, bbaParResult.output, scoreCmp);
#else
    printf("dfaSeq correctness: %d\n", testCorrectness(bnbSeqResult.output, dfaSeqResult.output, scoreCmp));
    printf("dfaPar correctness: %d\n", testCorrectness(bnbSeqResult.output, dfaParResult.output, scoreCmp));
    printf("bbaSeq correctness: %d\n", testCorrectness(bnbSeqResult.output, bbaSeqResult.output, scoreCmp));
    printf("bbaPar correctness: %d\n", testCorrectness(bnbSeqResult.output, bbaParResult.output, scoreCmp));
#endif*/

    return 0;
}
