#ifndef SHARED_H
#define SHARED_H

#include "utility.h"

template<typename score_t, typename data_t>
struct score_data_pair_t {
    score_t score;
    data_t data;
};

template<typename iter_t>
struct range_t {
    iter_t first;
    iter_t last;
};

template<typename elem_t, int DIMS>
struct bounding_box_t {
	elem_t min[DIMS];
	elem_t max[DIMS];
};

template<typename elem_t, int DIMS>
bool bbIntersect(const bounding_box_t<elem_t, DIMS>& lhs, const bounding_box_t<elem_t, DIMS>& rhs) {
	for (int dim = 0; dim < DIMS; ++dim) {
		if (lhs.min[dim] > rhs.max[dim] || rhs.min[dim] > lhs.max[dim]) {
			return false;
		}
	}
	return true;
}

template<typename elem_t, int DIMS>
elem_t euclideanDistanceSqr(const bounding_box_t<elem_t, DIMS>& lhs, const bounding_box_t<elem_t, DIMS>& rhs) {
    elem_t totalDistanceSqr = 0;
    for (int dim = 0; dim < DIMS; ++dim) {
        // if either distance is positive, we know they do not intersect on this axis and thus have a distance to contribute
        // first check if lhs is greater than rhs
        elem_t axisDistance = lhs.min[dim] - rhs.max[dim];
        if (axisDistance > 0) {
            totalDistanceSqr += axisDistance * axisDistance;
            continue;
        }
        // then check if rhs is greater than lhs
        axisDistance = rhs.min[dim] - lhs.max[dim];
        if (axisDistance > 0) {
            totalDistanceSqr += axisDistance * axisDistance;
        }
    }
    return totalDistanceSqr;
}

struct intersects_t {
    template<typename elem_t, int DIMS>
    bool operator()(const bounding_box_t<elem_t, DIMS>& lhs, const bounding_box_t<elem_t, DIMS>& rhs) const {
        return bbIntersect(lhs, rhs);
    }
};

template<typename elem_t>
struct euclidean_distance_test_t {
    elem_t thresholdSqr;

    euclidean_distance_test_t(elem_t threshold)
        : thresholdSqr(threshold * threshold) {}

    template<int DIMS>
    bool operator()(const bounding_box_t<elem_t, DIMS>& lhs, const bounding_box_t<elem_t, DIMS>& rhs) const {
        return euclideanDistanceSqr(lhs, rhs) <= this->thresholdSqr;
    }
};

template<typename elem_t, int DIMS, typename key_t, typename data_t>
struct alignas(32) record_t {
    typedef elem_t elem_t;
    typedef key_t key_t;
    typedef data_t data_t;
    static const int DIMS = DIMS;

	bounding_box_t<elem_t, DIMS> mbb;
	key_t key;
    union {
		data_t data;
        int childBlock;
    };
};

struct segment_t {
    size_t offset;
    size_t size;
    size_t paddedSize;
};

template<int MAX_HEIGHT>
struct rtree_layout_t {
    segment_t segments[MAX_HEIGHT];
    size_t height;
    size_t size;
    int fanout;
};

template<typename elem_t, int DIMS, typename key_t, typename data_t>
struct rtree_t {
    record_t<elem_t, DIMS, key_t, data_t> root;
    record_t<elem_t, DIMS, key_t, data_t>* records;
};

template<int MAX_HEIGHT>
bool calculateRTreeLayout(size_t inputSize, int fanout, int blockSize, rtree_layout_t<MAX_HEIGHT>& layout) {
    assert(fanout <= blockSize /* fanout cannot be greater than blockSize */);
    layout.fanout = fanout;
    if (inputSize == 0) {
        layout.height = 0;
        layout.size = 0;
        return false;
    }

    layout.height = 1;
    layout.segments[0].size = inputSize;
	size_t numParents = ceilDiv(inputSize, (size_t)fanout);
	size_t paddedSize = numParents * blockSize;
    size_t totalSize = paddedSize;
    layout.segments[0].paddedSize = paddedSize;
    while (numParents > 1) {
        if (layout.height >= MAX_HEIGHT) {
            return true;
        }

        layout.segments[layout.height].size = numParents;
		numParents = ceilDiv(numParents, (size_t)fanout);
		paddedSize = numParents * blockSize;
        totalSize += paddedSize;
        layout.segments[layout.height].paddedSize = paddedSize;
        ++layout.height;
    }


    size_t offset = totalSize;
    for (size_t level = 0; level < layout.height; ++level) {
        offset -= layout.segments[level].paddedSize;
        layout.segments[level].offset = offset;
    }

    layout.size = totalSize;

    return false;
}

template<int MAX_HEIGHT>
rtree_layout_t<MAX_HEIGHT> createRTreeLayout(size_t inputSize, int fanout, int blockSize) {
	rtree_layout_t<MAX_HEIGHT> layout;
	bool layoutError = calculateRTreeLayout(inputSize, fanout, blockSize, layout);
	assert(!layoutError);
	return layout;
}

#endif // SHARED_H
