#ifndef STR_PAR_H
#define STR_PAR_H

#include <algorithm>
#include "str.h"
#include "bp_par.h"
#include "shared.h"
#include "execution.h"

namespace str::par {
	struct segmented_block_partitioning_t {
		size_t segmentSize;
		size_t blockSize;
	};

	// Calculates a tile partitioning strategy for the sorting pass of a level of Sort-Tile-Recursive
	// In each level of Sort-Tile-Recursive, the nodes at the level are partitioned into spatially separate tiles of N nodes each
	// The partitioning into tiles is achieved by a series of segmented block partitioning operations, one for each dimension
	void calculateTilePartitionStrategy(segmented_block_partitioning_t* partitionings, int dims, size_t inputSize, size_t blockSize) {
		size_t blocksPerSegment = 1;
		size_t segmentSize = blockSize;
		while (segmentSize < inputSize) {
			++blocksPerSegment;
			segmentSize = blockSize;
			for (int dim = 0; dim < dims; ++dim) {
				segmentSize *= blocksPerSegment;
			}
		}

		for (int dim = 0; dim < dims; ++dim) {
			partitionings[dim].segmentSize = segmentSize;
			segmentSize /= blocksPerSegment;
			partitionings[dim].blockSize = segmentSize;
		}
	}

	// Performs a tile block partitioning operation
	// The input will be partitioned blocks which form spatially separate "tiles"
	template<typename elem_t, int DIMS, typename key_t, typename data_t, int AFFINITY, typename task_func_t>
	void tileBlockPartition(
		execution::fork_join_pool_t<AFFINITY, task_func_t>& forkJoinPool,
		record_t<elem_t, DIMS, key_t, data_t>* first,
		record_t<elem_t, DIMS, key_t, data_t>* last,
		segmented_block_partitioning_t* partitionings
	) {
		// The input is, at all times, divided into some amount of segments
		// For each dimension, a segmented block partitioning operation is performed to divide each segment into more segments by the coordinate on that dimension
		for (int dim = 0; dim < DIMS; ++dim) {
			segmented_block_partitioning_t partitioning = partitionings[dim];
			str::record_dim_cmp_t cmp(dim);
			bp::par::segmentedBlockPartition(forkJoinPool, first, last, partitioning.segmentSize, partitioning.blockSize, cmp);
		}
	}

	template<typename elem_t, int DIMS, typename key_t, typename data_t, int AFFINITY, typename task_func_t, typename key_agg_t>
	void createInnerNodes(
		execution::fork_join_pool_t<AFFINITY, task_func_t>& forkJoinPool,
		size_t blockSize,
		key_agg_t keyAgg,
		const record_t<elem_t, DIMS, key_t, data_t>* first,
		const record_t<elem_t, DIMS, key_t, data_t>* last,
		record_t<elem_t, DIMS, key_t, data_t>* output
	) {
		int blocksPerThread = ceilDiv((long)(last - first), (long)(blockSize * AFFINITY));
		forkJoinPool.run([&](int tid) {
			int firstBlock = tid * blocksPerThread;
			int childBlock = firstBlock;
			auto begin = first + (firstBlock * blockSize);
			auto end = std::min(begin + blocksPerThread * blockSize, last);
			auto i = begin;
			auto o = output + firstBlock;
			for (; i + blockSize < end; i += blockSize, ++o, ++childBlock) {
				*o = str::createInnerNode(keyAgg, i, i + blockSize, childBlock);
			}
			if (i < end) {
				*o = str::createInnerNode(keyAgg, i, end, childBlock);
			}
		});
	}

	template<typename elem_t, int DIMS, typename key_t, typename data_t, int MAX_HEIGHT, int AFFINITY, typename task_func_t, typename key_agg_t>
	record_t<elem_t, DIMS, key_t, data_t> runStr(
		execution::fork_join_pool_t<AFFINITY, task_func_t>& forkJoinPool,
		const rtree_layout_t<MAX_HEIGHT>& layout,
		key_agg_t keyAgg,
		const record_t<elem_t, DIMS, key_t, data_t>* input,
		record_t<elem_t, DIMS, key_t, data_t>* output
	) {
		// copy input to bottom output segment
		segment_t bottomSegment = layout.segments[0];
		std::copy_n(input, bottomSegment.size, output + bottomSegment.offset);

		segmented_block_partitioning_t partitionings[DIMS];

		for (int level = 0; level < layout.height - 1; ++level) {
			segment_t segment = layout.segments[level];
			segment_t nextSegment = layout.segments[level + 1];

			calculateTilePartitionStrategy(partitionings, DIMS, segment.size, layout.fanout);
			record_t<elem_t, DIMS, key_t, data_t>* segmentFirst = output + segment.offset;
			record_t<elem_t, DIMS, key_t, data_t>* segmentLast = segmentFirst + segment.size;
			record_t<elem_t, DIMS, key_t, data_t>* nextSegmentFirst = output + nextSegment.offset;

			tileBlockPartition(forkJoinPool, segmentFirst, segmentLast, partitionings);
			createInnerNodes(forkJoinPool, layout.fanout, keyAgg, segmentFirst, segmentLast, nextSegmentFirst);
		}
		segment_t lastSegment = layout.segments[layout.height - 1];
		return str::createInnerNode(keyAgg, output + lastSegment.offset, output + lastSegment.offset + lastSegment.size, 0);
	}

	template<typename elem_t, int DIMS, typename key_t, typename data_t, int MAX_HEIGHT, typename key_agg_t>
	struct str_input_t {
		rtree_layout_t<MAX_HEIGHT> layout;
		key_agg_t keyAgg;
		const record_t<elem_t, DIMS, key_t, data_t>* input;
		record_t<elem_t, DIMS, key_t, data_t>* output;
	};

	/* template<typename elem_t, int DIMS, typename key_t, typename data_t, int MAX_HEIGHT, int AFFINITY, typename task_func_t, typename key_agg_t>
	std::array<record_t<elem_t, DIMS, key_t, data_t>, 2> runMultiStr(
		execution::fork_join_pool_t<AFFINITY, task_func_t>& forkJoinPool,
		const std::array<str_input_t<elem_t, DIMS, key_t, data_t, MAX_HEIGHT, key_agg_t>, 2>& inputs
	) {
		std::array<std::array<segmented_block_partitioning_t, DIMS>, 2> partitioningStrategies;
		for (int inputIdx = 0; inputIdx < 2; ++inputIdx) {
			const str_input_t<elem_t, DIMS, key_t, data_t, MAX_HEIGHT, key_agg_t>& input = inputs[inputIdx];
			// copy input records to bottom output segment
			segment_t bottomSegment = input.layout.segments[0];
			std::copy_n(input.input, bottomSegment.size, input.output + bottomSegment.offset);
		}

		std::vector<bp::par::segmented_block_partition_input_t<record_t<elem_t, DIMS, key_t, data_t>*, record_dim_cmp_t>, 2> segmentedBlockPartitionInputs;

		for (int level = 0; level < layout.height - 1; ++level) {
			segment_t segment = layout.segments[level];
			segment_t nextSegment = layout.segments[level + 1];

			calculateTilePartitionStrategy(partitionings, DIMS, segment.size, layout.fanout);
			record_t<elem_t, DIMS, key_t, data_t>* segmentFirst = output + segment.offset;
			record_t<elem_t, DIMS, key_t, data_t>* segmentLast = segmentFirst + segment.size;
			record_t<elem_t, DIMS, key_t, data_t>* nextSegmentFirst = output + nextSegment.offset;

			tileBlockPartition(forkJoinPool, segmentFirst, segmentLast, partitionings);
			createInnerNodes(forkJoinPool, layout.fanout, keyAgg, segmentFirst, segmentLast, nextSegmentFirst);
		}
		segment_t lastSegment = layout.segments[layout.height - 1];
		return str::createInnerNode(keyAgg, output + lastSegment.offset, output + lastSegment.offset + lastSegment.size, 0);
	} */

	template<typename elem_t, int DIMS, typename key_t, typename data_t, int MAX_HEIGHT, int AFFINITY, typename task_func_t, typename key_agg_t>
	rtree_t<elem_t, DIMS, key_t, data_t> runStr(
		execution::fork_join_pool_t<AFFINITY, task_func_t>& forkJoinPool,
		const rtree_layout_t<MAX_HEIGHT>& layout,
		key_agg_t keyAgg,
		const record_t<elem_t, DIMS, key_t, data_t>* input
	) {
		rtree_t<elem_t, DIMS, key_t, data_t> result;
		result.records = new record_t<elem_t, DIMS, key_t, data_t>[layout.size];
		result.root = runStr(forkJoinPool, layout, keyAgg, input, result.records);
		return result;
	}
}

#endif //STR_PAR_H
