#ifndef STR_SEQ_H
#define STR_SEQ_H

#include <algorithm>
#include <execution>
#include "str.h"
#include "shared.h"
#include "bp_seq.h"

namespace str::seq {
	template<typename elem_t, int DIMS, typename key_t, typename data_t>
	void tileBlockPartition(size_t blockSize, size_t size, record_t<elem_t, DIMS, key_t, data_t>* first, record_t<elem_t, DIMS, key_t, data_t>* last) {
		size_t divisions = 1;
		size_t segmentSize = blockSize;
		while (segmentSize < size) {
			++divisions;
			segmentSize = blockSize;
			for (int dim = 0; dim < DIMS; ++dim) {
				segmentSize *= divisions;
			}
		}

		for (int dim = 0; dim < DIMS; ++dim, segmentSize /= divisions) {
			str::record_dim_cmp_t cmp(dim);
			bp::seq::segmentedBlockPartition(first, last, segmentSize, segmentSize / divisions, cmp);
		}
	}

	template<typename elem_t, int DIMS, typename key_t, typename data_t, typename key_agg_t>
	void createInnerNodes(int blockSize, key_agg_t keyAgg, const record_t<elem_t, DIMS, key_t, data_t>* first, const record_t<elem_t, DIMS, key_t, data_t>* last, record_t<elem_t, DIMS, key_t, data_t>* output) {
		auto i = first;
		auto o = output;
		int childBlock = 0;
		// create all full inner nodes first
		for (; i + blockSize < last; i += blockSize, ++o, ++childBlock) {
			*o = str::createInnerNode(keyAgg, i, i + blockSize, childBlock);
		}
		// create the last inner node (may also be full)
		if (i < last) {
			*o = str::createInnerNode(keyAgg, i, last, childBlock);
		}
	}

	template<typename elem_t, int DIMS, typename key_t, typename data_t, int MAX_HEIGHT, typename key_agg_t>
	record_t<elem_t, DIMS, key_t, data_t> runStr(
		const rtree_layout_t<MAX_HEIGHT>& layout,
		key_agg_t keyAgg,
		const record_t<elem_t, DIMS, key_t, data_t>* input,
		record_t<elem_t, DIMS, key_t, data_t>* output
	) {
		// copy input to bottom output segment
		segment_t bottomSegment = layout.segments[0];
		std::copy_n(input, bottomSegment.size, output + bottomSegment.offset);

		for (int level = 0; level < layout.height - 1; ++level) {
			segment_t segment = layout.segments[level];
			segment_t nextSegment = layout.segments[level + 1];
			record_t<elem_t, DIMS, key_t, data_t>* segmentFirst = output + segment.offset;
			record_t<elem_t, DIMS, key_t, data_t>* segmentLast = segmentFirst + segment.size;
			record_t<elem_t, DIMS, key_t, data_t>* nextSegmentFirst = output + nextSegment.offset;

			tileBlockPartition(layout.fanout, segment.size, segmentFirst, segmentLast);
			createInnerNodes(layout.fanout, keyAgg, segmentFirst, segmentLast, nextSegmentFirst);
		}
		segment_t lastSegment = layout.segments[layout.height - 1];
		return str::createInnerNode(keyAgg, output + lastSegment.offset, output + lastSegment.offset + lastSegment.size, 0);
	}

	template<typename elem_t, int DIMS, typename key_t, typename data_t, int MAX_HEIGHT, typename key_agg_t>
	rtree_t<elem_t, DIMS, key_t, data_t> runStr(
		const rtree_layout_t<MAX_HEIGHT>& layout,
		key_agg_t keyAgg,
		const record_t<elem_t, DIMS, key_t, data_t>* input
	) {
		rtree_t<elem_t, DIMS, key_t, data_t> result;
		result.records = new record_t<elem_t, DIMS, key_t, data_t>[layout.size];
		result.root = runStr(layout, keyAgg, input, result.records);
		return result;
	}
}

#endif //STR_SEQ_H
