#ifndef VALIDATION_H
#define VALIDATION_H

#include "output.h"

template<typename score_t, typename data_t, typename iter_t>
int countEqual(const score_data_pair_t<score_t, data_t>& value, iter_t first, iter_t last) {
    int result = 0;
	for (auto i = first; i < last; ++i) {
		if (value.score == i->score && value.data == i->data) {
            ++result;
		}
	}
    return result;
}

template<typename score_t, typename data_t, typename iter_t>
void assertUnorderedArrayEquality(iter_t expected, iter_t actual, size_t size) {
    for (iter_t reference = expected; reference < expected + size; ++reference) {
        const score_data_pair_t<score_t, data_t> value = *reference;
        int expectedMatches = countEqual<score_t, data_t, iter_t>(value, expected, expected + size);
        int actualMatches = countEqual<score_t, data_t, iter_t>(value, expected, expected + size);
        assert(expectedMatches == actualMatches /* each unique array element must appear (the same amount of times) in both arrays */);
    }
}

template<typename score_t, typename data_t, typename iter_t, typename score_cmp_t>
void assertTruncatedWeakOrderedArrayEquality(iter_t expected, iter_t actual, size_t size, score_cmp_t scoreCmp) {
    // we will split the expected results into ranges of equal scores
    // the same range applied to the actual results must create two unordered-equal arrays
    for (size_t rangeStart = 0; rangeStart < size;) {
		size_t rangeEnd = rangeStart + 1;
        for (; rangeEnd < size; ++rangeEnd) {
            // assert(!scoreCmp(expected[rangeEnd], expected[rangeEnd - 1]) /* actual result must be in order */);
            if (scoreCmp(expected[rangeStart].score, expected[rangeEnd].score)) {
                break;
            }
        }

        // a range ending at the boundary of the array may actually be part of a larger truncated range which we cannot truly validate
        if (rangeEnd != size) {
            assertUnorderedArrayEquality<score_t, data_t, iter_t>(expected + rangeStart, actual + rangeStart, rangeEnd - rangeStart);
        }
        rangeStart = rangeEnd;
    }
}
template<typename score_t, typename data_t, typename iter_t>
bool testUnorderedArrayEquality(iter_t expected, iter_t actual, size_t size) {
    for (iter_t reference = expected; reference < expected + size; ++reference) {
        const score_data_pair_t<score_t, data_t> value = *reference;
        int expectedMatches = countEqual<score_t, data_t, iter_t>(value, expected, expected + size);
        int actualMatches = countEqual<score_t, data_t, iter_t>(value, expected, expected + size);
        // each unique array element must appear (the same amount of times) in both arrays
        if (expectedMatches != actualMatches) return false;
    }
    return true;
}

template<typename score_t, typename data_t, typename iter_t, typename score_cmp_t>
bool testTruncatedWeakOrderedArrayEquality(iter_t expected, iter_t actual, size_t size, score_cmp_t scoreCmp) {
    // we will split the expected results into ranges of equal scores
    // the same range applied to the actual results must create two unordered-equal arrays
    for (size_t rangeStart = 0; rangeStart < size;) {
		size_t rangeEnd = rangeStart + 1;
        for (; rangeEnd < size; ++rangeEnd) {
            // assert(!scoreCmp(expected[rangeEnd], expected[rangeEnd - 1]) /* actual result must be in order */);
            if (scoreCmp(expected[rangeStart].score, expected[rangeEnd].score)) {
                break;
            }
        }

        // a range ending at the boundary of the array may actually be part of a larger truncated range which we cannot truly validate
        if (rangeEnd != size) {
            if (!testUnorderedArrayEquality<score_t, data_t, iter_t>(expected + rangeStart, actual + rangeStart, rangeEnd - rangeStart)) {
                return false;
            }
        }
        rangeStart = rangeEnd;
    }
    return true;
}


#endif // VALIDATION_H
