#ifndef BBA_CU_H
#define BBA_CU_H

#include <cuda_runtime.h>
#include <cooperative_groups.h>
#include "shared.h"
#include "heap.cuh"
#include "str_small.cuh"

// #define _DEBUG_IDS
// #define _DEBUG_QUEUES
// #define _DEBUG_OUTPUT
// #define _DEBUG_RTREES
#define _BBA_VECTORIZE_RECORD_ACCESS

namespace bba {
    template<int LOG_FANOUT, typename elem_t, int DIMS, typename key_t, typename data_t, typename key_cmp_t>
    __device__ void debugRTree(const rtree_t<elem_t, DIMS, key_t, data_t>& rtree, key_cmp_t keyCmp, int logBottomSize) {
        // for each child of a parent, keyCmp(child.key, parent.key) must be false
        int logNumChildren = logBottomSize;
        int logNumParents = logBottomSize - LOG_FANOUT;
        int firstChildIdx = 0;
        int firstParentIdx = 1 << logBottomSize;
        while (logNumChildren > LOG_FANOUT) {
            for (int i = 0; i < (1 << logNumParents); ++i) {
                record_t<elem_t, DIMS, key_t, data_t> parentRecord = rtree.records[firstParentIdx + i];

                for (int j = 0; j < (1 << LOG_FANOUT); ++j) {
                    record_t<elem_t, DIMS, key_t, data_t> childRecord = rtree.records[parentRecord.childBlock + j];
                    if (keyCmp(childRecord.key, parentRecord.key)) {
                        printf("Found a child with a better key than its parent!\n");
                    }

                    for (int dim = 0; dim < DIMS; ++dim) {
                        if (childRecord.mbb.min[dim] < parentRecord.mbb.min[dim]) {
							printf("Found a child with bounds outside parent!\n");
                        }
                        if (childRecord.mbb.max[dim] > parentRecord.mbb.max[dim]) {
							printf("Found a child with bounds outside parent!\n");
                        }
                    }
                }
            }

            for (int idx0 = 0; idx0 < (1 << logNumChildren); ++idx0) {
                record_t<elem_t, DIMS, key_t, data_t> record0 = rtree.records[firstChildIdx + idx0];
                data_t data0 = record0.data;
                for (int idx1 = idx0 + 1; idx1 < (1 << logNumChildren); ++idx1) {
					record_t<elem_t, DIMS, key_t, data_t> record1 = rtree.records[firstChildIdx + idx1];
					data_t data1 = record1.data;
                    if (data0 == data1) {
                        printf("Found duplicate data %d!\n", data0);
                    }
                }
            }

            logNumChildren = logNumParents;
            logNumParents -= LOG_FANOUT;
            firstChildIdx = firstParentIdx;
            firstParentIdx += (1 << logNumChildren);
        }

		for (int j = 0; j < (1 << logNumChildren); ++j) {
			record_t<elem_t, DIMS, key_t, data_t> childRecord = rtree.records[rtree.root.childBlock + j];
			if (keyCmp(childRecord.key, rtree.root.key)) {
				printf("Found a child with a better key than the root!\n");
			}

			for (int dim = 0; dim < DIMS; ++dim) {
				if (childRecord.mbb.min[dim] < rtree.root.mbb.min[dim]) {
					printf("Found a child with bounds outside the root!\n");
				}
				if (childRecord.mbb.max[dim] > rtree.root.mbb.max[dim]) {
					printf("Found a child with bounds outside the root!\n");
				}
			}
		}
    }


    struct bba_result_t {
        enum class status_t {
            OK, // the BBA was successful
            OUT_OF_MEMORY, // the BBA ran out of memory for r-trees
            QUEUE_OVERFLOW, // the joining of R-trees resulted in a queue overflow
            CUDA_ERROR // there was a CUDA error while performing dynamic parallelism
        };

        status_t status;
        int outputSize;
    };

    template<typename elem_t, int DIMS, typename key_t, typename data_t>
    __global__ void retrieveKeysKernel(record_t<elem_t, DIMS, key_t, data_t>* records, key_t* keys, size_t size) {
        cooperative_groups::thread_block group = cooperative_groups::this_thread_block();
        for (size_t idx = group.thread_rank() + blockIdx.x * group.size(); idx < size; idx += group.size() * gridDim.x) {
            keys[idx] = records[idx].key;
        }
    }

    template<typename elem_t, int DIMS, typename key_t, typename data_t>
	struct bba_input_t {
        const record_t<elem_t, DIMS, key_t, data_t>* source;
        rtree_t<elem_t, DIMS, key_t, data_t>* rtrees;
        int blockCount;
	};

    template<typename elem_t, int DIMS, typename key_t, typename data_t>
    struct bba_input_reader_t {
        const record_t<elem_t, DIMS, key_t, data_t>* source;
        rtree_t<elem_t, DIMS, key_t, data_t>* rtrees;
        int blockCount;
        int nextBlock;

        __device__ bba_input_reader_t(const bba_input_t<elem_t, DIMS, key_t, data_t>& input)
            : source(input.source)
            , rtrees(input.rtrees)
            , blockCount(input.blockCount)
            , nextBlock(0) {}

        __device__ const key_t& getRecordScore(int idx) const {
            return this->source[idx].key;
        }
    };

    template<typename elem_t, int DIMS, typename key_t, typename data_t, typename score_func_t, typename score_cmp_t>
    __device__ bool allocateNextBlock(const bba_input_reader_t<elem_t, DIMS, key_t, data_t> readers[2], int blockSize, score_func_t scoreFunc, score_cmp_t scoreCmp, key_t scoreThreshold, int allocations[2]) {
        // find which input whose next block should be allocated
        // the goal is maximize the upper bound for the combined score of any produced results
        int bestInputIdx = -1;
        key_t bestScoreUB = scoreThreshold;

		#pragma unroll
        for (int i = 0; i < 2; ++i) {
            int allocation = allocations[i];
            // can this input be read any further?
            if (allocation < readers[i].blockCount) {
                // if we allocate a block from the input and join it with allocated blocks of the other input, this is an upper bound for the score any produced results
                key_t scoreUB = scoreFunc(readers[i].getRecordScore(allocation * blockSize), readers[1 - i].getRecordScore(0));
                if (scoreCmp(scoreUB, bestScoreUB)) {
                    bestInputIdx = i;
                    bestScoreUB = scoreUB;
                }
            }
        }
        if (bestInputIdx != -1) {
            ++allocations[bestInputIdx];
            return true;
        }
        return false;
    }

	template<typename elem_t, int DIMS, typename key_t, typename data_t, typename score_func_t, typename score_cmp_t>
    __device__ int allocateNextBlocks(bba_input_reader_t<elem_t, DIMS, key_t, data_t> readers[2], int blockSize, score_func_t scoreFunc, score_cmp_t scoreCmp, key_t scoreThreshold, int maxTotalAllocations, int allocations[2]) {
        for (int i = 0; i < 2; ++i) {
            allocations[i] = readers[i].nextBlock;
        }

        // greedily allocate blocks until no more can be allocated
        int totalAllocations = 0;
        for (; totalAllocations < maxTotalAllocations; ++totalAllocations) {
            if (!allocateNextBlock(readers, blockSize, scoreFunc, scoreCmp, scoreThreshold, allocations)) {
                break;
            }
        }
        return totalAllocations;
    }

	struct record_dim_cmp_t {
		unsigned int dim;

		__device__ record_dim_cmp_t(unsigned int dim) : dim(dim) {}

		template<typename elem_t, int DIMS, typename key_t, typename data_t>
		__device__ bool operator()(const record_t<elem_t, DIMS, key_t, data_t>& lhs, const record_t<elem_t, DIMS, key_t, data_t>& rhs) const {
			return lhs.mbb.min[dim] < rhs.mbb.min[dim];
		}
	};

	template<typename elem_t, int DIMS, typename key_t, typename data_t, typename group_t>
	__device__ void tilePartitionInPlace(unsigned int laneId, record_t<elem_t, DIMS, key_t, data_t>* records, int logNumRecords, int logFanout, group_t group) {
        /* if (group.thread_rank() == 0) {
            printf("Before:\n");
            for (int i = 0; i < (1 << logNumRecords); ++i) {
                record_t<elem_t, DIMS, key_t, data_t>& record = records[i];
                printf("#%03d %03d %03d", i, record.key, record.data);
				for (int dim = 0; dim < DIMS; ++dim) {
					printf(" [%f, %f]", record.mbb.min[dim], record.mbb.max[dim]);
				}
                printf("\n");
            }
        }
        group.sync(); */

        // we want this amount of spatially separated groups
        int logNumGroups = logNumRecords - logFanout;
        // to achieve this, we divide each segment in each sorting pass by this dividend
        int logDividend = (logNumGroups + DIMS - 1) / DIMS;

        int logSegmentWidth = logNumRecords;
		#pragma unroll
		for (int dim = 0; dim < DIMS; ++dim) {
			int logBlockWidth = logSegmentWidth - logDividend;

			record_dim_cmp_t recordCmp(dim);

			str_small::arraySegmentedSort(records, recordCmp, logSegmentWidth, 1 << logNumRecords, group);
            group.sync();

			logSegmentWidth = logBlockWidth;
		}

        /* if (group.thread_rank() == 0) {
            printf("\nAfter:\n");
            for (int i = 0; i < (1 << logNumRecords); ++i) {
                record_t<elem_t, DIMS, key_t, data_t>& record = records[i];
                printf("#%03d %03d %03d", i, record.key, record.data);
				for (int dim = 0; dim < DIMS; ++dim) {
					printf(" [%f, %f]", record.mbb.min[dim], record.mbb.max[dim]);
				}
                printf("\n");
            }
        }
        group.sync(); */
	}

	template<typename elem_t, int DIMS, typename key_t, typename data_t, typename key_agg_t, typename group_t>
    __device__ record_t<elem_t, DIMS, key_t, data_t> shuffleCreateParent(const record_t<elem_t, DIMS, key_t, data_t>& childRecord, key_agg_t keyAgg, int count, group_t group) {
        record_t<elem_t, DIMS, key_t, data_t> parentRecord = childRecord;
		for (int f = 1; f < count; f <<= 1) {
			// bool shouldReduce = !(((f << 1) - 1) & group.thread_rank());
            bool shouldReduce = group.thread_rank() + f < count;
			#pragma unroll
			for (int dim = 0; dim < DIMS; ++dim) {
				elem_t altMin = group.shfl_down(parentRecord.mbb.min[dim], f);
				if (shouldReduce) {
					parentRecord.mbb.min[dim] = min(parentRecord.mbb.min[dim], altMin);
				}
			}

			#pragma unroll
			for (int dim = 0; dim < DIMS; ++dim) {
				elem_t altMax = group.shfl_down(parentRecord.mbb.max[dim], f);
				if (shouldReduce) {
					parentRecord.mbb.max[dim] = max(parentRecord.mbb.max[dim], altMax);
				}
			}

			key_t altKey = group.shfl_down(parentRecord.key, f);
			if (shouldReduce) {
				parentRecord.key = keyAgg(parentRecord.key, altKey);
			}
		}
        return parentRecord;
    }

    template<int LOG_FANOUT, typename elem_t, int DIMS, typename key_t, typename data_t, typename key_agg_t, typename parent_group_t, typename child_group_t>
    __device__ void shuffleCreateParents(record_t<elem_t, DIMS, key_t, data_t>* workspace, int numChildren, int firstChildIdx, key_agg_t keyAgg, parent_group_t& parentGroup, child_group_t& childGroup) {
        const int FANOUT = 1 << LOG_FANOUT;

        // in case we have more records than groups
		// we need the whole thread group to be synchronized here
        for (int groupIdx = 0; groupIdx < numChildren; groupIdx += parentGroup.size()) {
            int childRecordIdx = groupIdx + parentGroup.thread_rank();
            int parentRecordIdx = childRecordIdx >> LOG_FANOUT;
			record_t<elem_t, DIMS, key_t, data_t> childRecord;
            if (childRecordIdx < numChildren) {
                childRecord = workspace[childRecordIdx];
            }

            // we need all records to be read from the workspace before we overwrite it
            parentGroup.sync();

            // TODO: This does not handle uneven amounts of children
            if (childRecordIdx < numChildren) {
				record_t<elem_t, DIMS, key_t, data_t> parentRecord = shuffleCreateParent(childRecord, keyAgg, FANOUT, childGroup);
				if (childGroup.thread_rank() == 0) {
					// only the first thread in the child group has a valid parent record
					// write the parent record to the workspace
                    parentRecord.childBlock = firstChildIdx + childRecordIdx;
					workspace[parentRecordIdx] = parentRecord;
				}
            }
        }
    }

#ifdef _DEBUG_RTREES
	struct ge_t {
		template<typename item_t>
		__host__ __device__ bool operator()(const item_t& lhs, const item_t& rhs) const {
			return lhs > rhs;
		}
	};
#endif

    template<int LOG_FANOUT, typename elem_t, int DIMS, typename key_t, typename data_t, typename key_agg_t, typename parent_group_t, typename child_group_t>
    __device__ void sortTileRecursive(
        rtree_t<elem_t, DIMS, key_t, data_t>& output,
        record_t<elem_t, DIMS, key_t, data_t>* workspace,
        int logInputSize,
        key_agg_t keyAgg,
        parent_group_t parentGroup,
        child_group_t childGroup
    ) {
        const int FANOUT = 1 << LOG_FANOUT;

        record_t<elem_t, DIMS, key_t, data_t>* outputRecords = output.records;
        int firstChildIdx = 0;
        int logNumChildren = logInputSize;
        // while the parent is not the root!
        while (logNumChildren > LOG_FANOUT) {
            // tile partition the current records in the workspace
            tilePartitionInPlace(parentGroup.thread_rank(), workspace, logNumChildren, LOG_FANOUT, parentGroup);
            parentGroup.sync();

#ifndef _BBA_VECTORIZE_RECORD_ACCESS
			// write the sorted children out
            for (int idx = parentGroup.thread_rank(); idx < (1 << logNumChildren); idx += parentGroup.size()) {
				outputRecords[firstChildIdx + idx] = workspace[idx];
            }
#else
			static_assert(sizeof(record_t<elem_t, DIMS, key_t, data_t>) % sizeof(int4) == 0, "record size must be divisble by int4 size");
			const int INT4S_PER_RECORD = sizeof(record_t<elem_t, DIMS, key_t, data_t>) / sizeof(int4);
			for (int i = parentGroup.thread_rank(); i < (INT4S_PER_RECORD << logNumChildren); i += parentGroup.size()) {
				reinterpret_cast<int4*>(outputRecords + firstChildIdx)[i] = reinterpret_cast<const int4*>(workspace)[i];
			}
#endif

            // create parents in-place in the workspace
			shuffleCreateParents<LOG_FANOUT>(workspace, 1 << logNumChildren, firstChildIdx, keyAgg, parentGroup, childGroup);
            parentGroup.sync();

            firstChildIdx += (1 << logNumChildren);
            logNumChildren -= LOG_FANOUT;
        }

        // we have a number of children in the workspace equal to or below the fanout
        // they do not require any sorting, just write them straight out
#ifndef _BBA_VECTORIZE_RECORD_ACCESS
		if (parentGroup.thread_rank() < (1 << logNumChildren)) {
			outputRecords[firstChildIdx + parentGroup.thread_rank()] = workspace[parentGroup.thread_rank()];
		}
#else
		static_assert(sizeof(record_t<elem_t, DIMS, key_t, data_t>) % sizeof(int4) == 0, "record size must be divisble by int4 size");
		const int INT4S_PER_RECORD = sizeof(record_t<elem_t, DIMS, key_t, data_t>) / sizeof(int4);
		if (parentGroup.thread_rank() < (INT4S_PER_RECORD << logNumChildren)) {
			reinterpret_cast<int4*>(outputRecords + firstChildIdx)[parentGroup.thread_rank()] = reinterpret_cast<const int4*>(workspace)[parentGroup.thread_rank()];
		}
#endif

        // now create the root
        if (parentGroup.thread_rank() < FANOUT) {
            record_t<elem_t, DIMS, key_t, data_t> childRecord = workspace[childGroup.thread_rank()];
            {
                record_t<elem_t, DIMS, key_t, data_t> rootRecord = shuffleCreateParent(childRecord, keyAgg, 1 << logNumChildren, childGroup);
                // only the first thread in the child group has a valid parent record
                if (childGroup.thread_rank() == 0) {
                    rootRecord.childBlock = firstChildIdx;
                    output.root = rootRecord;
                }
            }
        }
    }

    template<int LOG_FANOUT, typename elem_t, int DIMS, typename key_t, typename data_t, typename key_agg_t>
    __global__ void blockInitializeKernel(
        bba_input_reader_t<elem_t, DIMS, key_t, data_t> readerL,
        bba_input_reader_t<elem_t, DIMS, key_t, data_t> readerR,
        int logBlockSize,
        int allocationL,
        key_agg_t keyAgg
    ) {
        const int FANOUT = 1 << LOG_FANOUT;

        __shared__ extern record_t<elem_t, DIMS, key_t, data_t> workspace[];

        const record_t<elem_t, DIMS, key_t, data_t>* inputRecords;
        rtree_t<elem_t, DIMS, key_t, data_t>* rtree;
        if (readerL.nextBlock + blockIdx.x < allocationL) {
            int block = readerL.nextBlock + blockIdx.x;
            inputRecords = readerL.source + (block << logBlockSize);
            rtree = readerL.rtrees + block;
        }
        else {
            int block = readerR.nextBlock + readerL.nextBlock + blockIdx.x - allocationL;
            inputRecords = readerR.source + (block << logBlockSize);
            rtree = readerR.rtrees + block;
        }

        cooperative_groups::thread_block block = cooperative_groups::this_thread_block();
		cooperative_groups::thread_block_tile<FANOUT> tile = cooperative_groups::tiled_partition<FANOUT>(block);

        // read the children into the workspace
        for (int idx = block.thread_rank(); idx < (1 << logBlockSize); idx += block.size()) {
			workspace[idx] = inputRecords[idx];
        }
        block.sync();

        sortTileRecursive<LOG_FANOUT>(*rtree, workspace, logBlockSize, keyAgg, block, tile);

#ifdef _DEBUG_RTREES
        block.sync();

        if (threadIdx.x == 0) {
            debugRTree<LOG_FANOUT>(*rtree, ge_t(), logBlockSize);
        }
#endif
    }

    struct dfa_entry_t {
        int level;
        int leftBlock;
        int rightBlock;
#ifdef _DEBUG_IDS
        int _debugId;
#endif

        __host__ __device__ bool operator==(const dfa_entry_t& other) const {
            return this->level == other.level && this->leftBlock == other.leftBlock && this->rightBlock == other.rightBlock
#ifdef _DEBUG_IDS
                && this->_debugId == other._debugId
#endif
                ;
        }
    };

    template<typename elem_t>
    struct default_shuffler_t {
        __device__ elem_t shfl(unsigned int mask, const elem_t& src, int srcLane) const {
            return __shfl_sync(mask, src, srcLane);
        }

        __device__ elem_t shfl_up(unsigned int mask, const elem_t& src, unsigned int delta) const {
            return __shfl_up_sync(mask, src, delta);
        }

        __device__ elem_t shfl_down(unsigned int mask, const elem_t& src, unsigned int delta) const {
            return __shfl_down_sync(mask, src, delta);
        }

        __device__ elem_t shfl_xor(unsigned int mask, const elem_t& src, int laneMask) const {
            return __shfl_xor_sync(mask, src, laneMask);
        }

        template<typename group_t>
        __device__ elem_t group_shfl(group_t group, const elem_t& src, int srcLane) const {
            return group.shfl(src, srcLane);
        }

        template<typename group_t>
        __device__ elem_t group_shfl_up(group_t group, const elem_t& src, unsigned int delta) const {
            return group.shfl_up(src, delta);
        }

        template<typename group_t>
        __device__ elem_t group_shfl_down(group_t group, const elem_t& src, unsigned int delta) const {
            return group.shfl_down(src, delta);
        }

        template<typename group_t>
        __device__ elem_t group_shfl_xor(group_t group, const elem_t& src, int laneMask) const {
            return group.shfl_xor(src, laneMask);
        }
    };

    struct dfa_entry_shuffler_t {
        __device__ dfa_entry_t shfl(unsigned int mask, const dfa_entry_t& src, int srcLane) const {
            return {
				__shfl_sync(mask, src.level, srcLane),
				__shfl_sync(mask, src.leftBlock, srcLane),
				__shfl_sync(mask, src.rightBlock, srcLane)
#ifdef _DEBUG_IDS
                , __shfl_sync(mask, src._debugId, srcLane)
#endif
            };
        }

        __device__ dfa_entry_t shfl_up(unsigned int mask, const dfa_entry_t& src, unsigned int delta) const {
            return {
				__shfl_up_sync(mask, src.level, delta),
				__shfl_up_sync(mask, src.leftBlock, delta),
				__shfl_up_sync(mask, src.rightBlock, delta)
#ifdef _DEBUG_IDS
				, __shfl_up_sync(mask, src._debugId, delta)
#endif
            };
        }

        __device__ dfa_entry_t shfl_down(unsigned int mask, const dfa_entry_t& src, unsigned int delta) const {
            return {
                __shfl_down_sync(mask, src.level, delta),
                __shfl_down_sync(mask, src.leftBlock, delta),
                __shfl_down_sync(mask, src.rightBlock, delta)
#ifdef _DEBUG_IDS
                , __shfl_down_sync(mask, src._debugId, delta)
#endif
            };
        }

        __device__ dfa_entry_t shfl_xor(unsigned int mask, const dfa_entry_t& src, int laneMask) const {
            return {
				__shfl_xor_sync(mask, src.level, laneMask),
				__shfl_xor_sync(mask, src.leftBlock, laneMask),
				__shfl_xor_sync(mask, src.rightBlock, laneMask)
#ifdef _DEBUG_IDS
                , __shfl_xor_sync(mask, src._debugId, laneMask)
#endif
            };
        }

        template<typename group_t>
        __device__ dfa_entry_t group_shfl(group_t group, const dfa_entry_t& src, int srcLane) const {
            return {
				group.shfl(src.level, srcLane),
				group.shfl(src.leftBlock, srcLane),
				group.shfl(src.rightBlock, srcLane)
#ifdef _DEBUG_IDS
				, group.shfl(src._debugId, srcLane)
#endif
            };
        }

        template<typename group_t>
        __device__ dfa_entry_t group_shfl_up(group_t group, const dfa_entry_t& src, unsigned int delta) const {
            return {
				group.shfl_up(src.level, delta),
				group.shfl_up(src.leftBlock, delta),
				group.shfl_up(src.rightBlock, delta)
#ifdef _DEBUG_IDS
				, group.shfl_up(src._debugId, delta)
#endif
            };
        }

        template<typename group_t>
        __device__ dfa_entry_t group_shfl_down(group_t group, const dfa_entry_t& src, unsigned int delta) const {
            return {
                group.shfl_down(src.level, delta),
                group.shfl_down(src.leftBlock, delta),
                group.shfl_down(src.rightBlock, delta)
#ifdef _DEBUG_IDS
                , group.shfl_down(src._debugId, delta)
#endif
            };
        }

        template<typename group_t>
        __device__ dfa_entry_t group_shfl_xor(group_t group, const dfa_entry_t& src, int laneMask) const {
            return {
				group.shfl_xor(src.level, laneMask),
				group.shfl_xor(src.leftBlock, laneMask),
				group.shfl_xor(src.rightBlock, laneMask)
#ifdef _DEBUG_IDS
                , group.shfl_xor(src._debugId, laneMask)
#endif
            };
        }
    };

    template<typename item_t, typename item_shuffler_t = default_shuffler_t<item_t>>
    struct tuple_shuffler_t {
        item_shuffler_t itemShuffler;

        __host__ __device__ tuple_shuffler_t(item_shuffler_t itemShuffler = item_shuffler_t()) : itemShuffler(itemShuffler) {}

        __device__ tuple_t<item_t> shfl(unsigned int mask, const tuple_t<item_t>& src, int srcLane) const {
            return {
                itemShuffler.shfl(mask, src.left, srcLane),
                itemShuffler.shfl(mask, src.right, srcLane)
            };
        }

        __device__ tuple_t<item_t> shfl_up(unsigned int mask, const tuple_t<item_t>& src, unsigned int delta) const {
            return {
                itemShuffler.shfl_up(mask, src.left, delta),
                itemShuffler.shfl_up(mask, src.right, delta)
            };
        }

        __device__ tuple_t<item_t> shfl_down(unsigned int mask, const tuple_t<item_t>& src, unsigned int delta) const {
            return {
                itemShuffler.shfl_down(mask, src.left, delta),
                itemShuffler.shfl_down(mask, src.right, delta)
            };
        }

        __device__ tuple_t<item_t> shfl_xor(unsigned int mask, const tuple_t<item_t>& src, int laneMask) const {
            return {
                itemShuffler.shfl_xor(mask, src.left, laneMask),
                itemShuffler.shfl_xor(mask, src.right, laneMask)
            };
        }

        template<typename group_t>
        __device__ tuple_t<item_t> group_shfl(group_t group, const tuple_t<item_t>& src, int srcLane) const {
            return {
                itemShuffler.group_shfl(group, src.left, srcLane),
                itemShuffler.group_shfl(group, src.right, srcLane)
            };
        }

        template<typename group_t>
        __device__ tuple_t<item_t> group_shfl_up(group_t group, const tuple_t<item_t>& src, unsigned int delta) const {
            return {
                itemShuffler.group_shfl_up(group, src.left, delta),
                itemShuffler.group_shfl_up(group, src.right, delta)
            };
        }

        template<typename group_t>
        __device__ tuple_t<item_t> group_shfl_down(group_t group, const tuple_t<item_t>& src, unsigned int delta) const {
            return {
                itemShuffler.group_shfl_down(group, src.left, delta),
                itemShuffler.group_shfl_down(group, src.right, delta)
            };
        }

        template<typename group_t>
        __device__ tuple_t<item_t> group_shfl_xor(group_t group, const tuple_t<item_t>& src, int laneMask) const {
            return {
                itemShuffler.group_shfl_xor(group, src.left, laneMask),
                itemShuffler.group_shfl_xor(group, src.right, laneMask)
            };
        }
    };

	struct dfa_entry_ref_t {
		int& level;
		int& leftBlock;
		int& rightBlock;
#ifdef _DEBUG_IDS
        int& _debugId;
#endif

		__device__ dfa_entry_ref_t(int& level, int& leftBlock, int& rightBlock
#ifdef _DEBUG_IDS
            , int& _debugId
#endif
        )
			: level(level)
			, leftBlock(leftBlock)
			, rightBlock(rightBlock)
#ifdef _DEBUG_IDS
            , _debugId(_debugId)
#endif
        {}

		__device__ dfa_entry_ref_t& operator=(const dfa_entry_t& other) {
			this->level = other.level;
			this->leftBlock = other.leftBlock;
			this->rightBlock = other.rightBlock;
#ifdef _DEBUG_IDS
            this->_debugId = other._debugId;
#endif
			return *this;
		}

        __device__ operator dfa_entry_t() {
            return {
                this->level,
                this->leftBlock,
                this->rightBlock
#ifdef _DEBUG_IDS
                , this->_debugId
#endif
            };
        }
	};

	template<int SIZE>
	struct dfa_entry_soa_t {
		int level[SIZE];
		int leftBlock[SIZE];
		int rightBlock[SIZE];
#ifdef _DEBUG_IDS
        int _debugId[SIZE];
#endif

		__device__ dfa_entry_ref_t operator[](int index) {
			return dfa_entry_ref_t(level[index], leftBlock[index], rightBlock[index]
#ifdef _DEBUG_IDS
                , _debugId[index]
#endif
            );
		}
	};

    template<typename item_t>
	struct tuple_ref_t {
		item_t& left;
		item_t& right;

		__device__ tuple_ref_t(item_t& left, item_t& right)
			: left(left)
			, right(right) {}

		__device__ tuple_ref_t<item_t>& operator=(const tuple_t<item_t>& other) {
			this->left = other.left;
			this->right = other.right;
			return *this;
		}

        __device__ operator tuple_t<item_t>() {
            return {
                left,
                right
            };
        }
	};

	template<typename item_t, int SIZE>
	struct tuple_soa_t {
		item_t left[SIZE];
		item_t right[SIZE];

		__device__ tuple_ref_t<item_t> operator[](int index) {
			return tuple_ref_t<item_t>(left[index], right[index]);
		}
	};

    template<typename cmp_t>
    struct rev_cmp_t {
        cmp_t cmp;

        __host__ __device__ rev_cmp_t(cmp_t cmp)
            : cmp(cmp) {}

        template<typename item_t>
        __host__ __device__ bool operator()(const item_t& lhs, const item_t& rhs) const {
            return cmp(rhs, lhs);
        }
    };

	template<typename elem_t, int DIMS, typename key_t, typename data_t>
	__device__ void printRecord(const record_t<elem_t, DIMS, key_t, data_t>& record) {
		printf("key = %d, data = %d\n", record.key, record.data);
		for (int dim = 0; dim < DIMS; ++dim) {
			printf("DIM %d [%f, %f]\n", dim, record.mbb.min[dim], record.mbb.max[dim]);
		}
	}

    // returns false if the queue overflows, true otherwise
	template<int LOG_FANOUT, int QUEUE_HEAP_LIMIT, typename elem_t, int DIMS, typename key_t, typename data_t, typename spatial_pred_t, typename score_func_t, typename score_cmp_t, typename key_shuffler_t, typename data_shuffler_t, typename group_t>
    __device__ bool processDfaEntry(
#ifdef _DEBUG_IDS
        int& _debugIdCounter,
#endif
        dfa_entry_t entry,
        int numChildren,
        int outputCapacity,
        const record_t<elem_t, DIMS, key_t, data_t>* g_recordsL,
        const record_t<elem_t, DIMS, key_t, data_t>* g_recordsR,
        spatial_pred_t spatialPred,
        score_func_t scoreFunc,
        score_cmp_t scoreCmp,
        key_shuffler_t keyShuffler,
        data_shuffler_t dataShuffler,
        key_t& scoreThreshold,
        heap::heap_t<key_t, tuple_t<data_t>, rev_cmp_t<score_cmp_t>, key_shuffler_t, tuple_shuffler_t<data_t, data_shuffler_t>>& outputHeap,
        heap::heap_t<key_t, dfa_entry_t, score_cmp_t, key_shuffler_t, dfa_entry_shuffler_t>& queueHeap,
        group_t group
    ) {
        const int FANOUT = 1 << LOG_FANOUT;

        __shared__ record_t<elem_t, DIMS, key_t, data_t> s_recordsL[FANOUT];
        __shared__ record_t<elem_t, DIMS, key_t, data_t> s_recordsR[FANOUT];
        __shared__ key_t s_pendingKeys[32];
        // __shared__ union {
		// 	dfa_entry_t s_pendingQueueValues[32];
		// 	tuple_t<data_t> s_pendingOutputValues[32];
        // };
		__shared__ dfa_entry_t s_pendingQueueValues[32];
		__shared__ tuple_t<data_t> s_pendingOutputValues[32];

        // read the records from global memory first
#ifndef _BBA_VECTORIZE_RECORD_ACCESS
		for (int idx = group.thread_rank(); idx < numChildren; idx += group.size()) {
			s_recordsL[idx] = g_recordsL[entry.leftBlock + idx];
			s_recordsR[idx] = g_recordsR[entry.rightBlock + idx];
		}
#else
        static_assert(sizeof(record_t<elem_t, DIMS, key_t, data_t>) % sizeof(int4) == 0, "record size must be divisble by int4 size");
        const int INT4S_PER_RECORD = sizeof(record_t<elem_t, DIMS, key_t, data_t>) / sizeof(int4);
        for (int i = group.thread_rank(); i < INT4S_PER_RECORD * numChildren; i += group.size()) {
            reinterpret_cast<int4*>(s_recordsL)[i] = reinterpret_cast<const int4*>(g_recordsL + entry.leftBlock)[i];
            reinterpret_cast<int4*>(s_recordsR)[i] = reinterpret_cast<const int4*>(g_recordsR + entry.rightBlock)[i];
        }
#endif
		group.sync();

		for (int groupIdx = 0; groupIdx < numChildren * numChildren; groupIdx += group.size()) {
            int myIdx = groupIdx + group.thread_rank();
			record_t<elem_t, DIMS, key_t, data_t> recordL = s_recordsL[myIdx % numChildren];
			record_t<elem_t, DIMS, key_t, data_t> recordR = s_recordsR[myIdx / numChildren];

			key_t childScore;
			int pred = 0;
			if (myIdx < numChildren * numChildren && spatialPred(recordL.mbb, recordR.mbb)) {
				childScore = scoreFunc(recordL.key, recordR.key);
				if (scoreCmp(childScore, scoreThreshold)) {
					pred = 1;
				}
			}

			int outputPosition;
			str_small::shufflePrefixScan(pred, outputPosition, sum_t(), 0, group);
            group.sync();

			if (pred) {
				s_pendingKeys[outputPosition] = childScore;
				if (entry.level != 1) {
					s_pendingQueueValues[outputPosition] = dfa_entry_t {
						entry.level - 1,
						recordL.childBlock,
						recordR.childBlock
#ifdef _DEBUG_IDS
						, _debugIdCounter + outputPosition
#endif
					};
				}
				else {
					s_pendingOutputValues[outputPosition] = tuple_t<data_t> {
						recordL.data,
						recordR.data
					};
				}
			}

#ifdef _DEBUG_IDS
			_debugIdCounter += __shfl_sync(0xffffffff, outputPosition + pred, 31);
#endif

            int pendingCount = group.shfl(outputPosition + pred, group.size() - 1);
            group.sync();

            /* if (group.thread_rank() == 0) {
                printf("Adding %d items\n", pendingCount);
            } */

			for (int j = 0; j < pendingCount; ++j) {
				if (entry.level != 1) {
#ifdef _DEBUG_QUEUES
                    queueHeap.debugAssertValueNotExists(s_pendingQueueValues[j]);
#endif
					if (queueHeap.getNumNodes() == QUEUE_HEAP_LIMIT && queueHeap.getLastNodeSize() == 32) {
                        return false;
					}
					queueHeap.insert(s_pendingKeys[j], s_pendingQueueValues[j]);
					/* if (group.thread_rank() == 0) {
						printf("Push s = %06d, level = %d, leftBlock = %06d, rightBlock = %06d\n", s_pendingKeys[j], s_pendingQueueValues[j].level, s_pendingQueueValues[j].leftBlock, s_pendingQueueValues[j].rightBlock);
					}
                    queueHeap.debugPrint(); */
				}
				else {
#ifdef _DEBUG_OUTPUT
                    outputHeap.debugAssertValueNotExists(s_pendingOutputValues[j]);
#endif
					if (outputHeap.size() != outputCapacity) {
						// there is still space for output items
						outputHeap.insert(s_pendingKeys[j], s_pendingOutputValues[j]);
					}
					else {
                        // the output heap is at capacity, we can only insert an item if it is better than the worst item in the heap
						if (scoreCmp(s_pendingKeys[j], outputHeap.peekKey())) {
							outputHeap.popInsert(s_pendingKeys[j], s_pendingOutputValues[j]);
						}
					}

                    if (outputHeap.size() == outputCapacity) {
						scoreThreshold = outputHeap.peekKey();
                    }
				}
			}
		}
#ifdef _DEBUG_QUEUES
		// queueHeap.debugDuplicateValues();
#endif
        return true;
    }

    // returns false if the queue ever overflows, true otherwise
	template<int LOG_FANOUT, int QUEUE_LIMIT, typename elem_t, int DIMS, typename key_t, typename data_t, typename spatial_pred_t, typename score_func_t, typename score_cmp_t, typename key_shuffler_t, typename data_shuffler_t, typename group_t>
    __device__ bool joinRTrees(
        const rtree_t<elem_t, DIMS, key_t, data_t>& rTreeL,
        const rtree_t<elem_t, DIMS, key_t, data_t>& rTreeR,
        int logBlockSize,
        int outputCapacity,
        spatial_pred_t spatialPred,
        score_func_t scoreFunc,
        score_cmp_t scoreCmp,
        key_shuffler_t keyShuffler,
        data_shuffler_t dataShuffler,
        key_t& scoreThreshold,
        heap::heap_t<key_t, tuple_t<data_t>, rev_cmp_t<score_cmp_t>, key_shuffler_t, tuple_shuffler_t<data_t, data_shuffler_t>>& outputHeap,
        group_t group
    ) {
        const int FANOUT = 1 << LOG_FANOUT;

        using queue_heap_t = heap::heap_t<key_t, dfa_entry_t, score_cmp_t, key_shuffler_t, dfa_entry_shuffler_t>;

        __shared__ key_t s_queueHeapMinKeys[QUEUE_LIMIT];
        __shared__ typename queue_heap_t::shared_state_t queueHeapSharedState;
        typename queue_heap_t::config_t queueHeapConfig(scoreCmp, keyShuffler, dfa_entry_shuffler_t());
        heap::node_t<key_t, dfa_entry_t> queueHeapNodes[QUEUE_LIMIT];
        queue_heap_t queueHeap(queueHeapConfig, queueHeapSharedState, s_queueHeapMinKeys, queueHeapNodes);

#ifdef _DEBUG_IDS
        int _debugIdCounter = blockIdx.x * 10000;
#endif

        if (spatialPred(rTreeL.root.mbb, rTreeR.root.mbb)) {
			key_t rootScore = scoreFunc(rTreeL.root.key, rTreeR.root.key);
            if (scoreCmp(rootScore, scoreThreshold)) {
                // by having the block size and fanout as powers of two, we have some very useful properties
                // the size of the leaf level (level = 0) is 2^logBlockSize.
                // the size of any other level L is the size of the below level L - 1 divided by the fanout
                // the level below the root level must have a size equal to or below 2^LOG_FANOUT
                // we can determine the level of the root by the amount of divisions that must be performed
                // we can determine the size of the level below the root by seeing which number remains after the divisions
                // in logarithmic space, this is fairly easy
                int rootLevel = 1 + (logBlockSize - 1) / LOG_FANOUT;
                int logRootFanout = 1 + (logBlockSize - 1) % LOG_FANOUT;
                dfa_entry_t rootEntry {
                    rootLevel,
                    rTreeL.root.childBlock,
                    rTreeR.root.childBlock
#ifdef _DEBUG_IDS
                    , ++_debugIdCounter
#endif
                };
                
                // process the root entry
                // we process this entry separately because it may have a different amount of children than the rest of the entries
                if (!processDfaEntry<LOG_FANOUT, QUEUE_LIMIT>(
#ifdef _DEBUG_IDS
                    _debugIdCounter,
#endif
                    rootEntry,
                    1 << logRootFanout,
                    outputCapacity,
                    rTreeL.records, rTreeR.records,
                    spatialPred, scoreFunc, scoreCmp,
                    keyShuffler, dataShuffler,
                    scoreThreshold, outputHeap, queueHeap,
                    group
				)) {
                    return false;
                }
            }
        }

        while (!queueHeap.isEmpty()) {
            key_t parentScore = queueHeap.peekKey();
            if (!scoreCmp(parentScore, scoreThreshold)) {
                // the score of the pair of parents is no better than the score threshold
                // all pairs of children produced by this pair of parents must also have a score no better than the score threshold
                // transitively, all pairs of leaf records directly/indirectly produced by this pair of parents must have a score no better than the score threshold
                // therefore, we have no need to expand this pair of parents
                // additionally, since all pairs of parents in the queue have a score no better than this one, the above statement holds true for ALL pairs of parents in the queue
                // we can therefore terminate. we have found all pairs of leaf records satisfying the score threshold
                break;
            }
            dfa_entry_t entry = queueHeap.peekValue();
            queueHeap.pop();
            /* if (group.thread_rank() == 0) {
                printf("Pop  s = %06d, level = %d, leftBlock = %06d, rightBlock = %06d\n", parentScore, entry.level, entry.leftBlock, entry.rightBlock);
            }
			queueHeap.debugPrint(); */

            if (!processDfaEntry<LOG_FANOUT, QUEUE_LIMIT>(
#ifdef _DEBUG_IDS
                _debugIdCounter,
#endif
                entry,
                FANOUT,
                outputCapacity,
                rTreeL.records, rTreeR.records,
                spatialPred, scoreFunc, scoreCmp,
                keyShuffler, dataShuffler,
                scoreThreshold, outputHeap, queueHeap,
                group
			)) {
                return false;
            }
        }

        return true;
    }

    // returns false if the queue overflows, true otherwise
	template<bool SWAP, int LOG_FANOUT, int QUEUE_LIMIT, typename elem_t, int DIMS, typename key_t, typename data_t, typename spatial_pred_t, typename score_func_t, typename score_cmp_t, typename key_shuffler_t, typename data_shuffler_t, typename group_t>
    __device__ bool joinBlocks(
        const rtree_t<elem_t, DIMS, key_t, data_t>& theRtree,
        const rtree_t<elem_t, DIMS, key_t, data_t>* otherRtrees,
        int numOtherInitialized,
        int numOtherAllocated,
        int logBlockSize,
        int outputCapacity,
        spatial_pred_t spatialPred,
        score_func_t scoreFunc,
        score_cmp_t scoreCmp,
        key_shuffler_t keyShuffler,
        data_shuffler_t dataShuffler,
        key_t& scoreThreshold,
        heap::heap_t<key_t, tuple_t<data_t>, rev_cmp_t<score_cmp_t>, key_shuffler_t, tuple_shuffler_t<data_t, data_shuffler_t>>& outputHeap,
        int* uninitializedJoinFlags,
        int uninitializedJoinFlagStride,
        group_t group
    ) {
        // first do uncontested joins with blocks that are already initialized
        for (int otherBlock = 0; otherBlock < numOtherInitialized; ++otherBlock) {
            rtree_t<elem_t, DIMS, key_t, data_t> rTreeL;
            rtree_t<elem_t, DIMS, key_t, data_t> rTreeR;
            if (!SWAP) {
                rTreeL = theRtree;
                rTreeR = otherRtrees[otherBlock];
            }
            else {
                rTreeL = otherRtrees[otherBlock];
                rTreeR = theRtree;
            }

            // can joining these R-trees produce any significant scores?
            if (scoreCmp(scoreFunc(rTreeL.root.key, rTreeR.root.key), scoreThreshold)) {
                // yes, there may be pairs of objects that belong in the result
                if (!joinRTrees<LOG_FANOUT, QUEUE_LIMIT>(rTreeL, rTreeR, logBlockSize, outputCapacity, spatialPred, scoreFunc, scoreCmp, keyShuffler, dataShuffler, scoreThreshold, outputHeap, group)) {
                    return false;
                }
            }
            else {
                // no, all pairs of objects do not belong in the result
                // all pairs of objects from subsequent joins with the following blocks also do not belong in the result
                return true;
            }
        }

        // then attempt contested joins with blocks that were just allocated
        for (int otherBlock = numOtherInitialized; otherBlock < numOtherAllocated; ++otherBlock) {
            bool shouldJoin;
            if (group.thread_rank() == 0) {
				int* uninitializedJoinFlag = uninitializedJoinFlags + (otherBlock - numOtherInitialized) * uninitializedJoinFlagStride;
                shouldJoin = !atomicCAS(uninitializedJoinFlag, 0, 1);
            }
            shouldJoin = group.shfl(shouldJoin, 0);

            if (shouldJoin) {
				rtree_t<elem_t, DIMS, key_t, data_t> rTreeL;
				rtree_t<elem_t, DIMS, key_t, data_t> rTreeR;
				if (!SWAP) {
					rTreeL = theRtree;
					rTreeR = otherRtrees[otherBlock];
				}
				else {
					rTreeL = otherRtrees[otherBlock];
					rTreeR = theRtree;
				}

				// can joining these R-trees produce any significant scores?
				if (scoreCmp(scoreFunc(rTreeL.root.key, rTreeR.root.key), scoreThreshold)) {
					// yes, there may be pairs of objects that belong in the result
                    if (!joinRTrees<LOG_FANOUT, QUEUE_LIMIT>(rTreeL, rTreeR, logBlockSize, outputCapacity, spatialPred, scoreFunc, scoreCmp, keyShuffler, dataShuffler, scoreThreshold, outputHeap, group)) {
                        return false;
                    }
				}
				else {
					// no, all pairs of objects do not belong in the result
					// all pairs of objects from subsequent joins with the following blocks also do not belong in the result
					return true;
				}
            }
        }

        return true;
    }

    // outputs the items in the heap into the buffer in the opposite order of the heap
    template<typename key_t, typename data_t, typename key_cmp_t, typename key_shuffler_t, typename data_shuffler_t>
    __device__ void deconstructHeap(heap::heap_t<key_t, data_t, key_cmp_t, key_shuffler_t, data_shuffler_t>& heap, score_data_pair_t<key_t, data_t>* output) {
        /* score_data_pair_t<key_t, data_t> sdp;
        int blockSize = (heap.size() - 1) % 32 + 1;
        for (int block = heap.size() / 32; block >= 0; --block) {
            for (int item = blockSize - 1; item >= 0; --item) {
                key_t score = heap.peekKey();
                data_t data = heap.peekValue();
                if (threadIdx.x == item) {
                    sdp.score = score;
                    sdp.data = data;
                }
				heap.pop();
            }

            if (threadIdx.x < blockSize) {
                output[block * 32 + threadIdx.x] = sdp;
            }
            blockSize = 32;
        } */
        int outputIdx = heap.size();
        while (outputIdx != 0) {
			--outputIdx;
            if (threadIdx.x == 0) {
                output[outputIdx].score = heap.peekKey();
                output[outputIdx].data = heap.nodes[0].value;
            }
            heap.pop();
        }
    }

	template<int LOG_FANOUT, int OUTPUT_LIMIT, int QUEUE_LIMIT, typename elem_t, int DIMS, typename key_t, typename data_t, typename spatial_pred_t, typename score_func_t, typename score_cmp_t>
    __global__ void blockJoinKernel(
        bba_input_reader_t<elem_t, DIMS, key_t, data_t> readerL,
        bba_input_reader_t<elem_t, DIMS, key_t, data_t> readerR,
        int allocationL,
        int allocationR,
        int logBlockSize,
        int outputCapacity,
        spatial_pred_t spatialPred,
        score_func_t scoreFunc,
        score_cmp_t scoreCmp,
        key_t scoreThreshold,
        void* blockOutputBufs,
        size_t blockOutputBufsPitch,
        void* blockStatuses,
        size_t blockStatusesPitch,
        int* uninitializedJoinFlags
    ) {
        using key_shuffler_t = default_shuffler_t<key_t>;
        using data_shuffler_t = default_shuffler_t<data_t>;
        using output_heap_t = heap::heap_t<key_t, tuple_t<data_t>, rev_cmp_t<score_cmp_t>, default_shuffler_t<key_t>, tuple_shuffler_t<data_t, data_shuffler_t>>;

        key_shuffler_t keyShuffler;
        data_shuffler_t dataShuffler;

        __shared__ key_t s_outputHeapMinKeys[OUTPUT_LIMIT];
        __shared__ typename output_heap_t::shared_state_t outputHeapSharedState;
        typename output_heap_t::config_t outputHeapConfig(rev_cmp_t<score_cmp_t>(scoreCmp), keyShuffler, tuple_shuffler_t<data_t, data_shuffler_t>(dataShuffler));
        heap::node_t<key_t, tuple_t<data_t>> outputHeapNodes[OUTPUT_LIMIT];
        output_heap_t outputHeap(outputHeapConfig, outputHeapSharedState, s_outputHeapMinKeys, outputHeapNodes);

        cooperative_groups::thread_block block = cooperative_groups::this_thread_block();
        cooperative_groups::thread_block_tile<32> group = cooperative_groups::tiled_partition<32>(block);

        bool success;

        if (readerL.nextBlock + blockIdx.x < allocationL) {
            success = joinBlocks<false, LOG_FANOUT, QUEUE_LIMIT>(
                readerL.rtrees[readerL.nextBlock + blockIdx.x],
                readerR.rtrees,
                readerR.nextBlock,
                allocationR,
                logBlockSize,
                outputCapacity,
                spatialPred,
                scoreFunc,
                scoreCmp,
                keyShuffler,
                dataShuffler,
                scoreThreshold,
                outputHeap,
                uninitializedJoinFlags + blockIdx.x,
                allocationL - readerL.nextBlock,
                group
			);
        }
        else {
            success = joinBlocks<true, LOG_FANOUT, QUEUE_LIMIT>(
                readerR.rtrees[readerR.nextBlock + blockIdx.x - (allocationL - readerL.nextBlock)],
                readerL.rtrees,
                readerL.nextBlock,
                allocationL,
                logBlockSize,
                outputCapacity,
                spatialPred,
                scoreFunc,
                scoreCmp,
                keyShuffler,
                dataShuffler,
                scoreThreshold,
                outputHeap,
                uninitializedJoinFlags + (blockIdx.x - (allocationL - readerL.nextBlock)) * (allocationL - readerL.nextBlock),
                1,
                group
			);
        }

        if (success) {
            score_data_pair_t<key_t, tuple_t<data_t>>* g_outputBuf = (score_data_pair_t<key_t, tuple_t<data_t>>*)(((char*)blockOutputBufs) + blockOutputBufsPitch * blockIdx.x);
            if (threadIdx.x == 0) {
                bba_result_t* result = (bba_result_t*)(((char*)blockStatuses) + blockStatusesPitch * blockIdx.x);
                result->status = bba_result_t::status_t::OK;
                result->outputSize = outputHeap.size();
            }
            deconstructHeap(outputHeap, g_outputBuf);
        }
        else {
            if (threadIdx.x == 0) {
                // NOTE: This printf must not be removed!
                // The optimizer may be overaggressive and remove a vital shuffle instruction in the heap...
                // ... Unless the kernel is compiled with a printf call!
				printf("Queue overflow!\n");
                bba_result_t* result = (bba_result_t*)(((char*)blockStatuses) + blockStatusesPitch * blockIdx.x);
                result->status = bba_result_t::status_t::QUEUE_OVERFLOW;
            }
        }
    }

    // returns false if any of the blocks ran out of queue space
    template<typename score_t, typename data_t, typename score_cmp_t, typename score_shuffler_t, typename data_shuffler_t, typename group_t>
    __device__ bool collectOutputs(
        int numBlocks,
        void* blockOutputBufs,
        int blockOutputBufsPitch,
        void* blockStatuses,
        int blockStatusesPitch,
        heap::heap_t<score_t, tuple_t<data_t>, rev_cmp_t<score_cmp_t>, score_shuffler_t, tuple_shuffler_t<data_t, data_shuffler_t>>& outputHeap,
        int outputCapacity,
        score_t& scoreThreshold,
        score_cmp_t scoreCmp,
        score_shuffler_t scoreShuffler,
        tuple_shuffler_t<data_t, data_shuffler_t> dataTupleShuffler,
        group_t group
    ) {
		const score_data_pair_t<score_t, tuple_t<data_t>>* iter = nullptr;
		const score_data_pair_t<score_t, tuple_t<data_t>>* last = nullptr;

        bba_result_t* result = (bba_result_t*)(((char*)blockStatuses) + group.thread_rank() * blockStatusesPitch);

        if (group.any(group.thread_rank() < numBlocks && result->status != bba_result_t::status_t::OK)) {
            return false;
        }

        if (group.thread_rank() < numBlocks) {
            iter = (score_data_pair_t<score_t, tuple_t<data_t>>*)(((char*)blockOutputBufs) + blockOutputBufsPitch * group.thread_rank());
            last = iter + result->outputSize;
        }

		while (true) {
			// select the best item from available blocks (if any)
			int bestBlock = -1;
			score_t bestScore = scoreThreshold;
			if (iter != last) {
				bestBlock = group.thread_rank();
				bestScore = iter->score;
			}

			for (int f = 1; f < numBlocks; f <<= 1) {
                int altBestBlock = group.shfl_down(bestBlock, f);
				score_t altBestScore = scoreShuffler.group_shfl_down(group, bestScore, f);
				if (group.thread_rank() + f < numBlocks && scoreCmp(altBestScore, bestScore)) {
					bestBlock = altBestBlock;
					bestScore = altBestScore;
				}
			}

            // the first thread knows the best block
			bestBlock = group.shfl(bestBlock, 0);

			if (bestBlock == -1) {
				// the "best" block is exhausted
                // this also implies that the best score is the score threshold
                // this may mean that we exhausted all the outputs
                // it may also mean that there is no output with a score better than the score threshold
				// we can return here because no more items can be collected satisyfing the score threshold
				return true;
			}

            // the first thread also knows the best score
			bestScore = scoreShuffler.group_shfl(group, bestScore, 0);
			tuple_t<data_t> bestValue;
			if (group.thread_rank() == bestBlock) {
				bestValue = iter->data;
				++iter;
			}
            // broadcast the best value from the thread holding the best block
			bestValue = dataTupleShuffler.group_shfl(group, bestValue, bestBlock);

			// we now have the best item in all threads, ready to place it into the heap
			// we can only place the item into the heap if it has a score better than the score threshold
            // the score threshold is equal to the score of the worst item in the heap (but may be some other threshold until then)
            if (scoreCmp(bestScore, scoreThreshold)) {
				// is the output heap at capacity?
				if (outputHeap.size() != outputCapacity) {
					// the output heap is not at capacity, so just insert the item
                    // this may cause the heap to reach capacity, which will update the score threshold below
					outputHeap.insert(bestScore, bestValue);
				}
				else {
					// the heap is at capacity, we must swap the worst item in the heap with this item
					// the best item is better than the worst item in the heap, so this may update the score threshold
					outputHeap.popInsert(bestScore, bestValue);
				}
#ifdef _DEBUG_OUTPUT
                outputHeap.debugDuplicateValues();
#endif
				
				if (outputHeap.size() == outputCapacity) {
					// the heap either just reached its capacity, or the previous worst item in the heap was replaced by a better one
					// update the score threshold
					scoreThreshold = outputHeap.peekKey();
				}
            }
            else {
				// the best item is no better than the score threshold (the worst item in the heap)
				// this is bound to also be true for all subsequent items
				// we can return here because no more items can be collected satisyfing the score threshold
                return true;
            }
		}
    }

	template<int LOG_FANOUT, int BLOCK_PARALLELISM, int OUTPUT_LIMIT, int QUEUE_LIMIT, typename elem_t, int DIMS, typename key_t, typename data_t, typename key_agg_t, typename spatial_pred_t, typename score_func_t, typename score_cmp_t, typename key_shuffler_t, typename value_shuffler_t, typename group_t>
    __device__ bba_result_t::status_t runBbaInternal(
        bba_input_reader_t<elem_t, DIMS, key_t, data_t> readers[2],
        int logBlockSize,
        key_agg_t keyAgg,
        spatial_pred_t spatialPred,
        score_func_t scoreFunc,
        score_cmp_t scoreCmp,
        key_t scoreThreshold,
        int outputCapacity,
        void* blockOutputBufs,
        size_t blockOutputBufsPitch,
        void* blockStatuses,
        size_t blockStatusesPitch,
        int* uninitializedJoinFlags,
        void* allocation,
        int allocationCapacity,
        heap::heap_t<key_t, tuple_t<data_t>, rev_cmp_t<score_cmp_t>, key_shuffler_t, value_shuffler_t>& outputHeap,
        group_t group
    ) {
        const int ALLOCATION_ALIGNMENT = 256;
        int allocatedBytes = 0;

        while (true) {
			int allocations[2];
            int totalAllocations = allocateNextBlocks(readers, 1 << logBlockSize, scoreFunc, scoreCmp, scoreThreshold, BLOCK_PARALLELISM, allocations);
            if (totalAllocations > 0) {
                cudaError_t err;
                if (group.thread_rank() == 0) {
                    int rtreeRecordCount = 0;
                    int segmentSize = 1 << logBlockSize;
                    while (segmentSize > 1) {
                        int nodeCount = (segmentSize + (1 << LOG_FANOUT) - 2) >> LOG_FANOUT;
                        rtreeRecordCount += nodeCount << LOG_FANOUT;
                        segmentSize = nodeCount;
                    }

                    // allocate space for the r-trees that will be produced
					#pragma unroll
                    for (int i = 0; i < 2; ++i) {
                        for (int j = readers[i].nextBlock; j < allocations[i]; ++j) {
                            int allocationSize = sizeof(record_t<elem_t, DIMS, key_t, data_t>) * rtreeRecordCount;
                            int paddedAllocationSize = ALLOCATION_ALIGNMENT * ((allocationSize + ALLOCATION_ALIGNMENT - 1) / ALLOCATION_ALIGNMENT);
                            if (allocatedBytes + paddedAllocationSize <= allocationCapacity) {
                                record_t<elem_t, DIMS, key_t, data_t>* records = (record_t<elem_t, DIMS, key_t, data_t>*)(((char*)allocation) + allocatedBytes);
                                allocatedBytes += paddedAllocationSize;
                                readers[i].rtrees[j].records = records;
                            }
                            else {
                                return bba_result_t::status_t::OUT_OF_MEMORY;
                            }
                        }
                    }

                    blockInitializeKernel<LOG_FANOUT><<<totalAllocations, 1 << logBlockSize, (sizeof(record_t<elem_t, DIMS, key_t, data_t>) << logBlockSize)>>>(readers[0], readers[1], logBlockSize, allocations[0], keyAgg);
                    cudaDeviceSynchronize();
                    err = cudaGetLastError();
                }

                err = (cudaError_t)group.shfl((int)err, 0);
                if (err != cudaSuccess) {
                    return bba_result_t::status_t::CUDA_ERROR;
                }

                if (group.thread_rank() == 0) {
					const int joinFlagsSize = BLOCK_PARALLELISM * BLOCK_PARALLELISM * sizeof(unsigned int);
					memset(uninitializedJoinFlags, 0, joinFlagsSize);

					blockJoinKernel<LOG_FANOUT, OUTPUT_LIMIT, QUEUE_LIMIT><<<totalAllocations, 32>>>(
						readers[0],
						readers[1],
						allocations[0],
						allocations[1],
						logBlockSize,
						outputCapacity,
						spatialPred,
						scoreFunc,
						scoreCmp,
						scoreThreshold,
						blockOutputBufs,
						blockOutputBufsPitch,
                        blockStatuses,
                        blockStatusesPitch,
						uninitializedJoinFlags
					);

					cudaDeviceSynchronize();
                    err = cudaGetLastError();

/* #ifdef _DEBUG_OUTPUT
                    for (int output = 0; output < totalAllocations; ++output) {
                        printf("\nOutput #%d has %d entries:\n", output, blockOutputSizes[output]);
						score_data_pair_t<key_t, tuple_t<data_t>>* iter = (score_data_pair_t<key_t, tuple_t<data_t>>*)(((char*)blockOutputBufs) + blockOutputBufsPitch * output);
						for (int i = 0; i < blockOutputSizes[output]; ++i) {
							printf("#%02d s: %04d, l: %04d, r: %04d, d: %08d\n", i + 1, iter[i].score, iter[i].data.left, iter[i].data.right, iter[i].data._debugId);
						}
                    }
#endif */
                }

                err = (cudaError_t)group.shfl((int)err, 0);
                if (err != cudaSuccess) {
                    return bba_result_t::status_t::CUDA_ERROR;
                }

                if (!collectOutputs(totalAllocations, blockOutputBufs, blockOutputBufsPitch, blockStatuses, blockStatusesPitch, outputHeap, outputCapacity, scoreThreshold, scoreCmp, outputHeap.config.keyShuffler, outputHeap.config.valueShuffler, group)) {
                    return bba_result_t::status_t::QUEUE_OVERFLOW;
                }

				#pragma unroll
                for (int i = 0; i < 2; ++i) {
                    readers[i].nextBlock = allocations[i];
                }
            }
            else {
                // we could not allocate any more blocks satisfying the score predicate
                // this means we can terminate
                break;
            }
        }

        return bba_result_t::status_t::OK;
    }

	template<int LOG_FANOUT, int BLOCK_PARALLELISM, int OUTPUT_LIMIT, int QUEUE_LIMIT, typename elem_t, int DIMS, typename key_t, typename data_t, typename key_agg_t, typename spatial_pred_t, typename score_func_t, typename score_cmp_t, typename group_t>
	__device__ void runBba(
        bba_result_t* result,
		bba_input_t<elem_t, DIMS, key_t, data_t> inputL,
		bba_input_t<elem_t, DIMS, key_t, data_t> inputR,
        int logBlockSize,
        key_agg_t keyAgg,
        spatial_pred_t spatialPred,
        score_func_t scoreFunc,
        score_cmp_t scoreCmp,
        key_t scoreThreshold,
		int outputCapacity,
		score_data_pair_t<key_t, tuple_t<data_t>>* g_outputBuf,
        void* blockOutputBufs,
        size_t blockOutputBufsPitch,
        void* blockStatuses,
        size_t blockStatusesPitch,
        int* uninitializedJoinFlags,
        void* allocation,
        int allocationCapacity,
        group_t group
	) {
        bba_input_reader_t<elem_t, DIMS, key_t, data_t> readers[2] = {
            bba_input_reader_t<elem_t, DIMS, key_t, data_t>(inputL),
            bba_input_reader_t<elem_t, DIMS, key_t, data_t>(inputR)
        };

        using key_shuffler_t = default_shuffler_t<key_t>;
        using data_shuffler_t = default_shuffler_t<data_t>;
        using data_tuple_shuffler_t = tuple_shuffler_t<data_t, data_shuffler_t>;
        using output_heap_t = heap::heap_t<key_t, tuple_t<data_t>, rev_cmp_t<score_cmp_t>, default_shuffler_t<key_t>, tuple_shuffler_t<data_t, data_shuffler_t>>;

        key_shuffler_t keyShuffler;
        data_shuffler_t dataShuffler;
        data_tuple_shuffler_t dataTupleShuffler(dataShuffler);

        __shared__ key_t s_outputHeapMinKeys[OUTPUT_LIMIT];
        __shared__ typename output_heap_t::shared_state_t outputHeapSharedState;
        typename output_heap_t::config_t outputHeapConfig(rev_cmp_t<score_cmp_t>(scoreCmp), keyShuffler, dataTupleShuffler);
        heap::node_t<key_t, tuple_t<data_t>> outputHeapNodes[OUTPUT_LIMIT];
        output_heap_t outputHeap(outputHeapConfig, outputHeapSharedState, s_outputHeapMinKeys, outputHeapNodes);

        bba_result_t::status_t status = runBbaInternal<LOG_FANOUT, BLOCK_PARALLELISM, OUTPUT_LIMIT, QUEUE_LIMIT, elem_t, DIMS, key_t, data_t, key_agg_t, spatial_pred_t, score_func_t, score_cmp_t, key_shuffler_t, tuple_shuffler_t<data_t, data_shuffler_t>, group_t>(
            readers,
            logBlockSize,
            keyAgg,
            spatialPred,
            scoreFunc,
            scoreCmp,
            scoreThreshold,
            outputCapacity,
            blockOutputBufs,
            blockOutputBufsPitch,
            blockStatuses,
            blockStatusesPitch,
            uninitializedJoinFlags,
            allocation,
            allocationCapacity,
            outputHeap,
            group
		);

        if (status == bba_result_t::status_t::OK) {
			if (threadIdx.x == 0) {
				result->status = bba_result_t::status_t::OK;
				result->outputSize = outputHeap.size();
			}

			deconstructHeap(outputHeap, g_outputBuf);
        }
        else {
            if (threadIdx.x == 0) {
				result->status = status;
            }
        }
	}

	template<int LOG_FANOUT, int BLOCK_PARALLELISM, int OUTPUT_LIMIT, int QUEUE_LIMIT, typename elem_t, int DIMS, typename key_t, typename data_t, typename key_agg_t, typename spatial_pred_t, typename score_func_t, typename score_cmp_t>
    __global__ void runBbaKernel(
        bba_result_t* result,
		bba_input_t<elem_t, DIMS, key_t, data_t> inputL,
		bba_input_t<elem_t, DIMS, key_t, data_t> inputR,
        int logBlockSize,
        key_agg_t keyAgg,
        spatial_pred_t spatialPred,
        score_func_t scoreFunc,
        score_cmp_t scoreCmp,
        key_t scoreThreshold,
		int outputCapacity,
		score_data_pair_t<key_t, tuple_t<data_t>>* g_outputBuf,
        void* blockOutputBufs,
        size_t blockOutputBufsPitch,
        void* blockStatuses,
        size_t blockStatusesPitch,
        int* uninitializedJoinFlags,
        void* allocation,
        int allocationSize
    ) {
        cooperative_groups::thread_block_tile<32> tile = cooperative_groups::tiled_partition<32>(cooperative_groups::this_thread_block());
        runBba<LOG_FANOUT, BLOCK_PARALLELISM, OUTPUT_LIMIT, QUEUE_LIMIT>(
            result,
            inputL,
            inputR,
            logBlockSize,
            keyAgg,
            spatialPred,
            scoreFunc,
            scoreCmp,
            scoreThreshold,
            outputCapacity,
            g_outputBuf,
            blockOutputBufs,
            blockOutputBufsPitch,
            blockStatuses,
            blockStatusesPitch,
            uninitializedJoinFlags,
            allocation,
            allocationSize,
			tile
		);
    }
}

#endif //BBA_CU_H
