#ifndef _BNB_SEQ_H
#define _BNB_SEQ_H

#include <array>
#include "shared.h"
#include "output.h"

namespace bnb {
    template<typename elem_t, int DIMS, typename key_t, typename data_t>
    struct bnb_input_reader_t {
        const record_t<elem_t, DIMS, key_t, data_t>* first;
        const record_t<elem_t, DIMS, key_t, data_t>* next;
        const record_t<elem_t, DIMS, key_t, data_t>* last;

        void initialize(range_t<const record_t<elem_t, DIMS, key_t, data_t>*> records);
        bool canRead() const;
        const key_t& getReadScoreUB() const;
        const key_t& getProbeScoreUB() const;
        const record_t<elem_t, DIMS, key_t, data_t>* readNext();
    };

    template<typename elem_t, int DIMS, typename key_t, typename data_t>
    void bnb_input_reader_t<elem_t, DIMS, key_t, data_t>::initialize(range_t<const record_t<elem_t, DIMS, key_t, data_t>*> records) {
        this->first = records.first;
        this->next = records.first;
        this->last = records.last;
    }

    template<typename elem_t, int DIMS, typename key_t, typename data_t>
    bool bnb_input_reader_t<elem_t, DIMS, key_t, data_t>::canRead() const {
        return this->next < this->last;
    }

    template<typename elem_t, int DIMS, typename key_t, typename data_t>
    const key_t& bnb_input_reader_t<elem_t, DIMS, key_t, data_t>::getReadScoreUB() const {
        return this->next->key;
    }

    template<typename elem_t, int DIMS, typename key_t, typename data_t>
    const key_t& bnb_input_reader_t<elem_t, DIMS, key_t, data_t>::getProbeScoreUB() const {
        return this->first->key;
    }

    template<typename elem_t, int DIMS, typename key_t, typename data_t>
    const record_t<elem_t, DIMS, key_t, data_t>* bnb_input_reader_t<elem_t, DIMS, key_t, data_t>::readNext() {
        return this->next++;
    }

    template<typename elem_t, int DIMS, typename key_t, typename data_t, typename score_func_t, typename score_cmp_t>
    bool selectInput(const std::array<bnb_input_reader_t<elem_t, DIMS, key_t, data_t>, 2>& readers, score_func_t scoreFunc, score_cmp_t scoreCmp, key_t scoreThreshold, size_t& readInput) {
        // find which input to read and which input to probe
        // the goal is maximize the upper bound for the combined score of any produced results
        bool anyFound = false;
        key_t bestScoreUB = scoreThreshold;
        for (size_t i = 0; i < 2; ++i) {
            if (readers[i].canRead()) {
                // if we read a block from inputs[i] and probe it against read blocks of inputs[1 - i], this is the upper bound for the combined score any produced results
                key_t scoreUB = scoreFunc(readers[i].getReadScoreUB(), readers[1 - i].getProbeScoreUB());
                if (scoreCmp(scoreUB, bestScoreUB)) {
                    anyFound = true;
                    readInput = i;
                    bestScoreUB = scoreUB;
                }
            }
        }
        return anyFound;
    }

    template<typename elem_t, int DIMS, typename key_t, typename data_t, typename spatial_pred_t, typename score_func_t, typename score_cmp_t>
    void runBnbInternal(
        const std::array<range_t<const record_t<elem_t, DIMS, key_t, data_t>*>, 2>& inputs,
        spatial_pred_t spatialPred,
        score_func_t scoreFunc,
        const output::accumulator_config_t<score_cmp_t>& accConfig,
        output::accumulator_data_t<key_t, tuple_t<data_t>>& accData
    ) {
        std::array<bnb_input_reader_t<elem_t, DIMS, key_t, data_t>, 2> readers{};
        for (int i = 0; i < 2; ++i) {
            readers[i].initialize(inputs[i]);
        }
        size_t readInput;
        while (selectInput(readers, scoreFunc, accConfig.scoreCmp, accData.lowerBound, readInput)) {
			size_t probeInput = 1 - readInput;

			std::array<const record_t<elem_t, DIMS, key_t, data_t>*, 2> records{};
			records[readInput] = readers[readInput].readNext();
			for (records[probeInput] = readers[probeInput].first; records[probeInput] < readers[probeInput].next; ++records[probeInput]) {
				if (spatialPred(records[0]->mbb, records[1]->mbb)) {
					key_t score = scoreFunc(records[0]->key, records[1]->key);
					if (accConfig.scoreCmp(score, accData.lowerBound)) {
						score_data_pair_t<key_t, tuple_t<data_t>> outputValue;
						outputValue.score = score;
						outputValue.data.left = records[0]->data;
						outputValue.data.right = records[1]->data;
						accData.insert(outputValue, accConfig);
					}
				}
			}
        }
    }

    template<typename elem_t, int DIMS, typename key_t, typename data_t, typename key_cmp_t>
    struct bnb_input_t {
        range_t<record_t<elem_t, DIMS, key_t, data_t>*> records;
        key_cmp_t keyCmp;

        bnb_input_t(range_t<record_t<elem_t, DIMS, key_t, data_t>*> records, key_cmp_t keyCmp);
    };

    template<typename elem_t, int DIMS, typename key_t, typename data_t, typename key_cmp_t>
    bnb_input_t<elem_t, DIMS, key_t, data_t, key_cmp_t>::bnb_input_t(range_t<record_t<elem_t, DIMS, key_t, data_t>*> records, key_cmp_t keyCmp)
        : records(records)
        , keyCmp(keyCmp)
    {}

    template<typename elem_t, int DIMS, typename key_t, typename data_t, typename key_cmp_t, typename spatial_pred_t, typename score_func_t, typename score_cmp_t>
    void runBnb(
        const std::array<bnb_input_t<elem_t, DIMS, key_t, data_t, key_cmp_t>, 2>& inputs,
        spatial_pred_t spatialPred,
        score_func_t scoreFunc,
        const output::accumulator_config_t<score_cmp_t>& accConfig,
        output::accumulator_data_t<key_t, tuple_t<data_t>>& accData
    ) {
        std::array<record_t<elem_t, DIMS, key_t, data_t>*, 2> copies {};
        std::array<range_t<const record_t<elem_t, DIMS, key_t, data_t>*>, 2> sortedRanges {};
        for (size_t i = 0; i < 2; ++i) {
            size_t inputSize = inputs[i].records.last - inputs[i].records.first;
            copies[i] = new record_t<elem_t, DIMS, key_t, data_t>[inputSize];
            sortedRanges[i].first = copies[i];
            sortedRanges[i].last = copies[i] + inputSize;
            std::copy(inputs[i].records.first, inputs[i].records.last, copies[i]);
            key_cmp_t keyCmp = inputs[i].keyCmp;
            std::sort(copies[i], copies[i] + inputSize, [&keyCmp](const record_t<elem_t, DIMS, key_t, data_t>& lhs, const record_t<elem_t, DIMS, key_t, data_t>& rhs) {
                return keyCmp(lhs.key, rhs.key);
			});
        }

        runBnbInternal(sortedRanges, spatialPred, scoreFunc, accConfig, accData);

        for (int i = 0; i < 2; ++i) {
            delete[] copies[i];
        }
    }
}

#endif //_BNB_SEQ_H
