#ifndef _HEAP_H
#define _HEAP_H

#include <cuda_runtime.h>

// #define _DEBUG_HEAP

namespace heap {
    template<typename key_cmp_t, typename key_shuffler_t, typename value_shuffler_t>
    struct heap_config_t {
        key_cmp_t keyCmp;
        key_shuffler_t keyShuffler;
        value_shuffler_t valueShuffler;

        __device__ heap_config_t(key_cmp_t keyCmp, key_shuffler_t keyShuffler, value_shuffler_t valueShuffler)
            : keyCmp(keyCmp)
            , keyShuffler(keyShuffler)
            , valueShuffler(valueShuffler) {}
    };

    template<typename key_t, typename value_t>
    struct kv_pair_t {
        key_t key;
        value_t value;
    };

    template<typename key_t, typename value_t>
    struct node_t {
        key_t key;
        value_t value;

        __device__ void initialize(const kv_pair_t<key_t, value_t> minItem) {
            if (threadIdx.x == 0) {
                this->key = minItem.key;
                this->value = minItem.value;
            }
        }

        __device__ static unsigned int sizeMask(int size) {
            return size == 32 ? 0xffffffff : ((1 << size) - 1);
        }

		template<typename key_cmp_t, typename key_shuffler_t, typename value_shuffler_t>
		__device__ void insert(const kv_pair_t<key_t, value_t> item, int size, const heap_config_t<key_cmp_t, key_shuffler_t, value_shuffler_t>& config) {
			unsigned int vote = __ballot_sync(0xffffffff, !config.keyCmp(item.key, this->key)) & sizeMask(size);
			unsigned int b;
			// find the position of first "1" in the warp
			asm("bfind.u32 %0, %1;" : "=r"(b) : "r"(vote));
            ++b;

			// we have keys greater than maxKey in the array, so evict the max item in the array by shifting the items up
			if (threadIdx.x >= b) {
                this->key = config.keyShuffler.shfl_up(~vote, this->key, 1);
                this->value = config.valueShuffler.shfl_up(~vote, this->value, 1);
			}
			if (threadIdx.x == b) {
				this->key = item.key;
				this->value = item.value;
			}
		}

        // effectively inserts an item into a full node, then pops the max item and returns it
        // the previous max item is guaranteed to be popped if maxItem.key >= item.key
		template<typename key_cmp_t, typename key_shuffler_t, typename value_shuffler_t>
		__device__ kv_pair_t<key_t, value_t> insertPopMax(const kv_pair_t<key_t, value_t> item, const heap_config_t<key_cmp_t, key_shuffler_t, value_shuffler_t>& config) {
			unsigned int vote = __ballot_sync(0xffffffff, !config.keyCmp(item.key, this->key));
			unsigned int b;
			// find the position of first "1" in the warp
			asm("bfind.u32 %0, %1;" : "=r"(b) : "r"(vote));
            ++b;

			if (b == 32) {
				// item.key is not smaller than any key in the node, thus the input item would be popped as the max item
                return item;
			}

			// we have keys larger than item.key in the array, so pop the max item in the array by shifting the items up
            // the current final item in the node is the max item
			// broadcast the current max item in the array so all threads can return it
            kv_pair_t<key_t, value_t> excessItem {
				config.keyShuffler.shfl(0xffffffff, this->key, 31),
				config.valueShuffler.shfl(0xffffffff, this->value, 31)
            };
			if (threadIdx.x >= b) {
                this->key = config.keyShuffler.shfl_up(~vote, this->key, 1);
                this->value = config.valueShuffler.shfl_up(~vote, this->value, 1);
			}
			if (threadIdx.x == b) {
				this->key = item.key;
				this->value = item.value;
			}
            return excessItem;
		}

        // effectively inserts an item into a full node, then pops the max item
        // the previous max item is guaranteed to be popped if maxItem.key >= item.key
		template<typename key_cmp_t, typename key_shuffler_t, typename value_shuffler_t>
		__device__ void insertPopMaxDiscard(const kv_pair_t<key_t, value_t> item, const heap_config_t<key_cmp_t, key_shuffler_t, value_shuffler_t>& config) {
			unsigned int vote = __ballot_sync(0xffffffff, !config.keyCmp(item.key, this->key));
			unsigned int b;
			// find the position of first "1" in the warp
			asm("bfind.u32 %0, %1;" : "=r"(b) : "r"(vote));
            ++b;

			if (b == 0) {
				// item.key is not smaller than all keys in the node, thus the input item would be popped as the max item
                return;
			}

			// we have keys larger than item.key in the array, so pop the max item in the array by shifting the items up
			if (threadIdx.x >= b) {
                this->key = config.keyShuffler.shfl_up(~vote, this->key, 1);
                this->value = config.valueShuffler.shfl_up(~vote, this->value, 1);
			}
			if (threadIdx.x == b) {
				this->key = item.key;
				this->value = item.value;
			}
		}

        // insert an item, given that the node is non-full and that the item is not larger than the current min item in the node
		template<typename key_cmp_t, typename key_shuffler_t, typename value_shuffler_t>
        __device__ void insertMin(const kv_pair_t<key_t, value_t> item, const heap_config_t<key_cmp_t, key_shuffler_t, value_shuffler_t>& config) {
            this->key = config.keyShuffler.shfl_up(0xffffffff, this->key, 1);
            this->value = config.valueShuffler.shfl_up(0xffffffff, this->value, 1);

            if (threadIdx.x == 0) {
                this->key = item.key;
                this->value = item.value;
            }
        }

        // pop the min item
		template<typename key_cmp_t, typename key_shuffler_t, typename value_shuffler_t>
        __device__ void popMin(const heap_config_t<key_cmp_t, key_shuffler_t, value_shuffler_t>& config) {
            this->key = config.keyShuffler.shfl_down(0xffffffff, this->key, 1);
            this->value = config.valueShuffler.shfl_down(0xffffffff, this->value, 1);
        }

        // effectively inserts the item into the node, then pops the min item of the node and returns it
        // the previous min item is guaranteed to be popped if minItem.key <= item.key
        // returns the popped min item
		template<typename key_cmp_t, typename key_shuffler_t, typename value_shuffler_t>
		__device__ kv_pair_t<key_t, value_t> insertPopMin(const kv_pair_t<key_t, value_t> item, int size, const heap_config_t<key_cmp_t, key_shuffler_t, value_shuffler_t>& config) {
			unsigned int vote = __ballot_sync(0xffffffff, !config.keyCmp(item.key, this->key)) & sizeMask(size);
			unsigned int b;
			// find the position of first "1" in the warp
			asm("bfind.u32 %0, %1;" : "=r"(b) : "r"(vote));

			if (b == 0xffffffff) {
				// item.key is already smaller than all keys in the array, no work needs to be done
                return item;
			}

			// we have keys smaller than minKey in the array, so evict the min item in the array by shifting the items down
			// broadcast the min item in the array so all threads can return it
            kv_pair_t<key_t, value_t> nextMin {
				config.keyShuffler.shfl(0xffffffff, this->key, 0),
				config.valueShuffler.shfl(0xffffffff, this->value, 0)
            };
			if (threadIdx.x <= b) {
                this->key = config.keyShuffler.shfl_down(vote, this->key, 1);
                this->value = config.valueShuffler.shfl_down(vote, this->value, 1);
			}
			if (threadIdx.x == b) {
				this->key = item.key;
				this->value = item.value;
			}
            return nextMin;
		}

        // effectively inserts the item into the node, then pops the min item of the node
        // the previous min item is guaranteed to be popped if minItem.key <= item.key
		template<typename key_cmp_t, typename key_shuffler_t, typename value_shuffler_t>
		__device__ void insertPopMinDiscard(const kv_pair_t<key_t, value_t> item, int size, const heap_config_t<key_cmp_t, key_shuffler_t, value_shuffler_t>& config) {
			unsigned int vote = __ballot_sync(0xffffffff, !config.keyCmp(item.key, this->key)) & sizeMask(size);
			unsigned int b;
			// find the position of first "1" in the warp
			asm("bfind.u32 %0, %1;" : "=r"(b) : "r"(vote));

			if (b == 0xffffffff) {
				// item.key is already smaller than all keys in the array, no work needs to be done
                return;
			}
			// printf("we did a discard!\n");

			// we have keys smaller than minKey in the array, so evict the min item in the array by shifting the items down
			if (threadIdx.x <= b) {
                this->key = config.keyShuffler.shfl_down(vote, this->key, 1);
                this->value = config.valueShuffler.shfl_down(vote, this->value, 1);
			}
			if (threadIdx.x == b) {
				this->key = item.key;
				this->value = item.value;
			}
		}

        // effectively pops the min item of a full node, then inserts the item into the node
        // returns the min item that was popped
		template<typename key_cmp_t, typename key_shuffler_t, typename value_shuffler_t>
		__device__ kv_pair_t<key_t, value_t> popMinInsert(const kv_pair_t<key_t, value_t> item, const heap_config_t<key_cmp_t, key_shuffler_t, value_shuffler_t>& config) {
			// broadcast the min item in the array so all threads can return it
            kv_pair_t<key_t, value_t> prevMin {
				config.keyShuffler.shfl(0xffffffff, this->key, 0),
				config.valueShuffler.shfl(0xffffffff, this->value, 0)
            };

			unsigned int vote = __ballot_sync(0xffffffff, !config.keyCmp(item.key, this->key)) | 1;
			unsigned int b;
			// find the position of first "1" in the warp
			asm("bfind.u32 %0, %1;" : "=r"(b) : "r"(vote));
            ++b;

			// evict the min item in the array by shifting the items down
			if (threadIdx.x <= b) {
                this->key = config.keyShuffler.shfl_down(vote, this->key, 1);
                this->value = config.valueShuffler.shfl_down(vote, this->value, 1);
			}
			if (threadIdx.x == b) {
				this->key = item.key;
				this->value = item.value;
			}
            return prevMin;
		}

        // effectively pops the min item of a full node, then inserts the item into the node
		template<typename key_cmp_t, typename key_shuffler_t, typename value_shuffler_t>
		__device__ void popMinInsertDiscard(const kv_pair_t<key_t, value_t> item, int size, const heap_config_t<key_cmp_t, key_shuffler_t, value_shuffler_t>& config) {
			unsigned int vote = __ballot_sync(0xffffffff, !config.keyCmp(item.key, this->key)) & sizeMask(size) | 1;
			unsigned int b;
			// find the position of first "1" in the warp
			asm("bfind.u32 %0, %1;" : "=r"(b) : "r"(vote));

			// evict the min item in the array by shifting the items down
			if (threadIdx.x <= b) {
                this->key = config.keyShuffler.shfl_down(vote, this->key, 1);
                this->value = config.valueShuffler.shfl_down(vote, this->value, 1);
			}
			if (threadIdx.x == b) {
				this->key = item.key;
				this->value = item.value;
			}
		}

		template<typename key_cmp_t, typename key_shuffler_t, typename value_shuffler_t>
        __device__ kv_pair_t<key_t, value_t> popMax(int size, const heap_config_t<key_cmp_t, key_shuffler_t, value_shuffler_t>& config) {
            return {
				config.keyShuffler.shfl(0xffffffff, this->key, size),
				config.valueShuffler.shfl(0xffffffff, this->value, size)
            };
        }

		template<typename key_cmp_t, typename key_shuffler_t, typename value_shuffler_t>
        __device__ int debugSortDiscontinuity(int size, const heap_config_t<key_cmp_t, key_shuffler_t, value_shuffler_t>& config) {
            key_t altKey = config.keyShuffler.shfl_down(0xffffffff, this->key, 1);
            return __ballot_sync(0xffffffff, (threadIdx.x < size - 1) & config.keyCmp(altKey, this->key));
        }
    };

// #define _ENABLE_SHARED_STATE

    template<typename key_t, typename value_t, typename key_cmp_t, typename key_shuffler_t, typename value_shuffler_t>
    struct heap_t {
        using config_t = heap_config_t<key_cmp_t, key_shuffler_t, value_shuffler_t>;
        struct shared_state_t {
#ifdef _ENABLE_SHARED_STATE
            int numNodes;
            int lastNodeSize;
#endif
        };

        const heap_config_t<key_cmp_t, key_shuffler_t, value_shuffler_t>& config;
        shared_state_t& sharedState;
        key_t* minKeys;
        node_t<key_t, value_t>* nodes;
#ifndef _ENABLE_SHARED_STATE
        int numNodes;
        int lastNodeSize;
#endif

        __device__ heap_t(const heap_config_t<key_cmp_t, key_shuffler_t, value_shuffler_t>& config, shared_state_t& sharedState, key_t* minKeys, node_t<key_t, value_t>* nodes)
            : config(config)
            , sharedState(sharedState)
            , minKeys(minKeys)
            , nodes(nodes)
#ifndef _ENABLE_SHARED_STATE
            , numNodes(0)
            , lastNodeSize(32)
#endif
        {
#ifdef _ENABLE_SHARED_STATE
            if (threadIdx.x == 0) {
                this->sharedState.numNodes = 0;
                this->sharedState.lastNodeSize = 32;
            }
            __syncwarp();
#endif
        }

        __device__ int getNumNodes() const {
#ifdef _ENABLE_SHARED_STATE
            return this->sharedState.numNodes;
#else
            return this->numNodes;
#endif
        }

        __device__ int getLastNodeSize() const {
#ifdef _ENABLE_SHARED_STATE
            return this->sharedState.lastNodeSize;
#else
            return this->lastNodeSize;
#endif
        }

        __device__ int size() const {
            return (this->getNumNodes() - 1) * 32 + this->getLastNodeSize();
        }

        __device__ bool isEmpty() const {
            return this->getNumNodes() == 0;
        }

        __device__ void updateMinKey(int index) {
			__syncwarp();
            if (threadIdx.x == 0) {
				this->minKeys[index] = this->nodes[index].key;
            }
            __syncwarp();
        }

        __device__ void setMinKey(int index, const key_t& key) {
            if (threadIdx.x == 0) {
				this->minKeys[index] = key;
            }
            __syncwarp();
        }

        __device__ key_t getMinKey(int index) const {
            return this->minKeys[index];
        }

        __device__ int parentOf(int index) const {
			return (index - 1) / 32;
        }

        __device__ int firstChildOf(int index) const {
            return 1 + index * 32;
        }

        __device__ void empty() {
#ifdef _ENABLE_SHARED_STATE
			__syncwarp();
            if (threadIdx.x == 0) {
                this->sharedState.numNodes = 0;
                this->sharedState.lastNodeSize = 32;
            }
            __syncwarp();
#else
            this->numNodes = 0;
            this->lastNodeSize = 32;
#endif
        }

        __device__ kv_pair_t<key_t, value_t> propagateUp(const kv_pair_t<key_t, value_t> item, int index) {
			const int STACK_HEIGHT = 4;
			int nodeStack[STACK_HEIGHT];

			// find the path to the first ancestor that is either the root or whose key is smaller than the key of the item being propagated up
			int stackPos = 0;
			while (index > 0) {
				index = this->parentOf(index);
				if (!this->config.keyCmp(item.key, this->getMinKey(index))) {
					break;
				}
				else {
					nodeStack[stackPos] = index;
					++stackPos;
				}
			}

            kv_pair_t<key_t, value_t> excessItem = item;

            // place the item in the ancestor, shifting max items down to the source node
			while (stackPos > 0) {
				--stackPos;

				index = nodeStack[stackPos];
                excessItem = this->nodes[index].insertPopMax(excessItem, this->config);
                this->updateMinKey(index);
			}

            return excessItem;
        }

        __device__ int minChild(int firstChild, int numChildren) const {
            if (numChildren == 1) {
                return firstChild;
            }

            key_t minKey;
            int minChildIndex = firstChild + threadIdx.x;

            if (threadIdx.x < numChildren) {
                minKey = this->getMinKey(minChildIndex);
            }

            for (int i = 1; i < numChildren; i <<= 1) {
                int altMinChildIndex = __shfl_down_sync(0xffffffff, minChildIndex, i);
                key_t altMinKey = this->config.keyShuffler.shfl_down(0xffffffff, minKey, i);
                if ((threadIdx.x + i < numChildren) && this->config.keyCmp(altMinKey, minKey)) {
                    minChildIndex = altMinChildIndex;
                    minKey = altMinKey;
                }
            }

			return __shfl_sync(0xffffffff, minChildIndex, 0);
        }

        __device__ void propagateDown() {
            int parentIndex = 0;
            kv_pair_t<key_t, value_t> parentMinItem;
            while (true) {
				int firstChild = this->firstChildOf(parentIndex);
				int numChildren = min(this->getNumNodes(), firstChild + 32) - firstChild;
                if (numChildren <= 0) {
                    // this node is a leaf node
                    break;
                }

                parentMinItem.key = this->getMinKey(parentIndex);
				int childIndex = this->minChild(firstChild, numChildren);
                key_t childMinKey = this->getMinKey(childIndex);
				if (!this->config.keyCmp(childMinKey, parentMinItem.key)) {
					// heap property is preserved
					break;
				}
                // heap property is not preserved, we have to swap min items between the parent and the child

                parentMinItem.value = this->config.valueShuffler.shfl(0xffffffff, this->nodes[parentIndex].value, 0);

                // directly replace the min entry of the parent with the min entry of the child
                // the min key of the child is smaller than the previous min key of the parent, therefore direct replacement preserves the order of items in the parent
				__syncwarp();
                if (threadIdx.x == 0) {
                    this->nodes[parentIndex].key = childMinKey;
                    this->nodes[parentIndex].value = this->nodes[childIndex].value;
					this->minKeys[parentIndex] = childMinKey;
                }
                __syncwarp();
                int size;
                if (childIndex == this->getNumNodes() - 1) {
                    size = this->getLastNodeSize();
                }
                else {
                    size = 32;
                }
                // place the previous min entry of the parent into the child and discard the previous min item of the child
                // NOTE: Normally, the item returned by childNode.insertPopMin should replace the min item in the parent to guarantee that we don't duplicate/lose any items
                // However, childNode.insertPopMin is guaranteed to truncate the previous min item because parentMinItem.key > childMinKey
				this->nodes[childIndex].insertPopMinDiscard(parentMinItem, size, this->config);
                this->updateMinKey(childIndex);

                parentIndex = childIndex;
            }
        }

        // effectively pops the min item then inserts another
        __device__ void popInsert(const key_t key, const value_t value) {
            if (this->getNumNodes() == 1) {
                // there is only the root node
                // pop the min item of the root node (the only node), then insert the replacement
                this->nodes[0].popMinInsertDiscard({ key, value }, this->getLastNodeSize(), this->config);
                this->updateMinKey(0);
            }
            else {
                // replace the min item of the root node with the last item
                this->nodes[0].popMinInsertDiscard({ key, value }, 32, this->config);
                this->updateMinKey(0);

                // by replacing the min item of the root node, we may violate the heap property
                // propagate down to restore the heap property
				this->propagateDown();
            }
#ifdef _DEBUG_HEAP
            this->debugTest();
#endif
        }

        __device__ void pop() {
            if (this->getNumNodes() == 1) {
                // there is only the root node
                // pop the min item of the root node (the only node)
                this->nodes[0].popMin(this->config);

#ifndef _ENABLE_SHARED_STATE
				if (this->lastNodeSize == 1) {
					--this->numNodes;
					this->lastNodeSize = 32;
				}
				else {
					--this->lastNodeSize;
				}
#endif
				__syncwarp();
				if (threadIdx.x == 0) {
                    this->minKeys[0] = this->nodes[0].key;
#ifdef _ENABLE_SHARED_STATE
					if (this->sharedState.lastNodeSize == 1) {
						--this->sharedState.numNodes;
						this->sharedState.lastNodeSize = 32;
					}
					else {
						--this->sharedState.lastNodeSize;
					}
#endif
				}
				__syncwarp();
            }
            else {
				// pop the last item of the last node
				kv_pair_t<key_t, value_t> lastItem = this->nodes[this->getNumNodes() - 1].popMax(this->getLastNodeSize() - 1, this->config);

#ifdef _ENABLE_SHARED_STATE
				__syncwarp();
				if (threadIdx.x == 0) {
					if (this->sharedState.lastNodeSize == 1) {
						--this->sharedState.numNodes;
						this->sharedState.lastNodeSize = 32;
					}
					else {
						--this->sharedState.lastNodeSize;
					}
				}
				__syncwarp();
#else
				if (this->lastNodeSize == 1) {
					--this->numNodes;
					this->lastNodeSize = 32;
				}
				else {
					--this->lastNodeSize;
				}
#endif

                // replace the min item of the root node with the last item
				this->nodes[0].insertPopMinDiscard(lastItem, 32, this->config);
                this->updateMinKey(0);

                // by replacing the min item of the root node, we may violate the heap property
                // propagate down to restore the heap property
				this->propagateDown();
            }
#ifdef _DEBUG_HEAP
            this->debugTest();
#endif
        }

        __device__ void insert(const key_t key, const value_t value) {
            if (this->getNumNodes() == 0) {
				// the heap is empty, just initialize the root node
                this->nodes[0].initialize({
                    key,
                    value
				});
#ifndef _ENABLE_SHARED_STATE
				++this->numNodes;
				this->lastNodeSize = 1;
#endif
				__syncwarp();
                if (threadIdx.x == 0) {
                    this->minKeys[0] = key;
#ifdef _ENABLE_SHARED_STATE
                    ++this->sharedState.numNodes;
                    this->sharedState.lastNodeSize = 1;
#endif
                }
				__syncwarp();
            }
            else {
                // the heap is not empty
                if (this->getLastNodeSize() != 32) {
                    // there is space for another item in the last node
                    // insert the item into the last node
					key_t lastNodeMinKey = getMinKey(this->getNumNodes() - 1);
                    if (!this->config.keyCmp(key, lastNodeMinKey)) {
						// because the min key of the last node is not larger than the key of the new item, we can easily preserve the heap property
                        // insert the item into the last node
                        this->nodes[this->getNumNodes() - 1].insert({ key, value }, this->getLastNodeSize(), this->config);
#ifdef _ENABLE_SHARED_STATE
						__syncwarp();
						if (threadIdx.x == 0) {
							++this->sharedState.lastNodeSize;
						}
						__syncwarp();
#else
						++this->lastNodeSize;
#endif
                    }
                    else {
						// place the item in a higher node by propagating up, producing an excess item that must be placed back into the heap
						// the excess item is guaranteed to be smaller than the min item of the parent node, so place it in the last node
						kv_pair_t<key_t, value_t> excessItem = this->propagateUp({ key, value }, this->getNumNodes() - 1);
						this->nodes[this->getNumNodes() - 1].insert(excessItem, this->getLastNodeSize(), this->config);
						__syncwarp();
						if (threadIdx.x == 0) {
							this->minKeys[this->getNumNodes() - 1] = this->nodes[this->getNumNodes() - 1].key;
#ifdef _ENABLE_SHARED_STATE
							++this->sharedState.lastNodeSize;
#endif
						}
						__syncwarp();
#ifndef _ENABLE_SHARED_STATE
						++this->lastNodeSize;
#endif
                    }
                }
                else {
                    // there is no space for another item in the last node
                    // place the item in a new node, then propagate it up
                    kv_pair_t<key_t, value_t> initialItem = this->propagateUp({ key, value }, this->getNumNodes());
                    this->nodes[this->getNumNodes()].initialize(initialItem);
					__syncwarp();
					if (threadIdx.x == 0) {
						this->minKeys[this->getNumNodes()] = initialItem.key;
#ifdef _ENABLE_SHARED_STATE
						++this->sharedState.numNodes;
						this->sharedState.lastNodeSize = 1;
#endif
					}
					__syncwarp();
#ifndef _ENABLE_SHARED_STATE
					++this->numNodes;
					this->lastNodeSize = 1;
#endif
                }
            }
#ifdef _DEBUG_HEAP
            this->debugTest();
#endif
        }

        __device__ key_t peekKey() const {
            return this->minKeys[0];
        }

        __device__ value_t peekValue() {
            return this->config.valueShuffler.shfl(0xffffffff, this->nodes[0].value, 0);
        }

        __device__ void debugTest() const {
            if (!this->isEmpty()) {
				for (int i = 0; i < this->getNumNodes() - 1; ++i) {
                    int discontinuity = this->nodes[i].debugSortDiscontinuity(32, this->config);
                    if (discontinuity) {
                        if (threadIdx.x == 0) {
                            printf("Node %d is not sorted!\n", i);
                        }
                    }
				}
				int discontinuity = this->nodes[this->getNumNodes() - 1].debugSortDiscontinuity(this->getLastNodeSize(), this->config);
				if (discontinuity) {
					if (threadIdx.x == 0) {
						printf("Node %d is not sorted!\n", this->getNumNodes() - 1);
					}
				}

                if (threadIdx.x == 0) {
                    for (int i = 0; i < this->getNumNodes(); ++i) {
                        key_t sharedKey = this->minKeys[i];
                        key_t localKey = this->nodes[i].key;
                        if (sharedKey != localKey) {
							printf("Shared key of node %d is out of sync! Should be %d, was %d\n", i, localKey, sharedKey);
                        }
                    }
                }

                for (int i = 1 + threadIdx.x; i < this->getNumNodes(); i += 32) {
                    int parent = this->parentOf(i);
                    key_t parentKey = this->getMinKey(parent);
                    key_t childKey = this->getMinKey(i);
                    if (this->config.keyCmp(childKey, parentKey)) {
						printf("Node %d is violating the heap property!\n", i);
                    }
                }
            }
        }

        __device__ bool debugValueExists(const value_t& value) const {
            bool valueExists = false;
            for (int node = 0; node < this->getNumNodes(); ++node) {
                int nodeSize = (node == this->getNumNodes() - 1) ? this->getLastNodeSize() : 32;
                if (threadIdx.x < nodeSize) {
					if (this->nodes[node].value == value) {
                        valueExists = true;
                        break;
					}
                }
            }

            return __any_sync(0xffffffff, valueExists);
        }

        __device__ void debugAssertValueNotExists(const value_t& value) const {
            for (int node = 0; node < this->getNumNodes(); ++node) {
                int nodeSize = (node == this->getNumNodes() - 1) ? this->getLastNodeSize() : 32;
                if (threadIdx.x < nodeSize) {
					if (this->nodes[node].value == value) {
						printf("A value in the heap already exists!\n");
					}
                }
            }
        }

        __device__ void debugDuplicateValues() const {
            for (int currentNode = 0; currentNode < this->getNumNodes(); ++currentNode) {
                int currentNodeSize = (currentNode == this->getNumNodes() - 1) ? this->getLastNodeSize() : 32;
                for (int currentItem = 0; currentItem < currentNodeSize; ++currentItem) {
                    value_t currentValue = this->config.valueShuffler.shfl(0xffffffff, this->nodes[currentNode].value, currentItem);
                    if (threadIdx.x > currentItem && threadIdx.x < currentNodeSize) {
                        if (this->nodes[currentNode].value == currentValue) {
                            printf("Found a duplicate value in the heap!\n");
                        }
                    }

                    for (int otherNode = currentNode + 1; otherNode < this->getNumNodes(); ++otherNode) {
						int otherNodeSize = (otherNode == this->getNumNodes() - 1) ? this->getLastNodeSize() : 32;
                        if (threadIdx.x < otherNodeSize) {
                            if (this->nodes[otherNode].value == currentValue) {
                                printf("Found a duplicate value in the heap!\n");
                            }
                        }
                    }
                }
            }
        }

        __device__ void debugPrint() const {
			int lastNode = this->getNumNodes() - 1;
            if (lastNode != -1) {
				for (int i = 0; i < this->getNumNodes() - 1; ++i) {
                    for (int j = 0; j < 32; ++j) {
                        key_t key = this->config.keyShuffler.shfl(0xffffffff, this->nodes[i].key, j);
                        value_t value = this->config.valueShuffler.shfl(0xffffffff, this->nodes[i].value, j);
                        if (threadIdx.x == 0) {
							printf("\ts = %06d, level = %d, leftBlock = %06d, rightBlock = %06d\n", key, value.level, value.leftBlock, value.rightBlock);
                        }
                    }
                    if (threadIdx.x == 0) {
                        printf("==============================\n");
                    }
				}

				for (int j = 0; j < this->getLastNodeSize(); ++j) {
					key_t key = this->config.keyShuffler.shfl(0xffffffff, this->nodes[lastNode].key, j);
					value_t value = this->config.valueShuffler.shfl(0xffffffff, this->nodes[lastNode].value, j);
					if (threadIdx.x == 0) {
						printf("\ts = %06d, level = %d, leftBlock = %06d, rightBlock = %06d\n", key, value.level, value.leftBlock, value.rightBlock);
					}
				}
            }
            if (threadIdx.x == 0) {
				printf("\n");
            }
        }
    };
}

#endif //_HEAP_H
