#include <cuda_runtime.h>
#include <stdio.h>
#include <memory>
#include <cassert>
#include <random>
#include <array>
#include <string>
#include <chrono>
#include "shared.h"
#include "str_small.cuh"
#include "bba.cuh"
#include <cub/device/device_radix_sort.cuh>
#include "bnb_seq.h"

// #define _USE_CPU_SORT

template<unsigned int RADIX_BITS, typename elem_t, int DIMS, typename key_t, typename data_t>
void str(record_t<elem_t, DIMS, key_t, data_t>* d_records, unsigned int inputSize, unsigned int fanout) {
    // we need space for all the records in shared memory + digit counters for each record for the radix sort
    size_t sharedMemorySize = (sizeof(record_t<elem_t, DIMS, key_t, data_t>) + (sizeof(unsigned short) << RADIX_BITS)) * inputSize;

    str_small::runStrShared<RADIX_BITS, elem_t, DIMS, key_t, data_t><<<1, inputSize, sharedMemorySize>>>(d_records, inputSize, fanout);
}

template<typename elem_t, int DIMS, typename key_t, typename data_t>
void allocateSortRecordsByKey(int numItems, cub::DoubleBuffer<record_t<elem_t, DIMS, key_t, data_t>>& d_records, cub::DoubleBuffer<key_t>& d_keys, void*& d_tempStorage, size_t& tempStorageBytes) {
    cudaMalloc(&(d_records.d_buffers[0]), sizeof(record_t<elem_t, DIMS, key_t, data_t>) * numItems);
    cudaMalloc(&(d_records.d_buffers[1]), sizeof(record_t<elem_t, DIMS, key_t, data_t>) * numItems);
    cudaMalloc(&(d_keys.d_buffers[0]), sizeof(key_t) * numItems);
    cudaMalloc(&(d_keys.d_buffers[1]), sizeof(key_t) * numItems);

    cub::DeviceRadixSort::SortPairsDescending(nullptr, tempStorageBytes, d_keys, d_records, numItems, 0, sizeof(key_t) * 8);
    cudaMalloc(&d_tempStorage, tempStorageBytes);
}

template<typename elem_t, int DIMS, typename key_t, typename data_t>
void freeSortRecordsByKey(cub::DoubleBuffer<record_t<elem_t, DIMS, key_t, data_t>>& d_records, cub::DoubleBuffer<key_t>& d_keys, void* d_tempStorage) {
    cudaFree(d_records.Alternate());
    cudaFree(d_keys.Current());
    cudaFree(d_keys.Alternate());
    cudaFree(d_tempStorage);
}

template<typename elem_t, int DIMS, typename key_t, typename data_t>
void sortRecordsByKeyDescending(cub::DoubleBuffer<record_t<elem_t, DIMS, key_t, data_t>>& d_records, cub::DoubleBuffer<key_t>& d_keys, void* d_tempStorage, size_t& tempStorageBytes, int numItems, cudaStream_t stream) {
    cub::DeviceRadixSort::SortPairsDescending(d_tempStorage, tempStorageBytes, d_keys, d_records, numItems, 0, sizeof(key_t) * 8, stream);
}

template<typename key_t, typename value_t>
struct key_value_pair_t {
    key_t key;
    value_t value;

    key_value_pair_t() = default;
    key_value_pair_t(key_t key, value_t value) : key(key), value(value) {}
};

struct kv_key_t {
	template<typename key_t, typename value_t>
    __host__ __device__ key_t& operator()(key_value_pair_t<key_t, value_t>& kvp) const {
        return kvp.key;
    }

	template<typename key_t, typename value_t>
    __host__ __device__ const key_t& operator()(const key_value_pair_t<key_t, value_t>& kvp) const {
        return kvp.key;
    }
};

struct sum_t {
    template<typename item_t>
    __host__ __device__ int operator()(const item_t& lhs, const item_t& rhs) const {
        return lhs + rhs;
    }
};

struct alignas(8) kv_le_t {
    template<typename key_t, typename value_t>
    __host__ __device__ bool operator()(const key_value_pair_t<key_t, value_t>& lhs, const key_value_pair_t<key_t, value_t>& rhs) const {
        return lhs.key < rhs.key;
    }
};

struct ge_t {
    template<typename item_t>
    __host__ __device__ bool operator()(const item_t& lhs, const item_t& rhs) const {
        return lhs > rhs;
    }
};

struct max_t {
    template<typename item_t>
    __host__ __device__ item_t operator()(const item_t& lhs, const item_t& rhs) const {
        return max(lhs, rhs);
    }
};

struct bb_intersect_t {
    template<typename elem_t, int DIMS>
    __host__ __device__ bool operator()(const bounding_box_t<elem_t, DIMS>& lhs, const bounding_box_t<elem_t, DIMS>& rhs) const {
		bool result = true;
		#pragma unroll
		for (int dim = 0; dim < DIMS; ++dim) {
			result &= lhs.min[dim] <= rhs.max[dim] && rhs.min[dim] <= lhs.max[dim];
		}
		return result;
    }
};

template<typename elem_t, int DIMS, typename key_t, typename data_t, typename key_cmp_t, typename spatial_pred_t, typename score_func_t, typename score_cmp_t>
void referenceImpl(
    record_t<elem_t, DIMS, key_t, data_t>* inputA,
    record_t<elem_t, DIMS, key_t, data_t>* inputB,
    int inputSizeA,
    int inputSizeB,
    int outputCapacity,
    key_t scoreThreshold,
    key_cmp_t keyCmp,
    spatial_pred_t spatialPred,
    score_func_t scoreFunc,
    score_cmp_t scoreCmp,
    score_data_pair_t<key_t, tuple_t<data_t>>* outputBuf,
    size_t& outputSize
) {
    range_t<record_t<elem_t, DIMS, key_t, data_t>*> rangeA { inputA, inputA + inputSizeA };
    range_t<record_t<elem_t, DIMS, key_t, data_t>*> rangeB { inputB, inputB + inputSizeB };
    std::array<bnb::bnb_input_t<elem_t, DIMS, key_t, data_t, key_cmp_t>, 2> bnbInputs {
        bnb::bnb_input_t<elem_t, DIMS, key_t, data_t, key_cmp_t>(rangeA, keyCmp),
        bnb::bnb_input_t<elem_t, DIMS, key_t, data_t, key_cmp_t>(rangeB, keyCmp)
    };
    output::accumulator_config_t<score_cmp_t> accConfig(outputCapacity, scoreCmp);
    output::accumulator_data_t<key_t, tuple_t<data_t>> accData(outputBuf, 0, scoreThreshold);

    bnb::runBnb<elem_t, DIMS, key_t, data_t, key_cmp_t, spatial_pred_t, score_func_t, score_cmp_t>(bnbInputs, spatialPred, scoreFunc, accConfig, accData);

    output::deheapify(accData.heap, accData.size, scoreCmp);
    outputSize = accData.size;
}

struct bba_result_t {
    bool success;
    float duration;
};

template<int LOG_FANOUT, int BLOCK_PARALLELISM, int OUTPUT_LIMIT, int QUEUE_LIMIT, typename elem_t, int DIMS, typename key_t, typename data_t, typename key_agg_t, typename spatial_pred_t, typename score_func_t, typename score_cmp_t>
bba_result_t gpuImpl(
    const std::array<record_t<elem_t, DIMS, key_t, data_t>*, 2>& h_records,
    const std::array<int, 2>& inputSizes,
    int outputCapacity,
    int logBlockSize,
    key_t scoreThreshold,
    key_agg_t keyAgg,
    spatial_pred_t spatialPred,
    score_func_t scoreFunc,
    score_cmp_t scoreCmp,
    score_data_pair_t<key_t, tuple_t<data_t>>* h_outputBuf,
    size_t& outputSize
) {
    cudaEvent_t start, bbaStart, bbaStop;
    cudaEventCreate(&start);
    cudaEventCreate(&bbaStart);
    cudaEventCreate(&bbaStop);

    // TODO: What is a good allocation size?
    int allocationSize = 0;
    for (int i = 0; i < 2; ++i) {
        allocationSize += inputSizes[i] * 2 * sizeof(record_t<elem_t, DIMS, key_t, data_t>);
    }

    std::array<int, 2> numBlocks { inputSizes[0] >> logBlockSize, inputSizes[1] >> logBlockSize };

    bba::bba_result_t h_result;

    bba::bba_result_t* d_result;
    std::array<record_t<elem_t, DIMS, key_t, data_t>*, 2> d_records;
    std::array<rtree_t<elem_t, DIMS, key_t, data_t>*, 2> d_rtrees;
    score_data_pair_t<key_t, tuple_t<data_t>>* d_outputBuf;
    void* d_blockOutputBufs;
    size_t blockOutputBufsPitch;
    void* d_blockStatuses;
    size_t blockStatusesPitch;
    int* d_uninitializedJoinFlags;
    void* d_allocation;

    cudaMalloc(&d_result, sizeof(bba::bba_result_t));
    for (int i = 0; i < 2; ++i) {
		cudaMalloc(&d_rtrees[i], sizeof(rtree_t<elem_t, DIMS, key_t, data_t>) * numBlocks[i]);
    }
    cudaMalloc(&d_outputBuf, sizeof(score_data_pair_t<key_t, tuple_t<data_t>>) * outputCapacity);
    cudaMallocPitch(&d_blockOutputBufs, &blockOutputBufsPitch, sizeof(score_data_pair_t<key_t, tuple_t<data_t>>) * outputCapacity, BLOCK_PARALLELISM);
    cudaMallocPitch(&d_blockStatuses, &blockStatusesPitch, sizeof(bba::bba_result_t), BLOCK_PARALLELISM);
    cudaMalloc(&d_uninitializedJoinFlags, sizeof(int) * BLOCK_PARALLELISM * BLOCK_PARALLELISM);
    cudaMalloc(&d_allocation, allocationSize);

#ifndef _USE_CPU_SORT
    cudaEvent_t sortStart[2];
    cudaEvent_t sortStop[2];
	cub::DoubleBuffer<record_t<elem_t, DIMS, key_t, data_t>> d_recordsDb[2];
	cub::DoubleBuffer<key_t> d_keysDb[2];
	void* d_tempStorage[2];

	cudaEventRecord(start);

	for (int i = 0; i < 2; ++i) {
		cudaEventCreate(&sortStart[i]);
		cudaEventCreate(&sortStop[i]);

        // use streams to carry out the sorts asynchronously
        cudaStream_t stream;
        size_t tempStorageBytes;
        cudaStreamCreate(&stream);
		// cudaStreamCreateWithFlags(&stream, cudaStreamNonBlocking);
		allocateSortRecordsByKey<elem_t, DIMS, key_t, data_t>(inputSizes[i], d_recordsDb[i], d_keysDb[i], d_tempStorage[i], tempStorageBytes);

		cudaEventRecord(sortStart[i], stream);
		cudaMemcpyAsync(d_recordsDb[i].Current(), h_records[i], sizeof(record_t<elem_t, DIMS, key_t, data_t>) * inputSizes[i], cudaMemcpyHostToDevice, stream);
		bba::retrieveKeysKernel<elem_t, DIMS, key_t, data_t><<<inputSizes[i] / 32, 32, 0, stream>>>(d_recordsDb[i].Current(), d_keysDb[i].Current(), inputSizes[i]);
		sortRecordsByKeyDescending(d_recordsDb[i], d_keysDb[i], d_tempStorage[i], tempStorageBytes, inputSizes[i], stream);
        d_records[i] = d_recordsDb[i].Current();
		cudaEventRecord(sortStop[i], stream);

        // make the BBA stream wait for this sort to complete
		cudaStreamWaitEvent(0, sortStop[i], 0);

        // cudaStreamSynchronize(stream);
		cudaStreamDestroy(stream);
		cudaError_t err = cudaGetLastError();
        if (err != cudaSuccess) {
            printf("hmm\n");
        }
	}
#else
    // DEBUG: sort on the CPU to reduce compile times and to make it easier to debug
    {
        ge_t keyCmp;

        for (int i = 0; i < 2; ++i) {
            std::unique_ptr<record_t<elem_t, DIMS, key_t, data_t>[]> h_sortedRecords = std::make_unique<record_t<elem_t, DIMS, key_t, data_t>[]>(inputSizes[i]);
            std::copy(h_records[i], h_records[i] + inputSizes[i], h_sortedRecords.get());
            std::sort(h_sortedRecords.get(), h_sortedRecords.get() + inputSizes[i], [&keyCmp](const record_t<elem_t, DIMS, key_t, data_t>& lhs, const record_t<elem_t, DIMS, key_t, data_t>& rhs) {
                return keyCmp(lhs.key, rhs.key);
			});
            cudaMalloc(&d_records[i], sizeof(record_t<elem_t, DIMS, key_t, data_t>) * inputSizes[i]);
            cudaMemcpy(d_records[i], h_sortedRecords.get(), sizeof(record_t<elem_t, DIMS, key_t, data_t>) * inputSizes[i], cudaMemcpyHostToDevice);
        }
    }
#endif

    bba::bba_input_t<elem_t, DIMS, key_t, data_t> inputL;
    inputL.source = d_records[0];
    inputL.rtrees = d_rtrees[0];
    inputL.blockCount = numBlocks[0];

    bba::bba_input_t<elem_t, DIMS, key_t, data_t> inputR;
    inputR.source = d_records[1];
    inputR.rtrees = d_rtrees[1];
    inputR.blockCount = numBlocks[1];

    cudaEventRecord(bbaStart);
    bba::runBbaKernel<LOG_FANOUT, BLOCK_PARALLELISM, OUTPUT_LIMIT, QUEUE_LIMIT><<<1, 32>>>(
        d_result,
        inputL,
        inputR,
        logBlockSize,
        keyAgg,
        spatialPred,
        scoreFunc,
        scoreCmp,
        scoreThreshold,
        outputCapacity,
        d_outputBuf,
        d_blockOutputBufs,
        blockOutputBufsPitch,
        d_blockStatuses,
        blockStatusesPitch,
        d_uninitializedJoinFlags,
        d_allocation,
        allocationSize
	);

    cudaMemcpy(&h_result, d_result, sizeof(bba::bba_result_t), cudaMemcpyDeviceToHost);

    cudaError_t err = cudaGetLastError();
    if (err == cudaSuccess) {
		if (h_result.status == bba::bba_result_t::status_t::OK) {
			outputSize = h_result.outputSize;
			cudaMemcpy(h_outputBuf, d_outputBuf, sizeof(score_data_pair_t<key_t, tuple_t<data_t>>) * outputSize, cudaMemcpyDeviceToHost);
		}
    }
    else {
        h_result.status = bba::bba_result_t::status_t::CUDA_ERROR;
    }

    cudaFree(d_result);
    for (int i = 0; i < 2; ++i) {
        cudaFree(d_records[i]);
        cudaFree(d_rtrees[i]);
    }
    cudaFree(d_outputBuf);
    cudaFree(d_blockOutputBufs);
    cudaFree(d_blockStatuses);
    cudaFree(d_uninitializedJoinFlags);
    cudaFree(d_allocation);

#ifndef _USE_CPU_SORT
	for (int i = 0; i < 2; ++i) {
		d_records[i] = d_recordsDb[i].Current();
		freeSortRecordsByKey(d_recordsDb[i], d_keysDb[i], d_tempStorage[i]);
	}
#endif

    cudaEventRecord(bbaStop);
    cudaEventSynchronize(bbaStop);

    bba_result_t result;
    result.success = h_result.status == bba::bba_result_t::status_t::OK;
    cudaEventElapsedTime(&result.duration, start, bbaStop);

    cudaEventDestroy(start);
    cudaEventDestroy(bbaStart);
    cudaEventDestroy(bbaStop);
#ifndef _USE_CPU_SORT
    for (int i = 0; i < 2; ++i) {
        cudaEventDestroy(sortStart[i]);
        cudaEventDestroy(sortStop[i]);
    }
#endif

    return result;
}

template<typename elem_t, int DIMS, typename key_t, typename data_t, typename key_cmp_t, typename key_agg_t>
struct ranked_join_input_t {
    range_t<record_t<elem_t, DIMS, key_t, data_t>*> records;
    key_cmp_t keyCmp;
    key_agg_t keyAgg;

    ranked_join_input_t(range_t<record_t<elem_t, DIMS, key_t, data_t>*> records, key_cmp_t keyCmp, key_agg_t keyAgg)
        : records(records)
        , keyCmp(keyCmp)
        , keyAgg(keyAgg) {}
};

template<typename elem_t, int DIMS, typename key_t, typename data_t, typename key_agg_t, typename score_func_t, typename score_cmp_t, typename spatial_pred_t>
struct top_k_spatial_join_t {
    std::array<record_t<elem_t, DIMS, key_t, data_t>*, 2> records;
    std::array<int, 2> inputSizes;
    key_t minScore;
    key_agg_t keyAgg;
    score_func_t scoreFunc;
    spatial_pred_t spatialPred;
    score_cmp_t scoreCmp;
    int capacity;

    top_k_spatial_join_t(
		std::array<record_t<elem_t, DIMS, key_t, data_t>*, 2> records,
		std::array<int, 2> inputSizes,
        key_t minScore,
        key_agg_t keyAgg,
        score_func_t scoreFunc,
        spatial_pred_t spatialPred,
        score_cmp_t scoreCmp,
        int capacity
    )
        : records(records)
        , inputSizes(inputSizes)
        , minScore(minScore)
        , keyAgg(keyAgg)
        , scoreFunc(scoreFunc)
        , spatialPred(spatialPred)
        , scoreCmp(scoreCmp)
		, capacity(capacity) {}
};

struct experiment_result_t {
    bool success;
    float avgCudaDuration;
    std::chrono::high_resolution_clock::duration avgCpuDuration;
    std::string name;
};

template<typename elem_t, int DIMS, typename key_t, typename data_t, typename key_cmp_t, typename key_agg_t, typename score_func_t, typename spatial_pred_t, typename task_func_t>
experiment_result_t runExperiment(const top_k_spatial_join_t<elem_t, DIMS, key_t, data_t, key_cmp_t, key_agg_t, score_func_t, spatial_pred_t>& join, task_func_t taskFunc, std::string name) {
    {
		score_data_pair_t<key_t, tuple_t<data_t>>* resultBuf = new score_data_pair_t<key_t, tuple_t<data_t>>[join.capacity];
        size_t outputSize;
		bba_result_t bbaResult = taskFunc(join, resultBuf, outputSize);
		delete[] resultBuf;
    }
    const int ITERATIONS = 5;

    bool success = true;
    float totalCudaDuration = 0.0F;
    std::chrono::high_resolution_clock::duration totalCpuDuration { 0 };
    for (int i = 0; i < ITERATIONS; ++i) {
		score_data_pair_t<key_t, tuple_t<data_t>>* resultBuf = new score_data_pair_t<key_t, tuple_t<data_t>>[join.capacity];
        size_t outputSize;

        auto cpuStart = std::chrono::high_resolution_clock::now();
		bba_result_t bbaResult = taskFunc(join, resultBuf, outputSize);
        auto cpuEnd = std::chrono::high_resolution_clock::now();

		delete[] resultBuf;

        if (!bbaResult.success) {
            success = false;
        }
		totalCudaDuration += bbaResult.duration;
        totalCpuDuration += cpuEnd - cpuStart;
    }

    return { success, totalCudaDuration / ITERATIONS, totalCpuDuration / ITERATIONS, name };
}

template<typename score_t, typename data_t>
void printOutput(score_data_pair_t<score_t, tuple_t<data_t>>* output, size_t outputSize) {
	for (int i = 0; i < outputSize; ++i) {
		// printf("#%02d s: %07d, l: %07d, r: %07d\n", i + 1, output[i].score, output[i].data.left, output[i].data.right);
		printf("s: %07d, l: %07d, r: %07d\n", output[i].score, output[i].data.left, output[i].data.right);
	}
}

template<typename elem_t, int DIMS, typename key_t, typename data_t>
__device__ __host__ void printRecord(const record_t<elem_t, DIMS, key_t, data_t>& record) {
    printf("key = %d, data = %d\n", record.key, record.data);
    for (int dim = 0; dim < DIMS; ++dim) {
        printf("DIM %d [%f, %f]\n", dim, record.mbb.min[dim], record.mbb.max[dim]);
    }
}

template<typename elem_t, int DIMS, typename key_t, typename data_t>
__device__ __host__ void printRecordSmall(int index, const record_t<elem_t, DIMS, key_t, data_t>& record) {
	printf("#%03d %03d %03d", index, record.key, record.data);
	for (int dim = 0; dim < DIMS; ++dim) {
		printf(" [%f, %f]", record.mbb.min[dim], record.mbb.max[dim]);
	}
	printf("\n");
}


template<typename elem_t, int DIMS, typename key_t, typename data_t>
__device__ bool areRecordsEqual(const record_t<elem_t, DIMS, key_t, data_t>& lhs, const record_t<elem_t, DIMS, key_t, data_t>& rhs) {
    if (lhs.key != rhs.key) return false;
    if (lhs.data != rhs.data) return false;
    for (int dim = 0; dim < DIMS; ++dim) {
        if (lhs.mbb.min[dim] != rhs.mbb.min[dim]) return false;
        if (lhs.mbb.max[dim] != rhs.mbb.max[dim]) return false;
    }
    return true;
}

template<int LOG_FANOUT, typename elem_t, int DIMS, typename key_t, typename data_t, typename key_agg_t>
__global__ void sortTileRecursiveKernel(
    rtree_t<elem_t, DIMS, key_t, data_t>* output,
    record_t<elem_t, DIMS, key_t, data_t>* input,
    int logInputSize,
    key_agg_t keyAgg
) {
    const int FANOUT = 1 << LOG_FANOUT;

    __shared__ extern record_t<elem_t, DIMS, key_t, data_t> workspace[];

    cooperative_groups::thread_block parentGroup = cooperative_groups::this_thread_block();
    cooperative_groups::thread_block_tile<FANOUT> childGroup = cooperative_groups::tiled_partition<FANOUT>(parentGroup);

    /* if (parentGroup.thread_rank() == 0) {
        printf("STR starting!\n");
    } */
    
#ifndef _BBA_VECTORIZE_RECORD_ACCESS
	for (int idx = parentGroup.thread_rank(); idx < (1 << logInputSize); idx += parentGroup.size()) {
		workspace[idx] = input[idx];
	}
#else
	static_assert(sizeof(record_t<elem_t, DIMS, key_t, data_t>) % sizeof(int4) == 0, "record size must be divisble by int4 size");
	const int INT4S_PER_RECORD = sizeof(record_t<elem_t, DIMS, key_t, data_t>) / sizeof(int4);
	for (int i = parentGroup.thread_rank(); i < (INT4S_PER_RECORD << logInputSize); i += parentGroup.size()) {
		reinterpret_cast<int4*>(workspace)[i] = reinterpret_cast<const int4*>(input)[i];
	}
#endif
    parentGroup.sync();

    bba::sortTileRecursive<LOG_FANOUT>(*output, workspace, logInputSize, keyAgg, parentGroup, childGroup);

    // parentGroup.sync();
    /* if (parentGroup.thread_rank() == 0) {
        for (int i = 0; i < (1 << logInputSize); ++i) {
			record_t<elem_t, DIMS, key_t, data_t>& record = output->records[i];
            printRecordSmall(i, record);
        }
    } */

    /* if (parentGroup.thread_rank() == 0) {
        printf("Debugging...\n");
		bba::debugRTree<LOG_FANOUT>(*output, ge_t(), logInputSize);

        for (int i = 0; i < (1 << logInputSize); ++i) {
            const record_t<elem_t, DIMS, key_t, data_t>& inputRecord = input[i];
            int numEqual = 0;
            for (int j = 0; j < (1 << logInputSize); ++j) {
				const record_t<elem_t, DIMS, key_t, data_t>& outputRecord = output->records[j];
                if (areRecordsEqual(inputRecord, outputRecord)) {
                    ++numEqual;
                }
            }
            if (numEqual != 1) {
                printf("Expected exactly 1 equal record in the output, found %d\n", numEqual);
            }
        }
    } */

    /* if (parentGroup.thread_rank() == 0) {
        printf("STR complete!\n");
    } */
}

template<int LOG_FANOUT, int OUTPUT_LIMIT, int QUEUE_LIMIT, typename elem_t, int DIMS, typename key_t, typename data_t, typename spatial_pred_t, typename score_func_t, typename score_cmp_t>
__global__ void joinRTreesKernel(
    rtree_t<elem_t, DIMS, key_t, data_t>* _rTreeL,
    rtree_t<elem_t, DIMS, key_t, data_t>* _rTreeR,
    int logBlockSize,
    int outputCapacity,
    spatial_pred_t spatialPred,
    score_func_t scoreFunc,
    score_cmp_t scoreCmp,
    key_t scoreThreshold,
    score_data_pair_t<key_t, tuple_t<data_t>>* output,
    bba::bba_result_t* result
) {
    rtree_t<elem_t, DIMS, key_t, data_t> rTreeL = *_rTreeL;
    rtree_t<elem_t, DIMS, key_t, data_t> rTreeR = *_rTreeR;

    cooperative_groups::thread_block_tile<32> group = cooperative_groups::tiled_partition<32>(cooperative_groups::this_thread_block());

    /*if (group.thread_rank() == 0) {
        printf("Join starting!\n");
    }*/

	using key_shuffler_t = bba::default_shuffler_t<key_t>;
	using data_shuffler_t = bba::default_shuffler_t<data_t>;
	using output_heap_t = heap::heap_t<key_t, tuple_t<data_t>, bba::rev_cmp_t<score_cmp_t>, bba::default_shuffler_t<key_t>, bba::tuple_shuffler_t<data_t, data_shuffler_t>>;

	key_shuffler_t keyShuffler;
	data_shuffler_t dataShuffler;

	__shared__ key_t s_outputHeapMinKeys[OUTPUT_LIMIT];
	__shared__ typename output_heap_t::shared_state_t outputHeapSharedState;
	typename output_heap_t::config_t outputHeapConfig(bba::rev_cmp_t<score_cmp_t>(scoreCmp), keyShuffler, bba::tuple_shuffler_t<data_t, data_shuffler_t>(dataShuffler));
	heap::node_t<key_t, tuple_t<data_t>> outputHeapNodes[OUTPUT_LIMIT];
	output_heap_t outputHeap(outputHeapConfig, outputHeapSharedState, s_outputHeapMinKeys, outputHeapNodes);

    bool success = bba::joinRTrees<LOG_FANOUT, QUEUE_LIMIT>(rTreeL, rTreeR, logBlockSize, outputCapacity, spatialPred, scoreFunc, scoreCmp, keyShuffler, dataShuffler, scoreThreshold, outputHeap, group);

    if (success) {
		if (group.thread_rank() == 0) {
            *result = { bba::bba_result_t::status_t::OK, outputHeap.size() };
		}
		bba::deconstructHeap(outputHeap, output);
    }
    else {
		if (group.thread_rank() == 0) {
            *result = { bba::bba_result_t::status_t::QUEUE_OVERFLOW, 0 };
		}
    }

    if (group.thread_rank() == 0) {
        printf("Join complete!\n");
    }
}

template<typename score_t, typename data_t>
bool compareOutputs(size_t expectedSize, const score_data_pair_t<score_t, tuple_t<data_t>>* expectedOutput, size_t actualSize, const score_data_pair_t<score_t, tuple_t<data_t>>* actualOutput) {
    if (expectedSize != actualSize) return false;
    for (size_t i = 0; i < expectedSize; ++i) {
        const score_data_pair_t<score_t, tuple_t<data_t>>& expected = expectedOutput[i];
        const score_data_pair_t<score_t, tuple_t<data_t>>& actual = actualOutput[i];

        if (expected.score != actual.score) return false;
        if (expected.data.left != actual.data.left) return false;
        if (expected.data.right != actual.data.right) return false;
    }
    return true;
}

template<typename elem_t, int DIMS>
__device__ elem_t euclideanDistanceSqr(const bounding_box_t<elem_t, DIMS>& lhs, const bounding_box_t<elem_t, DIMS>& rhs) {
    elem_t totalDistanceSqr = 0;
    for (int dim = 0; dim < DIMS; ++dim) {
        // if either distance is positive, we know they do not intersect on this axis and thus have a distance to contribute
        // first check if lhs is greater than rhs
        elem_t axisDistance = lhs.min[dim] - rhs.max[dim];
        if (axisDistance > 0) {
            totalDistanceSqr += axisDistance * axisDistance;
            continue;
        }
        // then check if rhs is greater than lhs
        axisDistance = rhs.min[dim] - lhs.max[dim];
        if (axisDistance > 0) {
            totalDistanceSqr += axisDistance * axisDistance;
        }
    }
    return totalDistanceSqr;
}

template<typename elem_t>
struct euclidean_distance_test_t {
    elem_t thresholdSqr;

    euclidean_distance_test_t(elem_t threshold)
        : thresholdSqr(threshold * threshold) {}

    template<int DIMS>
    __device__ bool operator()(const bounding_box_t<elem_t, DIMS>& lhs, const bounding_box_t<elem_t, DIMS>& rhs) const {
        return euclideanDistanceSqr(lhs, rhs) <= this->thresholdSqr;
    }
};

template<int DIMS, typename key_distribution_t, typename record_iter_t>
void generateRandomLeafRecords(record_iter_t first, record_iter_t last, std::mt19937& random, std::uniform_real_distribution<float> elemDistribution, key_distribution_t keyDistribution) {
    int data = 0;
    for (record_iter_t i = first; i < last; ++i, ++data) {
        for (int dim = 0; dim < DIMS; ++dim) {
            float coord = elemDistribution(random);
            i->mbb.min[dim] = coord;
            i->mbb.max[dim] = coord;
        }
        i->key = keyDistribution(random);
        i->data = data;
    }
}

template<int LOG_FANOUT, int BLOCK_PARALLELISM, int OUTPUT_LIMIT, int QUEUE_LIMIT>
struct bba_experiment_t {
    size_t logBlockSize;

    bba_experiment_t(size_t logBlockSize)
        : logBlockSize(logBlockSize) {}

	template<typename elem_t, int DIMS, typename key_t, typename data_t, typename key_agg_t, typename score_func_t, typename score_cmp_t, typename spatial_pred_t>
    bba_result_t operator()(const top_k_spatial_join_t<elem_t, DIMS, key_t, data_t, key_agg_t, score_func_t, score_cmp_t, spatial_pred_t>& join, score_data_pair_t<key_t, tuple_t<data_t>>* outputBuf, size_t& outputSize) const {
        return gpuImpl<LOG_FANOUT, BLOCK_PARALLELISM, OUTPUT_LIMIT, QUEUE_LIMIT, elem_t, DIMS, key_t, data_t, key_agg_t, spatial_pred_t, score_func_t, score_cmp_t>(join.records, join.inputSizes, join.capacity, logBlockSize, join.minScore, join.keyAgg, join.spatialPred, join.scoreFunc, join.scoreCmp, outputBuf, outputSize);
    }
};

int main() {
    // If l > r, l has a better score than r
    const auto keyAgg = max_t();
    const auto scoreFunc = sum_t();
    const auto scoreCmp = ge_t();
    const auto spatialPred = euclidean_distance_test_t<float>(1.0f);

    const int DIMS = 2;
    using key_agg_t = decltype(keyAgg);
    using score_func_t = decltype(scoreFunc);
    using score_cmp_t = decltype(scoreCmp);
    using spatial_pred_t = decltype(spatialPred);
    using elem_t = float;
    using key_t = int;
    using data_t = int;

	// const int INPUT_SIZES[] = { 1024 * 1024, 32 * 1024, 1024 };
    const int INPUT_SIZES[] = { 1024 * 1024 };
	// const int OUTPUT_SIZES[] = { 10000, 1000, 100 };
    const int OUTPUT_SIZES[] = { 10000 };
	// const int OUTPUT_SIZES[] = { 1000 };
	// const int LOG_BLOCK_SIZES[] = { 7, 8, 9, 10 };
	const int LOG_BLOCK_SIZES[] = { 10 };

	const float ELEM_DISTRIBUTION_SIZE = 1000.0f;
	const int KEY_DISTRIBUTION_SIZE = 100;
	std::mt19937 random(0);

	std::unique_ptr<record_t<elem_t, DIMS, key_t, data_t>[]> inputArrays[2];
	for (int i = 0; i < 2; ++i) {
		inputArrays[i] = std::make_unique<record_t<elem_t, DIMS, key_t, data_t>[]>(INPUT_SIZES[0]);
		std::uniform_real_distribution<elem_t> elemDistribution(0, ELEM_DISTRIBUTION_SIZE);
		std::uniform_int_distribution<key_t> keyDistribution(0, KEY_DISTRIBUTION_SIZE);
		generateRandomLeafRecords<DIMS>(inputArrays[i].get(), inputArrays[i].get() + INPUT_SIZES[0], random, elemDistribution, keyDistribution);
	}

    // const int BLOCK_PARALLELISM = 8;
    const int OUTPUT_LIMIT = (10000 + 31) / 32;
    const int QUEUE_LIMIT = 32;

    for (int inputSizeIdxL = 0; inputSizeIdxL < sizeof(INPUT_SIZES) / sizeof(int); ++inputSizeIdxL) {
        const int inputSizeL = INPUT_SIZES[inputSizeIdxL];
        for (int inputSizeIdxR = inputSizeIdxL; inputSizeIdxR < sizeof(INPUT_SIZES) / sizeof(int); ++inputSizeIdxR) {
            const int inputSizeR = INPUT_SIZES[inputSizeIdxR];
            for (int outputSize : OUTPUT_SIZES) {
                printf("Input %d x %d -> Output %d:\n", inputSizeL, inputSizeR, outputSize);

				std::vector<experiment_result_t> results;

				top_k_spatial_join_t<elem_t, DIMS, key_t, data_t, key_agg_t, score_func_t, score_cmp_t, spatial_pred_t> joinParameters(
					{ inputArrays[0].get(), inputArrays[1].get() },
                    { inputSizeL, inputSizeR },
					std::numeric_limits<int>::min(),
                    keyAgg,
					scoreFunc,
					spatialPred,
                    scoreCmp,
                    outputSize
				);

                for (int logBlockSize : LOG_BLOCK_SIZES) {
                    if ((logBlockSize << 1) <= inputSizeR) {
#define experiment(BLOCK_PARALLELISM) results.push_back(runExperiment(joinParameters, bba_experiment_t<3, BLOCK_PARALLELISM, OUTPUT_LIMIT, QUEUE_LIMIT>(logBlockSize), "bba fanout=2^3 blockSize=2^" + std::to_string(logBlockSize) + " blockParallelism=" + std::to_string(BLOCK_PARALLELISM)));

                        // experiment(1);
                        // experiment(2);
                        // experiment(3);
                        // experiment(4);
                        // experiment(5);
                        // experiment(6);
                        // experiment(7);
                        // experiment(8);
                        // experiment(9);
                        // experiment(10);
                        // experiment(11);
                        // experiment(12);
                        // experiment(13);
                        // experiment(14);
                        // experiment(15);
                        // experiment(16);
                        // experiment(17);
                        // experiment(18);
                        // experiment(19);
                        // experiment(20);
                        // experiment(21);
                        // experiment(22);
                        // experiment(23);
                        // experiment(24);
                        // experiment(25);
                        // experiment(26);
                        // experiment(27);
                        // experiment(28);
                        // experiment(29);
                        // experiment(30);
                        // experiment(31);
                        experiment(32);
                        /* results.push_back(runExperiment(
                            joinParameters, bba_experiment_t<3, 0, OUTPUT_LIMIT, QUEUE_LIMIT>(logBlockSize),
                            "bba fanout=2^3 blockSize=2^" + std::to_string(logBlockSize) + " blockParallelism=0"
                        )); */
                    }
                }

				/* std::sort(results.begin(), results.end(), [](const experiment_result_t& lhs, const experiment_result_t& rhs) {
					return lhs.avgCpuDuration < rhs.avgCpuDuration;
				}); */

				for (experiment_result_t& result : results) {
					printf("%lld ns: %s\n", std::chrono::duration_cast<std::chrono::nanoseconds>(result.avgCpuDuration).count(), result.name.c_str());
				}
				printf("\n");
            }
        }
    }

    /* const int DIMS = 2;
    const int LOG_FANOUT = 4;
    const int LOG_BLOCK_SIZE = 8;
    const int INPUT_SIZE = 1024 * 1024;
    const float DISTRIBUTION_SIZE = 1000.0f;
    const int OUTPUT_SIZE = 32;
	const int BLOCK_PARALLELISM = 20;
    const int QUEUE_LIMIT = 32;
    const int OUTPUT_LIMIT = (OUTPUT_SIZE + 31) / 32;

    using elem_t = float;
    using key_t = int;
    using data_t = int;
    using key_cmp_t = ge_t;
    using key_agg_t = max_t;
    using spatial_pred_t = bb_intersect_t;
    using score_func_t = sum_t;
    using score_cmp_t = ge_t;

    std::unique_ptr<record_t<elem_t, DIMS, key_t, data_t>[]> h_recordsA = std::make_unique<record_t<elem_t, DIMS, key_t, data_t>[]>(INPUT_SIZE);
    std::unique_ptr<record_t<elem_t, DIMS, key_t, data_t>[]> h_recordsB = std::make_unique<record_t<elem_t, DIMS, key_t, data_t>[]>(INPUT_SIZE);

    key_t scoreThreshold = std::numeric_limits<key_t>::min();
    key_cmp_t keyCmp;
    key_agg_t keyAgg;
    spatial_pred_t spatialPred;
    score_func_t scoreFunc;
    score_cmp_t scoreCmp;

    {
		std::mt19937 random(0);
		std::uniform_real_distribution<elem_t> elemDistribution(0, DISTRIBUTION_SIZE);
		std::uniform_int_distribution<key_t> keyDistribution(0, INPUT_SIZE);

		generateRandomLeafRecords<elem_t, DIMS>(h_recordsA.get(), h_recordsA.get() + INPUT_SIZE, random, elemDistribution, 4, keyDistribution);
		generateRandomLeafRecords<elem_t, DIMS>(h_recordsB.get(), h_recordsB.get() + INPUT_SIZE, random, elemDistribution, 4, keyDistribution);
    }

	std::unique_ptr<score_data_pair_t<key_t, tuple_t<data_t>>[]> referenceOutputBuf = std::make_unique<score_data_pair_t<key_t, tuple_t<data_t>>[]>(OUTPUT_SIZE);
	size_t referenceOutputSize;
	referenceImpl(h_recordsA.get(), h_recordsB.get(), INPUT_SIZE, INPUT_SIZE, OUTPUT_SIZE, scoreThreshold, keyCmp, spatialPred, scoreFunc, scoreCmp, referenceOutputBuf.get(), referenceOutputSize);

	printf("Reference output:\n");
	printOutput(referenceOutputBuf.get(), referenceOutputSize);
	printf("\n");

    std::unique_ptr<score_data_pair_t<key_t, tuple_t<data_t>>[]> gpuOutputBuf = std::make_unique<score_data_pair_t<key_t, tuple_t<data_t>>[]>(OUTPUT_SIZE);
    size_t gpuOutputSize;

    std::array<record_t<elem_t, DIMS, key_t, data_t>*, 2> records { h_recordsA.get(), h_recordsB.get() };
    std::array<int, 2> inputSizes { INPUT_SIZE, INPUT_SIZE };
    if (gpuImpl<LOG_FANOUT, BLOCK_PARALLELISM, OUTPUT_LIMIT, QUEUE_LIMIT>(records, inputSizes, OUTPUT_SIZE, LOG_BLOCK_SIZE, scoreThreshold, keyAgg, spatialPred, scoreFunc, scoreCmp, gpuOutputBuf.get(), gpuOutputSize)) {
		printf("BBA output:\n");
		printOutput(gpuOutputBuf.get(), gpuOutputSize);
		printf("\n");

        if (!compareOutputs(referenceOutputSize, referenceOutputBuf.get(), gpuOutputSize, gpuOutputBuf.get())) {
            printf("Invalid outputs!");
        }
        else {
            printf("Valid outputs!");
        }
    }
    else {
        printf("BBA failed!\n");
    } */

    /* {
		std::unique_ptr<score_data_pair_t<key_t, tuple_t<data_t>>[]> h_output = std::make_unique<score_data_pair_t<key_t, tuple_t<data_t>>[]>(OUTPUT_SIZE);

        record_t<elem_t, DIMS, key_t, data_t>* d_inputA;
        record_t<elem_t, DIMS, key_t, data_t>* d_inputB;
        record_t<elem_t, DIMS, key_t, data_t>* d_recordsA;
        record_t<elem_t, DIMS, key_t, data_t>* d_recordsB;
        rtree_t<elem_t, DIMS, key_t, data_t>* d_rtreeA;
        rtree_t<elem_t, DIMS, key_t, data_t>* d_rtreeB;
        score_data_pair_t<key_t, tuple_t<data_t>>* d_output;
        bba::bba_result_t* d_result;

        cudaMalloc(&d_inputA, sizeof(record_t<elem_t, DIMS, key_t, data_t>) << LOG_BLOCK_SIZE);
        cudaMalloc(&d_inputB, sizeof(record_t<elem_t, DIMS, key_t, data_t>) << LOG_BLOCK_SIZE);
        cudaMalloc(&d_recordsA, 2 * sizeof(record_t<elem_t, DIMS, key_t, data_t>) << LOG_BLOCK_SIZE);
        cudaMalloc(&d_recordsB, 2 * sizeof(record_t<elem_t, DIMS, key_t, data_t>) << LOG_BLOCK_SIZE);
        cudaMalloc(&d_rtreeA, sizeof(rtree_t<elem_t, DIMS, key_t, data_t>));
        cudaMalloc(&d_rtreeB, sizeof(rtree_t<elem_t, DIMS, key_t, data_t>));
        cudaMalloc(&d_output, sizeof(score_data_pair_t<key_t, tuple_t<data_t>>) * OUTPUT_SIZE);
        cudaMalloc(&d_result, sizeof(bba::bba_result_t));

        rtree_t<elem_t, DIMS, key_t, data_t> h_rtreeA;
        rtree_t<elem_t, DIMS, key_t, data_t> h_rtreeB;

        h_rtreeA.records = d_recordsA;
        h_rtreeB.records = d_recordsB;

        cudaMemcpy(d_inputA, h_recordsA.get(), sizeof(record_t<elem_t, DIMS, key_t, data_t>) << LOG_BLOCK_SIZE, cudaMemcpyHostToDevice);
        cudaMemcpy(d_inputB, h_recordsB.get(), sizeof(record_t<elem_t, DIMS, key_t, data_t>) << LOG_BLOCK_SIZE, cudaMemcpyHostToDevice);
        cudaMemcpy(d_rtreeA, &h_rtreeA, sizeof(rtree_t<elem_t, DIMS, key_t, data_t>), cudaMemcpyHostToDevice);
        cudaMemcpy(d_rtreeB, &h_rtreeB, sizeof(rtree_t<elem_t, DIMS, key_t, data_t>), cudaMemcpyHostToDevice);

        sortTileRecursiveKernel<LOG_FANOUT><<<1, 1 << LOG_BLOCK_SIZE, (sizeof(record_t<elem_t, DIMS, key_t, data_t>) << LOG_BLOCK_SIZE)>>>(d_rtreeA, d_inputA, LOG_BLOCK_SIZE, keyAgg);
        sortTileRecursiveKernel<LOG_FANOUT><<<1, 1 << LOG_BLOCK_SIZE, (sizeof(record_t<elem_t, DIMS, key_t, data_t>) << LOG_BLOCK_SIZE)>>>(d_rtreeB, d_inputB, LOG_BLOCK_SIZE, keyAgg);

        cudaError_t err = cudaGetLastError();
        if (err != cudaSuccess) {
            puts(cudaGetErrorName(err));
        }

        joinRTreesKernel<LOG_FANOUT, OUTPUT_LIMIT, QUEUE_LIMIT><<<1, 32>>>(d_rtreeA, d_rtreeB, LOG_BLOCK_SIZE, OUTPUT_SIZE, spatialPred, scoreFunc, scoreCmp, scoreThreshold, d_output, d_result);

        cudaDeviceSynchronize();

        err = cudaGetLastError();
        if (err != cudaSuccess) {
            puts(cudaGetErrorName(err));
        }

        bba::bba_result_t h_result;

        cudaMemcpy(&h_result, d_result, sizeof(bba::bba_result_t), cudaMemcpyDeviceToHost);

        if (h_result.status == bba::bba_result_t::status_t::OK) {
			cudaMemcpy(h_output.get(), d_output, sizeof(score_data_pair_t<key_t, tuple_t<data_t>>) * h_result.outputSize, cudaMemcpyDeviceToHost);
            printf("Test output:\n");
			printOutput(h_output.get(), h_result.outputSize);
        }
        else {
            printf("There was an error!\n");
        }

        cudaFree(d_inputA);
        cudaFree(d_inputB);
        cudaFree(d_rtreeA);
        cudaFree(d_rtreeB);
        cudaFree(d_rtreeA);
        cudaFree(d_rtreeB);
        cudaFree(d_output);
        cudaFree(d_result);
    } */

	return 0;
}
