#ifndef _OUTPUT_H
#define _OUTPUT_H

#include "shared.h"
#include <cassert>

namespace output {
	template<typename score_cmp_t, typename iter_t>
	void propagateUp(iter_t heap, int from, int size, score_cmp_t scoreCmp) {
		int i = from;
		while (i > 0) {
			int parent = (i - 1) >> 1;
			// we want the child element to be ranked higher than the parent element
			if (!scoreCmp(heap[i].score, heap[parent].score)) {
				// child is ranked lower than parent, swap and propagate from the parent
				std::swap(heap[i], heap[parent]);
				i = parent;
			}
			else {
				// child is ranked higher than parent, heap property is satisfied
				break;
			}
		}
	}

	template<typename score_cmp_t, typename iter_t>
	void propagateDown(iter_t heap, int from, int size, score_cmp_t scoreCmp) {
		int i = from;
		while (i < size) {
			int left = i * 2 + 1;
			int right = left + 1;
			if (left >= size) {
				// no children, terminate here
				break;
			}

			// find the minimum child, which is the child with the lowest rank, or the only child
			bool isLeftLessThanRight = right >= size || scoreCmp(heap[right].score, heap[left].score);
			int minChild = isLeftLessThanRight ? left : right;
			bool isParentLessThanMinChild = scoreCmp(heap[minChild].score, heap[i].score);

			if (!isParentLessThanMinChild) {
				// the parent is greater than the minimum child
				// swap the parent and the minimum child and continue from the minimum child
				std::swap(heap[i], heap[minChild]);
				i = minChild;
			}
			else {
				// child is ranked higher than parent, heap property is satisfied
				break;
			}
		}
	}

	template<typename score_t, typename data_t, typename score_cmp_t, typename iter_t>
	void insert(iter_t heap, const score_data_pair_t<score_t, data_t>& value, int capacity, int& size, score_cmp_t scoreCmp) {
		// is the heap being built, or is it at capacity?
		if (size < capacity) {
			// we're currently building the heap
			// grow and insert the value at end, then propagate up
			int end = size;
			heap[end] = value;
			++size;
			propagateUp(heap, end, size, scoreCmp);
		}
		else {
			// the heap is at capacity
			// replace the lowest ranked element in the heap and propagate down
			heap[0] = value;
			propagateDown(heap, 0, size, scoreCmp);
		}
	}

	template<typename score_t, typename data_t, typename score_cmp_t, typename iter_t>
	void insert(iter_t heap, const score_data_pair_t<score_t, data_t>& value, int capacity, int& size, score_t& lowerBound, score_cmp_t scoreCmp) {
		assert(!scoreCmp(lowerBound, value.score) /* score cannot be below lower bound */);

		// the score is greater than the current lower bound, so the value belongs somewhere in the heap
		insert(heap, value, capacity, size, scoreCmp);

		// have we found at least k values?
		if (size == capacity) {
			// all values in the heap must be greater than the lower bound
			// the minimum score value may be greater than the lower bound, so increase the lower bound
			lowerBound = heap[0].score;
		}
	}

	template<typename score_cmp_t, typename iter_t, typename source_iter_t>
	void insertAll(iter_t heap, source_iter_t first, source_iter_t last, int capacity, int& size, score_cmp_t scoreCmp) {
		for (source_iter_t i = first; i < last; ++i) {
			insert(heap, *i, capacity, size, scoreCmp);
		}
	}

	template<typename score_t, typename data_t, typename score_cmp_t, typename iter_t, typename source_iter_t>
	void insertAll(iter_t heap, source_iter_t first, source_iter_t last, int capacity, int& size, score_t& lowerBound, score_cmp_t scoreCmp) {
		for (source_iter_t i = first; i < last; ++i) {
			score_data_pair_t<score_t, data_t> value = *i;
			if (scoreCmp(value.score, lowerBound)) {
				insert(heap, value, capacity, size, scoreCmp);
			}
		}

		// have we found at least k values?
		if (size == capacity) {
			// all values in the heap must be greater than the lower bound
			// the minimum score value may be greater than the lower bound, so increase the lower bound
			lowerBound = heap[0].score;
		}
	}

	// Turns the heap into a sorted list in the opposite order of the heap
	template<typename iter_t, typename score_cmp_t>
	void deheapify(iter_t heap, int size, score_cmp_t scoreCmp) {
		for (int remainingItems = size - 1; remainingItems > 0; --remainingItems) {
			std::swap(heap[0], heap[remainingItems]);
			propagateDown(heap, 0, remainingItems, scoreCmp);
		}
	}

	// Deheapifies the min-heap into the output iterator
	template<typename score_cmp_t, typename input_iter_t, typename output_iter_t>
	void read(input_iter_t heap, int size, score_cmp_t scoreCmp, output_iter_t output) {
		std::copy_n(heap, size, output);
		deheapify(output, size, scoreCmp);
	}

	template<typename score_cmp_t>
	struct accumulator_config_t {
		int capacity;
		score_cmp_t scoreCmp;

		accumulator_config_t(int capacity, score_cmp_t scoreCmp)
			: capacity(capacity)
			, scoreCmp(scoreCmp) {}
	};

	template<typename score_t, typename data_t>
	struct accumulator_data_t {
		score_data_pair_t<score_t, data_t>* heap;
		int size;
		score_t lowerBound;

		accumulator_data_t(score_data_pair_t<score_t, data_t>* heap, int size, score_t lowerBound)
			: heap(heap)
			, size(size)
			, lowerBound(lowerBound) {}

		template<typename score_cmp_t>
		void insert(const score_data_pair_t<score_t, data_t>& value, const accumulator_config_t<score_cmp_t>& config) {
			output::insert(this->heap, value, config.capacity, this->size, this->lowerBound, config.scoreCmp);
		}

		template<typename score_cmp_t>
		void insertAll(const accumulator_data_t<score_t, data_t>& other, const accumulator_config_t<score_cmp_t>& config) {
			output::insertAll<score_t, data_t>(this->heap, other.heap, other.heap + other.size, config.capacity, this->size, this->lowerBound, config.scoreCmp);
		}

		template<typename iter_t, typename score_cmp_t>
		void read(iter_t output, const accumulator_config_t<score_cmp_t>& config) const {
			output::read(this->heap, this->size, config.scoreCmp, output);
		}
	};
}

// accumulator for a top-k result. stores its results in a binary min-heap
template<typename score_t, typename data_t, typename score_cmp_t>
struct top_k_accumulator_t {
	output::accumulator_config_t<score_cmp_t> config;
	output::accumulator_data_t<score_t, data_t> data;

	top_k_accumulator_t(score_cmp_t scoreCmp, int capacity, score_t lowerBound, score_data_pair_t<score_t, data_t>* heap)
		: config(capacity, scoreCmp)
		, data(heap, 0, lowerBound) {}

	bool isScoreBelowBound(score_t score) const {
		return !this->config.scoreCmp(score, this->data.lowerBound);
	}

	void insert(const score_data_pair_t<score_t, data_t>& value) {
		this->data.insert(value, this->config);
	}

	template<typename iter_t>
	int read(iter_t output) const {
		return output::read(this->data.heap, this->data.size, this->config.scoreCmp, output);
	}
};

#endif // _OUTPUT_H
