#ifndef STR_H
#define STR_H

#include <cuda_runtime.h>
#include "cub/block/block_radix_sort.cuh"
#include "cub/block/block_load.cuh"
#include "cub/block/block_store.cuh"
#include "cub/device/device_radix_sort.cuh"
#include "shared.h"

namespace str {
    template<int BLOCK_DIM_X, int ITEMS_PER_THREAD, typename elem_t, int DIMS, typename key_t, typename data_t>
    __global__ void sortRecordsByDim(record_t<elem_t, DIMS, key_t, data_t>* input, record_t<elem_t, DIMS, key_t, data_t>* output, int dim, int totalSize) {
        using block_load_t = cub::BlockLoad<record_t<elem_t, DIMS, key_t, data_t>, BLOCK_DIM_X, ITEMS_PER_THREAD>;
        using block_radix_sort_t = cub::BlockRadixSort<elem_t, BLOCK_DIM_X, ITEMS_PER_THREAD, record_t<elem_t, DIMS, key_t, data_t>>;
        using block_store_t = cub::BlockStore<record_t<elem_t, DIMS, key_t, data_t>, BLOCK_DIM_X, ITEMS_PER_THREAD>;

        union temp_storage_t {
            typename block_load_t::TempStorage blockLoad;
            typename block_radix_sort_t::TempStorage blockRadixSort;
            typename block_store_t::TempStorage blockStore;
        };

        __shared__ temp_storage_t tempStorage;

        int firstItemIdx = blockIdx.x * BLOCK_DIM_X * ITEMS_PER_THREAD;
        int lastItemIdx = min(firstItemIdx + BLOCK_DIM_X * ITEMS_PER_THREAD, totalSize);
        int numItems = lastItemIdx - firstItemIdx;

        elem_t keys[ITEMS_PER_THREAD];
        record_t<elem_t, DIMS, key_t, data_t> values[ITEMS_PER_THREAD];

        block_load_t blockLoad(tempStorage.blockLoad);
        blockLoad.Load(input + firstItemIdx, values, numItems);
        __syncthreads();

        for (int i = 0; i < ITEMS_PER_THREAD; ++i) {
            keys[i] = values[i].mbb.min[dim];
        }

        block_radix_sort_t blockRadixSort(tempStorage.blockRadixSort);
        blockRadixSort.Sort(keys, values);
        __syncthreads();

        block_store_t blockStore(tempStorage.blockStore);
        blockStore.Store(output, values, numItems);
    }

    template<int BLOCK_DIM_X, int ITEMS_PER_THREAD, typename elem_t, int DIMS, typename key_t, typename data_t>
    __global__ void getRecordKeys(record_t<elem_t, DIMS, key_t, data_t>* input, elem_t* output, int dim, int totalSize) {
        using block_load_t = cub::BlockLoad<record_t<elem_t, DIMS, key_t, data_t>, BLOCK_DIM_X, ITEMS_PER_THREAD>;
        using block_store_t = cub::BlockStore<record_t<elem_t, DIMS, key_t, data_t>, BLOCK_DIM_X, ITEMS_PER_THREAD>;

        union temp_storage_t {
            typename block_load_t::TempStorage blockLoad;
            typename block_store_t::TempStorage blockStore;
        };

        __shared__ temp_storage_t tempStorage;

        int firstItemIdx = blockIdx.x * BLOCK_DIM_X * ITEMS_PER_THREAD;
        int lastItemIdx = min(firstItemIdx + BLOCK_DIM_X * ITEMS_PER_THREAD, totalSize);
        int numItems = lastItemIdx - firstItemIdx;

        record_t<elem_t, DIMS, key_t, data_t> values[ITEMS_PER_THREAD];
        elem_t keys[ITEMS_PER_THREAD];

        block_load_t blockLoad(tempStorage.blockLoad);
        blockLoad.Load(input + firstItemIdx, values, numItems);
        __syncthreads();

        for (int i = 0; i < ITEMS_PER_THREAD; ++i) {
            keys[i] = values[i].mbb.min[dim];
        }

        block_store_t blockStore(tempStorage.blockStore);
        blockStore.Store(output, keys, numItems);
    }

	// Calculates a tile partitioning strategy for the sorting pass of a level of Sort-Tile-Recursive
	// In each level of Sort-Tile-Recursive, the nodes at the level are partitioned into spatially separate tiles of N nodes each
	// The partitioning into tiles is achieved by a series of segmented block partitioning operations, one for each dimension
    template<int DIMS>
	void calculateTilePartitionStrategy(int blockSizes[DIMS], int inputSize, int minBlockSize) {
		int blocksPerSegment = 1;
		int segmentSize = minBlockSize;
		while (segmentSize < inputSize) {
			++blocksPerSegment;
			segmentSize = minBlockSize;
			for (int dim = 0; dim < DIMS; ++dim) {
				segmentSize *= blocksPerSegment;
			}
		}

		for (int dim = 0; dim < DIMS; ++dim) {
			segmentSize /= blocksPerSegment;
			blockSizes[dim] = segmentSize;
		}
	}

    template<typename item_t>
    item_t* allocateArraySegment(void*& storage, int size) {
        item_t* result = static_cast<item_t*>(storage);
        storage = static_cast<void*>(static_cast<char*>(storage) + sizeof(item_t) * size);
        return result;
    }

    template<typename elem_t, int DIMS, typename key_t, typename data_t, int MAX_HEIGHT>
    struct str_t {
        struct temp_storage_allocations_t {
            record_t<elem_t, DIMS, key_t, data_t>* valuesAlt;
            elem_t* keys;
            elem_t* keysAlt;
            int* offsets;

            temp_storage_allocations_t(record_t<elem_t, DIMS, key_t, data_t>* valuesAlt, elem_t* keys, elem_t* keysAl, int* offsets)
                : valuesAlt(valuesAlt)
                , keys(keys)
                , keysAlt(keysAlt)
                , offsets(offsets) {}
        };

        rtree_layout_t<MAX_HEIGHT> layout;
        int blockSizes[MAX_HEIGHT][DIMS];

        str_t(rtree_layout_t<MAX_HEIGHT> layout)
			: layout(layout) {
            for (int level = 0; level < layout.height; ++level) {
                segment_t segment = layout.segments[level];
                calculateTilePartitionStrategy<DIMS>(blockSizes[level], segment.size, layout.fanout);
            }
        }

        temp_storage_allocations_t allocateTempStorage(void*& storage) const {
            segment_t bottomSegment = this->layout.segments[this->layout.height - 1];
            return temp_storage_allocations_t(
                allocateArraySegment<record_t<elem_t, DIMS, key_t, data_t>>(storage, bottomSegment.size),
                allocateArraySegment<elem_t>(storage, bottomSegment.size),
                allocateArraySegment<elem_t>(storage, bottomSegment.size),
				allocateArraySegment<int>(storage, 1 + (bottomSegment.size + this->layout.fanout - 1) / this->layout.fanout)
			);
        }

        size_t getTempStorageSize() const {
            void* storage = nullptr;
            this->allocateTempStorage(storage);
            return static_cast<char*>(storage) - (char*)nullptr;
        }

        void execute(record_t<elem_t, DIMS, key_t, data_t>* input, record_t<elem_t, DIMS, key_t, data_t>* output, void* tempStorage) {
            temp_storage_allocations_t tempStorageHandles = allocateTempStorage(tempStorage);

			const int BLOCK_SIZE = 1024;

			int dim = 0;
			int size = layout.segments[layout.height - 1].size;
			sortRecordsByDim<BLOCK_SIZE, 1><<<BLOCK_SIZE, ceilDiv(size, BLOCK_SIZE)>>>(input, output, dim, size);
        }
    };
}

#endif //STR_H
