#ifndef STR_SMALL_H
#define STR_SMALL_H

// TODO: Remove this, I guess...
#ifndef __CUDACC__
#define __CUDACC__
#endif

#include <cuda_runtime.h>
#include <cooperative_groups.h>
#include "shared.h"

namespace str_small {
	template<unsigned int NUM_VALUES, typename value_t, typename acc_func_t>
	__device__ void threadScanUpsweep(value_t values[NUM_VALUES], acc_func_t accFunc) {
		#pragma unroll
		for (unsigned int i = 2; i <= NUM_VALUES; i <<= 1) {
			for (unsigned int idx0 = i - 1; idx0 < NUM_VALUES; idx0 += i) {
				unsigned idx1 = idx0 - (i >> 1);
				values[idx0] = accFunc(values[idx0], values[idx1]);
			}
		}
	}

	template<unsigned int NUM_VALUES, typename value_t, typename acc_func_t>
	__device__ void threadScanDownsweep(value_t values[NUM_VALUES], acc_func_t accFunc) {
		#pragma unroll
		for (unsigned int i = NUM_VALUES; i > 1; i >>= 1) {
			for (unsigned int idx0 = i - 1; idx0 < NUM_VALUES; idx0 += i) {
				unsigned idx1 = idx0 - (i >> 1);
				value_t t = values[idx0];
				values[idx0] = accFunc(values[idx0], values[idx1]);
				values[idx1] = t;
			}
		}
	}

	template<unsigned int NUM_VALUES, typename value_t, typename acc_func_t>
	__device__ value_t threadPrefixScan(value_t input[NUM_VALUES], value_t output[NUM_VALUES], acc_func_t accFunc, value_t identity) {
		value_t acc = identity;
#pragma unroll
		for (int item = 0; item < NUM_VALUES; ++item) {
			output[item] = acc;
			acc = accFunc(acc, input[item]);
		}
		return acc;
	}

	template<typename value_t, typename acc_func_t, typename group_t>
	__device__ void workEfficientScanUpsweep(value_t* values, acc_func_t accFunc, unsigned int numValues, group_t group) {
		/* for (unsigned int i = 1; i < numValues; i <<= 1) {
			for (int workId = group.thread_rank(); workId < numValues / (i * 2); workId += group.size()) {
				unsigned idx0 = (workId + 1) * i * 2 - 1;
				unsigned idx1 = idx0 - i;
				values[idx0] = accFunc(values[idx0], values[idx1]);
			}
		} */

		for (unsigned int i = 2; i <= numValues; i <<= 1) {
			group.sync();
			for (unsigned int idx0 = i * (group.thread_rank() + 1) - 1; idx0 < numValues; idx0 += i * group.size()) {
				unsigned int idx1 = idx0 - (i >> 1);
				values[idx0] = accFunc(values[idx0], values[idx1]);
			}
		}
	}

	template<typename value_t, typename acc_func_t, typename group_t>
	__device__ void workEfficientScanDownsweep(value_t* values, acc_func_t accFunc, unsigned int numValues, group_t group) {
		/* for (unsigned int i = numValues; i > 1; i >>= 1) {
			group.sync();
			for (int workId = group.thread_rank(); workId < numValues / i; workId += group.size()) {
				unsigned int idx0 = (workId + 1) * i - 1;
				unsigned int idx1 = idx0 - i / 2;
				value_t t = values[idx0];
				values[idx0] = accFunc(values[idx0], values[idx1]);
				values[idx1] = t;
			}
		} */

		for (unsigned int i = numValues; i > 1; i >>= 1) {
			group.sync();
			for (int idx0 = i * (group.thread_rank() + 1) - 1; idx0 < numValues; idx0 += i * group.size()) {
				unsigned int idx1 = idx0 - (i >> 1);
				value_t t = values[idx0];
				values[idx0] = accFunc(values[idx0], values[idx1]);
				values[idx1] = t;
			}
		}
	}

	template<typename value_t, typename acc_func_t, typename group_t>
	__device__ void workEfficientPrefixScan(value_t* values, acc_func_t accFunc, value_t identity, unsigned int numValues, group_t group) {
		workEfficientScanUpsweep(values, accFunc, numValues, group);

		group.sync();
		// zero out the last upsweep value
		if (group.thread_rank() == 0) {
			values[numValues - 1] = identity;
		}

		workEfficientScanDownsweep(values, accFunc, numValues, group);
	}

	template<typename value_t, typename acc_func_t, typename group_t>
	__device__ void shuffleScanUpsweep(value_t& value, acc_func_t accFunc, group_t group) {
		for (unsigned int i = 1; i < group.size(); i <<= 1) {
			value_t n = group.shfl_up(value, i);
			unsigned int mask = (i << 1) - 1;
			if ((group.thread_rank() & mask) == mask) {
				value = accFunc(value, n);
			}
		}
	}

	template<typename value_t, typename acc_func_t, typename group_t>
	__device__ void shuffleScanDownsweep(value_t& value, acc_func_t accFunc, group_t group) {
		for (unsigned int i = group.size(); i > 0; i >>= 1) {
			unsigned int mask = (i << 1) - 1;
			value_t n = group.shfl_xor(value, mask);
			if ((group.thread_rank() & mask) == mask) {
				// the higher ranked thread accumulates the value of the lower ranked thread
				value = accFunc(value, n);
			}
			else if ((group.thread_rank() & mask) == mask >> 1) {
				// the lower ranked thread inherits the value of the higher ranked thread
				value = n;
			}
		}
	}

	template<typename value_t, typename acc_func_t, typename group_t>
	__device__ void shufflePrefixScan(const value_t input, value_t& output, acc_func_t accFunc, value_t identity, group_t group) {
		// shuffle all the items up and zero the first one
		output = group.shfl_up(input, 1);
		if (group.thread_rank() == 0) {
			output = identity;
		}

		for (unsigned int i = 1; i < group.size(); i <<= 1) {
			value_t n = group.shfl_up(output, i);
			if (group.thread_rank() >= i) {
				output = accFunc(output, n);
			}
		}
	}

	template<unsigned int VALUES_PER_THREAD, typename value_t, typename acc_func_t, typename group_t>
	__device__ void shuffleScanBlockedUpsweep(value_t values[VALUES_PER_THREAD], acc_func_t accFunc, group_t group) {
		threadScanUpsweep(values, accFunc);
		shuffleScanUpsweep(values[VALUES_PER_THREAD - 1], accFunc, group);
	}

	template<unsigned int VALUES_PER_THREAD, typename value_t, typename acc_func_t, typename group_t>
	__device__ void shuffleScanBlockedDownsweep(value_t values[VALUES_PER_THREAD], acc_func_t accFunc, group_t group) {
		shuffleScanDownsweep(values[VALUES_PER_THREAD - 1], accFunc, group);
		threadScanDownsweep(values, accFunc);
	}

	template<unsigned int VALUES_PER_THREAD, typename value_t, typename acc_func_t, typename group_t>
	__device__ void shufflePrefixScanBlocked(const value_t input[VALUES_PER_THREAD], value_t output[VALUES_PER_THREAD], acc_func_t accFunc, value_t identity, group_t group) {
		// perform a prefix scan on the thread's own values
		for (unsigned int i = 0; i < VALUES_PER_THREAD; ++i) {
			output[i] = input[i];
		}

		shuffleScanBlockedUpsweep(output, accFunc, group);

		if (group.thread_rank() == group.size()) {
			output[VALUES_PER_THREAD - 1] = identity;
		}

		shuffleScanBlockedDownsweep(output, accFunc, group);

		/* value_t threadSum = threadPrefixScan(input, output, accFunc, identity);

		// perform a prefix scan on the sums of each thread
		value_t threadPrefixSum;
		shufflePrefixScan(threadSum, threadPrefixSum, accFunc, identity, group);

		// add the thread's prefix sum to each value
		#pragma unroll
		for (unsigned int item = 0; item < VALUES_PER_THREAD; ++item) {
			output[item] = accFunc(threadPrefixSum, output[item]);
		} */
	}

	/* template<unsigned int VALUES_PER_THREAD, typename value_t, typename acc_func_t, typename group_t>
	__device__ void combinedPrefixScanStriped(const value_t input[VALUES_PER_THREAD], value_t output[VALUES_PER_THREAD], value_t* tileSums, acc_func_t accFunc, value_t identity, unsigned numValues, group_t group) {
		cooperative_groups::thread_block_tile<32> tile = cooperative_groups::tiled_partition<32>(group);

		for (unsigned int i = 0; i < VALUES_PER_THREAD; ++i) {
			// perform a shuffle prefix scan for each tile
			shufflePrefixScan(input[i], output[i], accFunc, identity, tile);

			// collect the sums of each tile
			// the last thread in each tile writes the sum
			if (tile.thread_rank() == tile.size() - 1) {
				tileSums[group.thread_rank() >> 5] = accFunc(output[i], input[i]);
			}

			// perform a prefix scan on the tile sums
			workEfficientPrefixScan(tileSums, accFunc, identity, numValues >> 5, group);
			group.sync();

			// add the tile's prefix scan sum to the value
			output[i] = accFunc(tileSums[group.thread_rank() >> 5], output[i]);
		}
	} */

	template<unsigned int VALUES_PER_THREAD, typename value_t, typename acc_func_t, typename group_t>
	__device__ void combinedScanBlockedUpsweep(value_t values[VALUES_PER_THREAD], value_t* tilePrefixes, acc_func_t accFunc, unsigned int numValues, group_t group) {
		cooperative_groups::thread_block_tile<32> tile = cooperative_groups::tiled_partition<32>(group);
		shuffleScanBlockedUpsweep(values, accFunc, tile);

		// collect the sums of each tile
		// the last thread in each tile writes the sum
		if (tile.thread_rank() == tile.size() - 1) {
			tilePrefixes[group.thread_rank() >> 5] = values[VALUES_PER_THREAD - 1];
		}

		workEfficientScanUpsweep(tilePrefixes, accFunc, numValues, group);
	}

	template<unsigned int VALUES_PER_THREAD, typename value_t, typename acc_func_t, typename group_t>
	__device__ void combinedScanBlockedDownsweep(value_t values[VALUES_PER_THREAD], value_t* tilePrefixes, acc_func_t accFunc, unsigned int numValues, group_t group) {
		workEfficientScanDownsweep(tilePrefixes, accFunc, numValues, group);

		cooperative_groups::thread_block_tile<32> tile = cooperative_groups::tiled_partition<32>(group);
		// collect the sums of each tile
		// the last thread in each tile reads the sum
		if (tile.thread_rank() == tile.size() - 1) {
			values[VALUES_PER_THREAD - 1] = tilePrefixes[group.thread_rank() >> 5];
		}

		shuffleScanBlockedDownsweep(values, accFunc, tile);
	}

	template<unsigned int VALUES_PER_THREAD, typename value_t, typename acc_func_t, typename group_t>
	__device__ void combinedPrefixScanBlocked(const value_t input[VALUES_PER_THREAD], value_t output[VALUES_PER_THREAD], value_t* tilePrefixes, acc_func_t accFunc, value_t identity, unsigned int numValues, group_t group) {
		for (unsigned int i = 0; i < VALUES_PER_THREAD; ++i) {
			output[i] = input[i];
		}
		combinedScanBlockedUpsweep(output, tilePrefixes, accFunc, numValues, group);

		if (group.thread_rank() == 0) {
			tilePrefixes[numValues >> 5 - 1] = identity;
		}

		combinedScanBlockedDownsweep(output, tilePrefixes, accFunc, numValues, group);

		/* cooperative_groups::thread_block_tile<32> tile = cooperative_groups::tiled_partition<32>(group);
		shufflePrefixScanBlocked(input, output, accFunc, identity, tile);

		// collect the sums of each tile
		// the last thread in each tile writes the sum
		if (tile.thread_rank() == tile.size() - 1) {
			tilePrefixes[group.thread_rank() >> 5] = accFunc(output[VALUES_PER_THREAD - 1], input[VALUES_PER_THREAD - 1]);
		}

		// perform a prefix scan on the tile sums
		workEfficientPrefixScan(tilePrefixes, accFunc, identity, numValues >> 5, group);

		group.sync();

		// add the tile's prefix scan sum to each value
		for (unsigned int i = 0; i < VALUES_PER_THREAD; ++i) {
			output[i] = accFunc(tilePrefixes[group.thread_rank() >> 5], output[i]);
		} */
	}

	template<unsigned int SEGMENT_WIDTH, typename value_t, typename acc_func_t>
	__global__ void runSegmentedPrefixScanFixed(value_t* values, acc_func_t accFunc, value_t identity) {
		cooperative_groups::thread_block wholeBlock = cooperative_groups::this_thread_block();
		cooperative_groups::thread_block_tile<SEGMENT_WIDTH> tile = cooperative_groups::tiled_partition<SEGMENT_WIDTH>(wholeBlock);

		int tid = wholeBlock.thread_rank() + blockIdx.x * wholeBlock.size();
		value_t input[1] = { values[tid] };
		value_t output[1];

		shufflePrefixScanBlocked(input, output, accFunc, identity, tile);

		values[tid] = output[0];
	}

	template<typename value_t, typename acc_func_t>
	__global__ void runSegmentedPrefixScanShared(value_t* values, acc_func_t accFunc, value_t identity, int segmentWidth) {
		__shared__ extern value_t tileSums[];

		cooperative_groups::thread_block wholeBlock = cooperative_groups::this_thread_block();

		int tid = wholeBlock.thread_rank() + blockIdx.x * wholeBlock.size();
		value_t input[1] = { values[tid] };
		value_t output[1];

		combinedPrefixScanBlocked(input, output, tileSums, accFunc, identity, segmentWidth, wholeBlock);

		values[tid] = output[0];
	}

	template<typename item_t>
	struct double_buffer_t {
		item_t* buffers[2];
		int selector;

		double_buffer_t(item_t* buffer1, item_t* buffer2) {
			buffers[0] = buffer1;
			buffers[1] = buffer2;
			selector = 0;
		}

		item_t* current() {
			return buffers[selector];
		}

		void swap() {
			selector ^= 1;
		}
	};

	template<typename key_t, typename value_t, typename cmp_t, typename group_t>
	__device__ void shuffleCmpSwap(key_t& key, value_t& value, unsigned tradeLaneId, cmp_t cmp, group_t group) {
		key_t altKey = group.shfl(key, tradeLaneId);
		value_t altValue = group.shfl(value, tradeLaneId);
		if (group.thread_rank() < tradeLaneId ? !cmp(key, altKey) : !cmp(altKey, key)) {
			key = altKey;
			value = altValue;
		}
	}

	template<typename key_t, typename value_t, typename cmp_t, unsigned int TILE_SIZE>
	__device__ void shuffleSortBitonicSequence(key_t& key, value_t& value, cmp_t cmp, cooperative_groups::thread_block_tile<TILE_SIZE> tile) {
		int tradeLaneId = tile.thread_rank() ^ (TILE_SIZE / 2);
		shuffleCmpSwap(key, value, tradeLaneId, cmp, tile);
		shuffleSortBitonicSequence(key, value, cmp, cooperative_groups::tiled_partition<TILE_SIZE / 2>(tile));
	}

	template<typename key_t, typename value_t, typename cmp_t>
	__device__ void shuffleSortBitonicSequence<key_t, value_t, cmp_t, 1>(key_t& key, value_t& value, cmp_t cmp, cooperative_groups::thread_block_tile<1> tile) {
	}

	template<typename key_t, typename value_t, typename cmp_t, typename group_t>
	__device__ void shuffleCreateBitonicSequence(key_t& key, value_t& value, cmp_t cmp, group_t group) {
		int tradeLaneId = (group.size() - group.thread_rank()) - 1;
		shuffleCmpSwap(key, value, tradeLaneId, cmp, group);
	}

	template<typename key_t, typename value_t, typename cmp_t, unsigned int TILE_SIZE>
	__device__ void shuffleMerge(key_t& key, value_t& value, cmp_t cmp, cooperative_groups::thread_block_tile<TILE_SIZE> tile) {
		// join the subsequences into one bitonic sequence
		shuffleCreateBitonicSequence(key, value, cmp, tile);
		// sort the bitonic sequence
		shuffleSortBitonicSequence(key, value, cmp, cooperative_groups::tiled_partition<TILE_SIZE / 2>(tile));
	}

	// performs a bitonic sort recursively
	template<typename key_t, typename value_t, typename cmp_t, unsigned int TILE_SIZE>
	__device__ void shuffleSort(key_t& key, value_t& value, cmp_t cmp, cooperative_groups::thread_block_tile<TILE_SIZE> tile) {
		// divide the sequence into two subsequences and sort each
		shuffleSort(key, value, cmp, cooperative_groups::tiled_partition<TILE_SIZE / 2>(tile));
		// merge the subsequences
		shuffleMerge(key, value, cmp, tile);
	}

	// special case: a sequence of one item (per tile) is already sorted
	template<typename key_t, typename value_t, typename cmp_t>
	__device__ void shuffleSort<key_t, value_t, cmp_t, 1>(key_t& key, value_t& value, cmp_t cmp, cooperative_groups::thread_block_tile<1> tile) {
	}

	template<unsigned int SEGMENT_WIDTH, typename key_t, typename value_t, typename cmp_t>
	__global__ void runSegmentedSortFixed(key_t* keys, value_t* values, cmp_t cmp) {
		cooperative_groups::thread_block wholeBlock = cooperative_groups::this_thread_block();

		int tid = wholeBlock.thread_rank() + blockIdx.x * wholeBlock.size();
		key_t key = keys[tid];
		value_t value = values[tid];

		shuffleSort(key, value, cmp, cooperative_groups::tiled_partition<SEGMENT_WIDTH>(wholeBlock));

		keys[tid] = key;
		values[tid] = value;
	}

	template<typename value_t, typename cmp_t>
	__device__ void cmpSwap(value_t& lhs, value_t& rhs, cmp_t cmp) {
		if (cmp(rhs, lhs)) {
			value_t tmp = lhs;
			lhs = rhs;
			rhs = tmp;
		}
	}

	template<typename value_t, typename cmp_t, typename group_t>
	__device__ void arrayMergeRedistribute(value_t* values, cmp_t cmp, const group_t& group) {
		unsigned halfSize = group.size() >> 1;
		if (group.thread_rank() < halfSize) {
			cmpSwap(values[group.thread_rank()], values[group.size() - group.thread_rank() - 1], cmp);
		}
	}

	template<typename value_t, typename cmp_t, typename group_t>
	__device__ void arraySortSplit(value_t* values, cmp_t cmp, const group_t& group) {
		// base case: a sequence of one item is already sorted
		if (group.size() == 1) {
			return;
		}

		unsigned halfSize = group.size() >> 1;
		if (group.thread_rank() < halfSize) {
			cmpSwap(values[group.thread_rank()], values[group.thread_rank() | halfSize], cmp);
		}
		cooperative_groups::thread_group tile = cooperative_groups::tiled_partition(group, halfSize);
		unsigned offset = group.thread_rank() & halfSize;
		arraySortSplit(values + offset, cmp, tile);
	}

	template<typename value_t, typename cmp_t, typename group_t>
	__device__ void arraySortedMerge(value_t* values, cmp_t cmp, const group_t& group) {
		// distribute elements among the arrays so that all items in the first compare below all items in the second
		arrayMergeRedistribute(values, cmp, group);
		group.sync();

		// sort each sub-array
		unsigned halfSize = group.size() >> 1;
		unsigned offset = group.thread_rank() & halfSize;
		cooperative_groups::thread_group tile = cooperative_groups::tiled_partition(group, halfSize);
		arraySortSplit(values + offset, cmp, tile);
	}

	template<typename value_t, typename cmp_t, typename group_t>
	__device__ void arraySort(value_t* values, cmp_t cmp, const group_t& group) {
		// base case: a sequence of one item is already sorted
		if (group.size() == 1) {
			return;
		}

		// divide the sequence in half and sort each half
		unsigned halfSize = group.size() >> 1;
		unsigned offset = group.thread_rank() & halfSize;
		cooperative_groups::thread_group tile = cooperative_groups::tiled_partition(group, halfSize);
		arraySort(values + offset, cmp, tile);

		group.sync();

		// merge the sorted halves
		arraySortedMerge(values, cmp, group);
	}

	template<typename value_t, typename cmp_t, typename group_t>
	__device__ void arraySegmentedMergeRedistribute(value_t* values, cmp_t cmp, unsigned logSegmentWidth, unsigned inputSize, const group_t& group) {
		if ((group.thread_rank() << 1) < inputSize) {
			unsigned laneMask = (1 << (logSegmentWidth - 1)) - 1;
			unsigned segmentOffset = (group.thread_rank() & ~laneMask) << 1;
			unsigned itemIdx0 = segmentOffset | (group.thread_rank() & laneMask);
			unsigned itemIdx1 = itemIdx0 ^ ((laneMask << 1) | 1);
			cmpSwap(values[itemIdx0], values[itemIdx1], cmp);
		}
	}

	template<typename value_t, typename cmp_t, typename group_t>
	__device__ void arraySegmentedSortSplit(value_t* values, cmp_t cmp, unsigned logSegmentWidth, unsigned inputSize, const group_t& group) {
		for (unsigned currentLogSegmentWidth = logSegmentWidth; currentLogSegmentWidth > 5; --currentLogSegmentWidth) {
			if ((group.thread_rank() << 1) < inputSize) {
				unsigned laneMask = (1 << (currentLogSegmentWidth - 1)) - 1;
				unsigned segmentOffset = (group.thread_rank() & ~laneMask) << 1;
				unsigned itemIdx0 = segmentOffset | (group.thread_rank() & laneMask);
				unsigned itemIdx1 = itemIdx0 | (1 << (currentLogSegmentWidth - 1));
				cmpSwap(values[itemIdx0], values[itemIdx1], cmp);
			}

			group.sync();
		}

		cooperative_groups::thread_block_tile<32> tile = cooperative_groups::tiled_partition<32>(group);
		// does this tile cover a piece of the input?
		if ((group.thread_rank() & ~31) < inputSize) {
			// yes, the tile has a piece of the input to sort
			for (unsigned int currentLogSegmentWidth = min(logSegmentWidth, 5); currentLogSegmentWidth > 0; --currentLogSegmentWidth) {
				if ((group.thread_rank() << 1) < inputSize) {
					unsigned laneMask = (1 << (currentLogSegmentWidth - 1)) - 1;
					unsigned segmentOffset = (group.thread_rank() & ~laneMask) << 1;
					unsigned itemIdx0 = segmentOffset | (group.thread_rank() & laneMask);
					unsigned itemIdx1 = itemIdx0 | (1 << (currentLogSegmentWidth - 1));
					cmpSwap(values[itemIdx0], values[itemIdx1], cmp);
				}

				tile.sync();
			}
		}
	}

	template<typename value_t, typename cmp_t, typename group_t>
	__device__ void arraySegmentedSortedMerge(value_t* values, cmp_t cmp, unsigned logSegmentWidth, unsigned inputSize, const group_t& group) {
		// distribute elements among the arrays so that all items in the first compare below all items in the second
		arraySegmentedMergeRedistribute(values, cmp, logSegmentWidth, inputSize, group);

		group.sync();

		// sort the redistributed arrays
		arraySegmentedSortSplit(values, cmp, logSegmentWidth - 1, inputSize, group);
	}

	template<typename value_t, typename cmp_t, typename group_t>
	__device__ void arraySegmentedSort(value_t* values, cmp_t cmp, unsigned logSegmentWidth, unsigned inputSize, const group_t& group) {
		// do the first few passes with a warp cooperative group for better synchronisation performance
		{
			cooperative_groups::thread_block_tile<32> tile = cooperative_groups::tiled_partition<32>(group);
			unsigned int firstTileValue = group.thread_rank() & ~31;
			unsigned int lastTileValue = min(firstTileValue + 32, inputSize);
			if (lastTileValue > firstTileValue) {
				for (unsigned int currentLogSegmentWidth = 1; currentLogSegmentWidth <= min(logSegmentWidth, 5); ++currentLogSegmentWidth) {
					arraySegmentedSortedMerge(values + firstTileValue, cmp, currentLogSegmentWidth, lastTileValue - firstTileValue, tile);
					tile.sync();
				}
			}
		}

		for (unsigned currentLogSegmentWidth = 5; currentLogSegmentWidth <= logSegmentWidth; ++currentLogSegmentWidth) {
			group.sync();
			arraySegmentedSortedMerge(values, cmp, currentLogSegmentWidth, inputSize, group);
		}
	}

	template<typename value_t, typename cmp_t>
	__global__ void runSegmentedSortShared(value_t* values, cmp_t cmp, unsigned logSegmentWidth) {
		__shared__ extern value_t sharedValues[];

		cooperative_groups::thread_block wholeBlock = cooperative_groups::this_thread_block();

		unsigned offset = blockIdx.x * wholeBlock.size();
		sharedValues[wholeBlock.thread_rank()] = values[offset + wholeBlock.thread_rank()];

		arraySegmentedSort(sharedValues, cmp, logSegmentWidth, 1 << logSegmentWidth, wholeBlock);

		values[offset + wholeBlock.thread_rank()] = sharedValues[wholeBlock.thread_rank()];
	}

	struct sum_t {
		template<typename item_t>
		__device__ int operator()(const item_t& lhs, const item_t& rhs) {
			return lhs + rhs;
		}
	};

	template<unsigned int RADIX_BITS, typename value_t, typename key_func_t, typename group_t>
	__device__ void radixSortInPlace(unsigned int laneId, value_t* values, key_func_t keyFunc, unsigned short* digitCounters, unsigned short numValues, const group_t& group) {
		unsigned int RADIX_DIGITS = 1 << RADIX_BITS;
		unsigned int MASK = RADIX_DIGITS - 1;
		unsigned int firstBit = 0;
		unsigned int lastBit = 32;

		for (unsigned int currentBit = firstBit; currentBit < lastBit; currentBit += RADIX_BITS) {
			// zero out the digit counters
			for (unsigned int workId = laneId; workId < numValues; workId += group.size()) {
				for (unsigned int digit = 0; digit < RADIX_DIGITS; ++digit) {
					digitCounters[digit * numValues + workId] = 0;
				}
			}

			// fill digit counters
			for (unsigned int workId = laneId; workId < numValues; workId += group.size()) {
				value_t value = values[workId];
				unsigned int digit = (keyFunc(value) >> currentBit) & MASK;
				digitCounters[digit * numValues + workId] += 1;
			}

			// prefix scan to rank the values
			workEfficientPrefixScan<unsigned short>(digitCounters, sum_t(), 0, numValues << RADIX_BITS, group);
			group.sync();

			for (unsigned int workId = laneId; workId < numValues; workId += group.size()) {
				value_t value = values[workId];
				unsigned int digit = (keyFunc(value) >> currentBit) & MASK;
				unsigned short writeIdx = digitCounters[digit * numValues + workId];
				values[writeIdx] = value;
			}
		}
	}

	template<unsigned int RADIX_BITS, typename value_t, typename key_func_t>
	__global__ void runSegmentedRadixSortShared(value_t* values, key_func_t keyFunc, unsigned int segmentWidth) {
		__shared__ extern value_t sharedValues[];
		unsigned short* digitCounters = (unsigned short*)&sharedValues[segmentWidth];

		cooperative_groups::thread_block wholeBlock = cooperative_groups::this_thread_block();

		value_t* segmentValues = values + (blockIdx.x * segmentWidth);
		sharedValues[wholeBlock.thread_rank()] = segmentValues[wholeBlock.thread_rank()];

		radixSortInPlace<RADIX_BITS>(wholeBlock.thread_rank(), sharedValues, keyFunc, digitCounters, segmentWidth, wholeBlock);

		segmentValues[wholeBlock.thread_rank()] = sharedValues[wholeBlock.thread_rank()];
	}

	struct record_dim_key_func_t {
		unsigned int dim;

		__device__ record_dim_key_func_t(unsigned int dim) : dim(dim) {}

		template<typename elem_t, int DIMS, typename key_t, typename data_t>
		__device__ const elem_t& operator()(const record_t<elem_t, DIMS, key_t, data_t>& record) const {
			return record.mbb.min[dim];
		}

		template<typename elem_t, int DIMS, typename key_t, typename data_t>
		__device__ elem_t& operator()(record_t<elem_t, DIMS, key_t, data_t>& record) const {
			return record.mbb.min[dim];
		}
	};

	struct record_dim_cmp_t {
		unsigned int dim;

		__device__ record_dim_cmp_t(unsigned int dim) : dim(dim) {}

		template<typename elem_t, int DIMS, typename key_t, typename data_t>
		__device__ bool operator()(const record_t<elem_t, DIMS, key_t, data_t>& lhs, const record_t<elem_t, DIMS, key_t, data_t>& rhs) const {
			return lhs.mbb.min[dim] < rhs.mbb.min[dim];
		}
	};

	template<unsigned int RADIX_BITS, typename elem_t, int DIMS, typename key_t, typename data_t, typename group_t>
	__device__ void tilePartitionInPlace(unsigned int laneId, record_t<elem_t, DIMS, key_t, data_t>* records, unsigned short* digitCounters, unsigned int numRecords, unsigned int fanout, group_t group) {
		unsigned int segmentWidth = numRecords;
		for (unsigned int dim = 0; dim < DIMS; ++dim) {
			unsigned int blockWidth = segmentWidth / fanout;

			/* unsigned int segmentId = laneId / segmentWidth;
			unsigned int segmentLaneId = laneId % segmentWidth;
			unsigned int segmentOffset = segmentId * segmentWidth;
			record_t<elem_t, DIMS, key_t, data_t>* segmentRecords = records + segmentOffset;
			unsigned short* segmentDigitCounters = digitCounters + (segmentOffset << RADIX_BITS);

			record_dim_key_func_t keyFunc(dim);

			radixSortInPlace<RADIX_BITS>(segmentLaneId, segmentRecords, keyFunc, segmentDigitCounters, segmentWidth, group); */

			record_dim_cmp_t recordCmp(dim);

			arraySegmentedSort(records, recordCmp, segmentWidth, numRecords, group);

			segmentWidth = blockWidth;
		}
	}

	template<unsigned int RADIX_BITS, typename elem_t, int DIMS, typename key_t, typename data_t>
	__global__ void runStrShared(record_t<elem_t, DIMS, key_t, data_t>* records, unsigned int segmentWidth, unsigned int fanout) {
		__shared__ extern record_t<elem_t, DIMS, key_t, data_t> sharedRecords[];
		unsigned short* digitCounters = (unsigned short*)&sharedRecords[segmentWidth];

		cooperative_groups::thread_block wholeBlock = cooperative_groups::this_thread_block();

		record_t<elem_t, DIMS, key_t, data_t>* segmentRecords = records + (blockIdx.x * segmentWidth);
		sharedRecords[wholeBlock.thread_rank()] = segmentRecords[wholeBlock.thread_rank()];

		tilePartitionInPlace<RADIX_BITS>(wholeBlock.thread_rank(), sharedRecords, digitCounters, segmentWidth, fanout, wholeBlock);

		segmentRecords[wholeBlock.thread_rank()] = sharedRecords[wholeBlock.thread_rank()];
	}
}

#endif //STR_SMALL_H
